////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// VMHistLite.cpp: implementation of the CVMHistLite class.
//
//////////////////////////////////////////////////////////////////////

#include <math.h>
#include <iostream>
#include <fstream>
#include <time.h>
//#include <limits.h>

#include "VMLib/include/vmlib.h"

#ifdef VM_HEP_LIB_FULL
/*
#include "VMOutput.h"
#include "VMHistLite.h"
#include "SysSettings.h"
*/

/////////////////////////////////////////////////////////////////////// IVMPlotData2D
const int IVMPlotData2D::PL_HIST		=1;
const int IVMPlotData2D::PL_SCAT		=2;
const int IVMPlotData2D::PL_FREEDRAW	=3;

IVMPlotData2D::IVMPlotData2D(const char* name) {
	m_name=IVMPlotData2D::GetStrCopy(name); if(m_name==NULL)m_name=IVMPlotData2D::GetStrCopy("Error");
}
IVMPlotData2D::~IVMPlotData2D() {
	if (m_name!=NULL) delete[] m_name;
}
const char* IVMPlotData2D::GetName() { return m_name; }
const char* IVMPlotData2D::GetLegendTxt() { return GetName(); }

// local library function to allow stand-alone class usage & possibly filter symbols
#define CVMHist_StrCpy_MAX	1000
char* IVMPlotData2D::GetStrCopy(const char* s) {
	if(s==NULL)return NULL;
	long l=-1; long i;
	for(i=0;i<CVMHist_StrCpy_MAX;i++){ if(s[i]=='\0'){l=i;break;} } if(l<0)return NULL;
	char* res=new char[l+2]; for(i=0;i<l+1;i++){res[i]=s[i];} // no sprintf here to allow %-symbols
	return res;
}
void IVMPlotData2D::SetName(const char* name) {
	if(m_name != name){ //Important: 'otherwise' may occur
		if(m_name!=NULL){ delete[] m_name; m_name=NULL; }
		m_name=IVMPlotData2D::GetStrCopy(name); if(m_name==NULL)m_name=IVMPlotData2D::GetStrCopy("Error");
	}
}

/////////////////////////////////////////////////////////////////////// CVMHistLite
CVMHistLite::CVMHistLite():IVMPlotData2D("Not_inited") {
	// To be initialized via Init() call
	m_nBins=0;
	m_values=NULL; m_valsSqr=NULL; m_valSdSqr=NULL; m_margins=NULL; m_nBinEvents=NULL;
	m_nEvents=0; m_SumValue=0.; m_SumValueAll=0.; m_NormalizeIntegral=-1; m_yMaxValue=0; m_txtLegend=NULL;
	m_bAutoXMode=false; m_nAutoXTestSize=-1; m_nAutoXTestEvt=0; m_bufAutoXFills=NULL; m_bLogXAfterAutoMode=false;
	m_Ready=false;
}

/*
*	Warning: this function content should exacttly meet Init() function
*	Note: Calling Init() from constructor seems not very good idea.
*/
CVMHistLite::CVMHistLite(const char* name, unsigned nBins, double xMin, double xMax, bool bLogX):IVMPlotData2D(name) {
	m_nBins=nBins; if(m_nBins<=0)m_nBins=1;
	m_values=new double[m_nBins+2]; m_valsSqr=new double[m_nBins+2]; m_valSdSqr=new double[m_nBins+2]; m_margins=new double[m_nBins+1]; m_nBinEvents=new unsigned long[m_nBins+2];
	m_nEvents=0; m_SumValue=0.; m_SumValueAll=0.; m_NormalizeIntegral=-1; m_yMaxValue=0; m_txtLegend=NULL;
	unsigned i; for(i=0; i<(m_nBins+2); i++) { m_values[i]=0.; m_valsSqr[i]=0.; m_valSdSqr[i]=0.; m_nBinEvents[i]=0; }
	if ((bLogX) && (xMax>0)) {
		if(xMin<=0)xMin=xMax/1000;
		double lnMin=log(xMin); double lnMax=log(xMax); for(i=1; i<m_nBins; i++) { m_margins[i]=exp(lnMin+(lnMax-lnMin)*i/(m_nBins));}
		m_margins[0]=xMin; m_margins[m_nBins]=xMax; // fill separatelly to avoid 1-a+a artifacts
	} else {
		for(i=0; i<m_nBins; i++) { m_margins[i]=xMin+(xMax-xMin)*i/m_nBins; }
		m_margins[m_nBins]=xMax; // fill separatelly to avoid 1-a+a artifacts
	}
	m_bAutoXMode=false; m_nAutoXTestSize=-1; m_nAutoXTestEvt=0; m_bufAutoXFills=NULL; m_bLogXAfterAutoMode=false;
	m_Ready=true;
}

CVMHistLite::CVMHistLite(const char* name, unsigned nBins, double* xBinMargins):IVMPlotData2D(name) {
	m_nBins=nBins;
	m_values=new double[m_nBins+2]; m_valsSqr=new double[m_nBins+2]; m_valSdSqr=new double[m_nBins+2]; m_margins=new double[m_nBins+1]; m_nBinEvents=new unsigned long[m_nBins+2];
	m_nEvents=0; m_SumValue=0.; m_SumValueAll=0.; m_NormalizeIntegral=-1; m_yMaxValue=0; m_txtLegend=NULL;
	unsigned i;
	for(i=0; i<(m_nBins+2); i++) { m_values[i]=0.; m_valsSqr[i]=0.; m_valSdSqr[i]=0.; m_nBinEvents[i]=0; }
	for(i=0; i<=m_nBins; i++) { m_margins[i]=xBinMargins[i]; }
	m_bAutoXMode=false; m_nAutoXTestSize=-1; m_nAutoXTestEvt=0; m_bufAutoXFills=NULL; m_bLogXAfterAutoMode=false;
	m_Ready=true;
}

CVMHistLite::CVMHistLite(const CVMHistLite& hist):IVMPlotData2D(hist.m_name) {
	m_nBins=hist.m_nBins;
	m_values=new double[m_nBins+2]; m_valsSqr=new double[m_nBins+2]; m_valSdSqr=new double[m_nBins+2]; m_margins=new double[m_nBins+1]; m_nBinEvents=new unsigned long[m_nBins+2];
	m_nEvents=hist.m_nEvents; m_SumValue=hist.m_SumValue; m_SumValueAll=hist.m_SumValueAll; m_NormalizeIntegral=hist.m_NormalizeIntegral; m_yMaxValue=hist.m_yMaxValue;
	m_txtLegend=((hist.m_txtLegend!=NULL)?IVMPlotData2D::GetStrCopy(hist.m_txtLegend):NULL);
	for(unsigned i=0; i<(m_nBins+2); i++) { m_values[i]=hist.m_values[i]; m_valsSqr[i]=hist.m_valsSqr[i]; m_valSdSqr[i]=hist.m_valSdSqr[i]; m_nBinEvents[i]=hist.m_nBinEvents[i]; if(i<=m_nBins)m_margins[i]=hist.m_margins[i]; }
	m_bAutoXMode=false; m_nAutoXTestSize=-1; m_nAutoXTestEvt=0; m_bufAutoXFills=NULL; m_bLogXAfterAutoMode=false;
	m_Ready=true;
}

CVMHistLite::~CVMHistLite() {
	if (m_values!=NULL) delete[] m_values;
	if (m_valsSqr!=NULL) delete[] m_valsSqr;
	if (m_valSdSqr!=NULL) delete[] m_valSdSqr;
	if (m_margins!=NULL) delete[] m_margins;
	if (m_nBinEvents!=NULL) delete[] m_nBinEvents;
	if (m_txtLegend!=NULL) delete[] m_txtLegend;
	if (m_bufAutoXFills!=NULL) { delete[] m_bufAutoXFills; m_bufAutoXFills=NULL; }
}
int CVMHistLite::GetClassId() { return IVMPlotData2D::PL_HIST; }

void CVMHistLite::Init(const char* name, unsigned nBins, double xMin, double xMax, bool bLogX) {
	if (m_Ready || m_bAutoXMode) { std::cerr<<"ERROR: CVMHistLite::Init call for non-clear histogram"<<std::endl; return; } // Error
	m_nBins=nBins; if(m_nBins<=0)m_nBins=1;
	m_values=new double[m_nBins+2]; m_valsSqr=new double[m_nBins+2]; m_valSdSqr=new double[m_nBins+2]; m_margins=new double[m_nBins+1]; m_nBinEvents=new unsigned long[m_nBins+2];
	m_nEvents=0; m_SumValue=0.; m_SumValueAll=0.; m_NormalizeIntegral=-1; m_yMaxValue=0;
	SetName(name);
	unsigned i;
	for(i=0; i<(m_nBins+2); i++) { m_values[i]=0.; m_valsSqr[i]=0.; m_valSdSqr[i]=0.; m_nBinEvents[i]=0; }
	if ((bLogX) && (xMax>0)) {
		if(xMin<=0)xMin=xMax/1000;
		double lnMin=log(xMin); double lnMax=log(xMax); for(i=1; i<m_nBins; i++) { m_margins[i]=exp(lnMin+(lnMax-lnMin)*i/(m_nBins));}
		m_margins[0]=xMin; m_margins[m_nBins]=xMax; // fill separatelly to avoid 1-a+a artifacts
	} else {
		for(i=0; i<m_nBins; i++) { m_margins[i]=xMin+(xMax-xMin)*i/m_nBins; }
		m_margins[m_nBins]=xMax; // fill separatelly to avoid 1-a+a artifacts
	}
	m_Ready=true;
}
void CVMHistLite::InitAutoX(const char* name, unsigned nBins, unsigned nTestEvts, bool bLogX) {
	if (m_Ready || m_bAutoXMode) { std::cerr<<"ERROR: CVMHistLite::InitAutoX call for non-clear histogram"<<std::endl; return; } // Error
	if(m_bufAutoXFills!=NULL) { delete[] m_bufAutoXFills; m_bufAutoXFills=NULL; }
	m_nAutoXTestSize=nTestEvts; m_nAutoXTestEvt=0; m_bLogXAfterAutoMode=bLogX;
	if(m_nAutoXTestSize > 1000) { std::cerr<<"WARNING in CVMHistLite::InitAutoX: maximum number of test events set to 1000"<<std::endl; m_nAutoXTestSize=1000; }
	if(m_nAutoXTestSize>=1) {
		m_bufAutoXFills = new double[m_nAutoXTestSize*3];
		if(m_bufAutoXFills != NULL) {
			SetName(name); m_nBins=nBins;
			m_bAutoXMode=true;
		}
	}
}

void CVMHistLite::SetLegendText(const char* txt) { // histogram name is used by default
	if (m_txtLegend!=NULL) { delete[] m_txtLegend; m_txtLegend=NULL; }
	m_txtLegend=((txt!=NULL)?IVMPlotData2D::GetStrCopy(txt):NULL);
}
void CVMHistLite::Reset() {
	if(!m_Ready)return;
	m_nEvents=0; m_SumValue=0.; m_SumValueAll=0.; m_NormalizeIntegral=-1; m_yMaxValue=0;
	for(unsigned i=0; i<(m_nBins+2); i++) { m_values[i]=0.; m_valsSqr[i]=0.; m_valSdSqr[i]=0.; m_nBinEvents[i]=0; }
}
// returns: 1+n, (m_margins[n]<=value<m_margins[n+1]), n==0 is underflow, n==m_nBins+1 is overflow, -1 for error
int CVMHistLite::GetBinNoInternal(double value) {
	if ((!m_Ready) || (m_nBins==0)) return -1; int res=-1;
	if (value < m_margins[0]) { res=0; }
	else if (value >= m_margins[m_nBins]) { res=m_nBins+1; }
	else {
		int i1=0; int i2=m_nBins;
		for (;;) {
			res=i1+(i2-i1)/2;
			if(value>m_margins[res]) { i1=res; if(i2>(i1+1))continue; }
			else if (value<m_margins[res]) { i2=res; if(i2>(i1+1))continue; res=i1; }
			break;
		}
		res++; // to fit interval [1..m_nBins]
	}
	return res;
}

void CVMHistLite::Fill(double val, double wgt, double wgtErr) {
	if(m_bAutoXMode) {
		m_bufAutoXFills[m_nAutoXTestEvt*3]=val; m_bufAutoXFills[m_nAutoXTestEvt*3+1]=wgt; m_bufAutoXFills[m_nAutoXTestEvt*3+2]=wgtErr;
		m_nAutoXTestEvt++;
		if(m_nAutoXTestEvt>=m_nAutoXTestSize) {
			//Select xMin & xMax
			double xMin=0.; double xMax=0.; unsigned i;
			for(i=0; i<m_nAutoXTestSize; i++) {
				double x=m_bufAutoXFills[i*3]; if(i==0){ xMin=x; xMax=x; }else{ if(xMin>x)xMin=x; if(xMax<x)xMax=x; }
			}
			//
			double gapFactor=1.3;
			//avoid margins (0.7..1.3) for plots like (0.99..1.01)
			if((xMax>0.) && (xMin>0.)) { if((xMax/xMin-1.)*2.<(gapFactor-1.)) gapFactor=1.+(xMax/xMin-1.)*2.; }
			if((xMax<0.) && (xMin<0.)) { if((xMin/xMax-1.)*2.<(gapFactor-1.)) gapFactor=1.+(xMin/xMax-1.)*2.; }
			if(xMin>=0){ xMin/=gapFactor; }else{ xMin*=gapFactor; }
			if(xMax>=0){ xMax*=gapFactor; }else{ xMax/=gapFactor; }
			if((xMax>0.) && (xMin>0.)) { if(xMax/xMin>10.)xMin=0.; }
			if((xMax<0.) && (xMin<0.)) { if(xMin/xMax>10.)xMax=0.; }
			if(xMin==xMax){ xMin=0.; xMax=1.; std::cerr<<"[ERROR] Not enough data to estimate X-margins of "<<m_name<<" histogram"<<std::endl; }
			m_bAutoXMode=false; if(m_Ready){ std::cerr<<"Internal error in CVMHistLite::Fill"<<std::endl; m_Ready=false; }
			//Init histogram
			Init(m_name,m_nBins,xMin,xMax,m_bLogXAfterAutoMode);
			//Re-fill all values
			for(i=0; i<m_nAutoXTestSize; i++) {
				double x=m_bufAutoXFills[i*3]; double wgt=m_bufAutoXFills[i*3+1]; double wgtErr=m_bufAutoXFills[i*3+2];
				Fill(x,wgt,wgtErr);
			}
			//Finalize mode
			delete[] m_bufAutoXFills; m_bufAutoXFills=NULL;
		}
		return;
	}

	// Check numbers //TODO: uniform it!
	// Same check to (CAlHepLib::IsNaN(val) || CAlHepLib::IsInf(val))
	if ((wgt!=wgt) || ((((unsigned long*)&wgt)[1] & 0x7ff80000) == 0x7ff00000)) return;
	if ((wgtErr!=wgtErr) || ((((unsigned long*)&wgtErr)[1] & 0x7ff80000) == 0x7ff00000)) return;

	if (!m_Ready) return;
	m_SumValueAll += wgt;
	int iBin=GetBinNoInternal(val); // iBin==0 is underflow, iBin==m_nBins+1 is overflow
	if (iBin>=0) {
		m_values[iBin]+=wgt; m_valsSqr[iBin]+=wgt*wgt; m_valSdSqr[iBin]+=wgtErr*wgtErr; m_nBinEvents[iBin]++;
		m_nEvents++; // including outmargins
		if((iBin>0) && (iBin<=(int)m_nBins)){ //real bin
			m_SumValue+=wgt; // excluding outmargins
			if((m_nEvents==0) || (m_values[iBin]>m_yMaxValue)) m_yMaxValue=m_values[iBin];
		}
	}
}
/*
*	FillMax: replace previous weight with 'wgt' if newest is greather.
*	N[evetns/bin] is set to 1.
*	TODO: move this function to 'WgtMax[i]' array of usual Fill()?
*		But unusual HistView-method will be required.
*/
void CVMHistLite::FillMax(double val, double wgt) {
	// Check numbers //TODO: uniform it!
	// Same check to (CAlHepLib::IsNaN(val) || CAlHepLib::IsInf(val))
	if ((wgt!=wgt) || ((((unsigned long*)&wgt)[1] & 0x7ff80000) == 0x7ff00000)) return;

	if (!m_Ready) return;
	int iBin=GetBinNoInternal(val); // iBin==0 is underflow, iBin==m_nBins+1 is overflow
	if ((iBin>=0) && (m_values[iBin]<wgt)) {
		double wgtOld=m_values[iBin]; m_values[iBin]=wgt; m_valsSqr[iBin]=wgt*wgt; m_valSdSqr[iBin]=0.;
		if(m_nBinEvents[iBin]==0)m_nEvents++; //don't change nEvents if weight is replaced
		m_nBinEvents[iBin]=1;
		if((iBin>0) && (iBin<=(int)m_nBins)){ //real bin
			m_SumValue+=(wgt-wgtOld);
			if((m_nEvents==0) || (m_values[iBin]>m_yMaxValue)) m_yMaxValue=m_values[iBin];
		}
		m_SumValueAll+=(wgt-wgtOld);
	}
}
double CVMHistLite::GetValue(unsigned i, double* err) {
	if (!m_Ready) return 0.;
	double res=0.;
	if (i<m_nBins) { //unsigned i ==> (i>=0)
		res=m_values[i+1];
		if(err!=NULL) {
			double _Err=((m_valSdSqr[i+1]>0.)?sqrt(m_valSdSqr[i+1]):0.); // To prevent Err=10^-15 artifacts
// Note: Stat. error is probably included into m_valSdSqr[] values, see Notebook (22.3)
			bool bIncludeStatError=false; //bool bIncludeStatError=(_Err==0.);
			if(bIncludeStatError) {//TODO: calculate stat. error correctly
				unsigned long n=m_nBinEvents[i+1];
				if(n>0) {
					_Err+=sqrt(fabs(m_valsSqr[i+1]/n-(res*res)/(n*n))); // Sig=sqrt(<x2>-<x><x>) --> 0 for n==1
					//if(n==1)_Err+=res; // Sig~sqrt(1/n) --> //TODO: calculate error correctly!
				}
			}
			if((m_NormalizeIntegral>=0.)&&(m_SumValue>0.))_Err*=m_NormalizeIntegral/m_SumValue;
			*err=_Err;
		}
		if((m_NormalizeIntegral>=0.)&&(m_SumValue>0.))res*=m_NormalizeIntegral/m_SumValue;
	}
	return res;
}
double CVMHistLite::GetValueOrOverflow(unsigned i) { // 0<=i<=(nDim+1), first&last=outmargins
	if ((!m_Ready) || (i>(m_nBins+1))) return 0.;
	double res=m_values[i]; if((m_NormalizeIntegral>=0.)&&(m_SumValue>0.))res*=m_NormalizeIntegral/m_SumValue;
	return res;
}
double CVMHistLite::operator()(unsigned i) { return GetValue(i,NULL); }
unsigned CVMHistLite::GetNBins() { return m_nBins; }
unsigned long CVMHistLite::GetNumEvents() { return m_nEvents; }
unsigned long CVMHistLite::GetNumEventsBin(unsigned i) { if(!m_Ready)return 0; return ((i<m_nBins)?m_nBinEvents[i+1]:0); } //unsigned i ==> (i>=0)
unsigned long CVMHistLite::GetNumOutmargins() { return (m_nBinEvents[0]+m_nBinEvents[m_nBins+1]); }
double CVMHistLite::GetMaxXValue() { if(!m_Ready)return -1; return ((m_nBins>0)?m_margins[m_nBins]:0); }
double CVMHistLite::GetMinXValue() { if(!m_Ready)return -1; return ((m_nBins>0)?m_margins[0]:0); }
double CVMHistLite::GetMaxYValue() {
	if(!m_Ready)return -1;
	double res=m_yMaxValue;
	if((m_NormalizeIntegral>=0.)&&(m_SumValue>0.))res*=m_NormalizeIntegral/m_SumValue;
	return res;
}
double CVMHistLite::GetMinYValue() {
	if(!m_Ready)return -1;
	double res=m_yMaxValue;
	if(m_nEvents>0) {
		for(unsigned i=1; i<=m_nBins; i++) {
			if ((m_nBinEvents[i]>0) && (m_values[i]<res)) res=m_values[i];
		}
	} else { res=0; } // no data
	if((m_NormalizeIntegral>=0.)&&(m_SumValue>0.))res*=m_NormalizeIntegral/m_SumValue;
	return res;
}
double CVMHistLite::GetBinXMargin(unsigned i) { if(!m_Ready)return -1; return ((i<=m_nBins)?m_margins[i]:0); } //unsigned i ==> (i>=0)
double CVMHistLite::GetBinXCenter(unsigned i) { if(!m_Ready)return -1; return ((i<m_nBins)?(m_margins[i]+m_margins[i+1])/2.:0); } //unsigned i ==> (i>=0)
double CVMHistLite::GetBinXWidth(unsigned i) { if(!m_Ready)return 0; return ((i<m_nBins)?(m_margins[i+1]-m_margins[i]):0); } //unsigned i ==> (i>=0)
bool CVMHistLite::FitXRegion(double x) { return ((x>=GetMinXValue())&&(x<=GetMaxXValue())); }
const char* CVMHistLite::GetLegendTxt() { return((m_txtLegend!=NULL)?m_txtLegend:m_name); }
bool CVMHistLite::IsEmpty() { return (m_nEvents<=0); }
//
double CVMHistLite::GetIntegral() { //[excl. outmargins]
	return ((m_NormalizeIntegral>=0.)?m_NormalizeIntegral:m_SumValue);
}
//Normalization operations
/*
*	Normalize: Set integral value (as global factor).
*	Doesn't modify bins ==> may be called BEFORE filling.
*	Normalize(-1) removes previously set value.
*/
void CVMHistLite::Normalize(double _Integral) { // value*=_NormValue/m_SumValue;
	m_NormalizeIntegral=((_Integral<0)?-1.:_Integral);
}
/*
*	NormalizeAdv: Scale to new integral value (scaling all bins).
*	Modifies bins ==> may be called AFTER filling only.
*/
void CVMHistLite::NormalizeAdv(double _Integral) {
	if((!m_Ready) || (m_nEvents<=0) || (_Integral<0.))return;
	if(m_SumValueAll<=0.){ std::cout<<"[ERROR] Can't normalize histogram. Previous weight <= 0."<<std::endl; return; }
	double fScale=_Integral/m_SumValueAll; double fScale2=fScale*fScale; unsigned iStart=0;
	for(unsigned i=iStart; i<m_nBins; i++) {
		m_values[i+1]*=fScale; m_valsSqr[i+1]*=fScale2; m_valSdSqr[i+1]*=fScale2;
		if((i==iStart) || (m_values[i+1]>m_yMaxValue))m_yMaxValue=m_values[i+1]; // update m_yMaxValue
	}
	if(m_SumValue!=m_SumValueAll){m_SumValue*=fScale;}else{m_SumValue=_Integral;} m_SumValueAll=_Integral;
}
/*
*	[NormalizeToBinWidth] To df/dx (after filled with weight=='Integal/Nevts'):  BinValue/=dx
*	May not be followed by filling, averaging or re-normalization.
*	WARNING: fill, average or re-normalize is NOT automatically forbidden after this call.
*/
void CVMHistLite::NormalizeToBinWidth() {
	if((!m_Ready) || (m_nEvents<=0))return;
	unsigned iStart=0;
	for(unsigned i=iStart; i<m_nBins; i++) {
		double dx=GetBinXWidth(i); if(dx<=0.)continue;
		m_values[i+1]/=dx; m_valsSqr[i+1]/=(dx*dx); m_valSdSqr[i+1]/=(dx*dx);
		if((i==iStart) || (m_values[i+1]>m_yMaxValue))m_yMaxValue=m_values[i+1]; // update m_yMaxValue
	}
}
/*
*	NormalizeByDiffBins: Restore integral value after 'df/dx' is filled in every bin (1 event/bin).
*/
void CVMHistLite::NormalizeByDiffBins() {
	if(!m_Ready)return;
	if(m_nEvents!=m_nBins) { std::cout<<"[ERROR in CVMHistLite::NormalizeByDiffBins()]. Historgam must filled be filled with 1 event per bin."<<std::endl; return; }
	double _Integral=0.;
	for(unsigned i=0; i<m_nBins; i++) {
		if(m_nBinEvents[i+1]!=1){ std::cout<<"[ERROR in CVMHistLite::NormalizeByDiffBins()]. Historgam must filled be filled with 1 event per bin."<<std::endl; return; }
		double dx=GetBinXWidth(i); if(dx<=0.)continue;
		_Integral+=m_values[i+1]*dx;
	}
	if(m_SumValueAll != m_SumValue){ //Never occurs if (m_nEvents==m_nBins).
		if(m_SumValue>0.){ m_SumValueAll*=(_Integral/m_SumValue); }else{ std::cout<<"[ERROR in CVMHistLite::NormalizeByDiffBins()]. Invalid previous weights sum value."<<std::endl; m_SumValueAll=-1.; }
	}else{ m_SumValueAll=_Integral; }
	m_SumValue=_Integral;
}
/*
*	[Factor] Factorize all values in histogram by f (multiply previously filled weights: 'wgt'->'wgt*f').
*/
void CVMHistLite::Factor(double f) {
	if(f<0.){ std::cout<<"[ERROR in CVMHistLite::Factor()]. Negative factor."<<std::endl; return; }
	//[Readiness of the histogram is not required] if((!m_Ready) || (m_nEvents<=0) || (_Integral<0.))return;
	m_SumValueAll*=f; m_SumValue*=f;
	for(unsigned i=0; i<(m_nBins+2); i++) { m_values[i]*=f; m_valsSqr[i]*=(f*f); m_valSdSqr[i]*=(f*f); }
	m_yMaxValue*=f;
}
void CVMHistLite::AverageBinValues() { // BinValue=BinValue/NEventsInBin
	if(!m_Ready)return;
	//TODO: update total sum, outmargins & others
	unsigned iStart=1;
	for(unsigned i=iStart; i<=m_nBins; i++) { // don't average overflow values to keep correct N-Outmargins
		if (m_nBinEvents[i]>1) { m_values[i]/=m_nBinEvents[i]; m_nBinEvents[i]=1; } // nBinEvt=1 to avoid second averaging
		if ((i==iStart) || (m_values[i]>m_yMaxValue)) m_yMaxValue=m_values[i]; // update m_yMaxValue
	}
}
//Histogram operations
void CVMHistLite::AddHisto(CVMHistLite* h, double hFactor){
	if((!m_Ready) || (h==NULL))return;
	if((m_nBins == h->GetNBins()) && (GetMinXValue() == h->GetMinXValue()) && (GetMaxXValue() == h->GetMaxXValue())) {
		double val=0.; double err=0.; double fBinsSum=0.;
		//TODO: clear outmargins
		for(unsigned i=0; i<m_nBins; i++) {
			val=h->GetValue(i,&err); val*=hFactor; err*=hFactor; double valNew=m_values[i+1]+val;
			m_values[i+1]=valNew;
			m_valSdSqr[i+1]+=err*err;
			//truncate statistics
			m_nBinEvents[i+1]=1; m_valsSqr[i+1]=valNew*valNew; fBinsSum+=valNew;
			//TODO: implement 'sum{sqr}', update or (!) truncate statistics & leave 'm_nBins' events
			if((i==0) || (m_values[i+1]>m_yMaxValue))m_yMaxValue=m_values[i+1]; // update m_yMaxValue
		}
		//truncate statistics & normalization
		m_NormalizeIntegral=-1.; m_nEvents=m_nBins; m_SumValueAll=fBinsSum; m_SumValue=fBinsSum; //set to 'sum-of-bins'
	}else{ std::cout<<"Error in CVMHistLite::AddHisto(): imcompartible histograms"<<std::endl; }
}
void CVMHistLite::DivideHisto(CVMHistLite* h){
	if((!m_Ready) || (h==NULL))return;
	if((m_nBins == h->GetNBins()) && (GetMinXValue() == h->GetMinXValue()) && (GetMaxXValue() == h->GetMaxXValue())) {
		double val=0.; double err=0.; double fBinsSum=0.;
		//TODO: clear outmargins
		for(unsigned i=0; i<m_nBins; i++) {
			val=h->GetValue(i,&err); double valNew=((val!=0.)?(m_values[i+1]/val):0.);
			m_values[i+1]=valNew;
			m_valSdSqr[i+1]=0.; //TODO: implement 'err'
			//truncate statistics
			m_nBinEvents[i+1]=1; m_valsSqr[i+1]=valNew*valNew; fBinsSum+=valNew;
			//TODO: implement 'sum{sqr}', update or (!) truncate statistics & leave 'm_nBins' events
			if((i==0) || (m_values[i+1]>m_yMaxValue))m_yMaxValue=m_values[i+1]; // update m_yMaxValue
		}
		//truncate statistics & normalization
		m_NormalizeIntegral=-1.; m_nEvents=m_nBins; m_SumValueAll=fBinsSum; m_SumValue=fBinsSum; //set to 'sum-of-bins'
	}else{ std::cout<<"Error in CVMHistLite::DivideHisto(): imcompartible histograms"<<std::endl; }
}
//Save
bool CVMHistLite::SaveAsTable(const char* fn, bool bAddColsNumber) {
	if(!m_Ready)return false;
	if (fn==NULL) return false;
	std::ofstream ofs(fn); if(!ofs)return false;
	if (bAddColsNumber) { ofs<<2<<std::endl; } else { ofs<<"Title: "<<m_name<<std::endl; } // add title string
	for(unsigned i=0; i<m_nBins; i++) { ofs<<GetBinXCenter(i)<<" "<<GetValue(i)<<std::endl; }
	ofs.close();
	return true;
}
bool CVMHistLite::SaveAsObject(CVMFileOut* f) {
	if(f==NULL)return false;
	if(f->IsBinary()){ std::cout<<"ERROR in CVMHistLite::SaveAsObject: binary file"<<std::endl; return false; }
	*f<<"["<<((long)CVMHistLiteVER)<<"]";
	*f<<"["<<CVMUtil::strLen(m_name)<<"<"<<m_name<<">]";
	const char* strLegend=((m_txtLegend!=NULL)?m_txtLegend:"");
	*f<<"["<<CVMUtil::strLen(strLegend)<<"<"<<strLegend<<">]S\n";
	*f<<(long)m_nBins<<" "<<(long)m_nEvents<<" "<<m_SumValue<<" "<<m_SumValueAll<<" "<<m_NormalizeIntegral<<" "<<m_yMaxValue<<"\n";
	for(unsigned int i=0; i<(m_nBins+2); i++){
		*f<<m_values[i]<<" "<<m_valsSqr[i]<<" "<<m_valSdSqr[i]<<" "<<(long)(m_nBinEvents[i]); if(i<=m_nBins){*f<<" "<<m_margins[i];} *f<<"\n";
	}
	*f<<"E\n";
	return true;
}
bool CVMHistLite::LoadAsObject(CVMFileIn* f) {
	m_Ready=false;
	if(f==NULL)return false; if(!f->IsReady()){ std::cout<<"CVMHistLite::LoadAsObject: error reading file"<<std::endl; return false; }
	if(f->IsBinary()){ std::cout<<"ERROR in CVMHistLite::LoadAsObject: binary file"<<std::endl; return false; }
	char c; long iVer,szName,szLegend;
	if(!f->ReadNotEmptySymbol(c) || (c!='[')){ std::cout<<"CVMHistLite::LoadAsObject: '[' expected"<<std::endl; return false; }
	*f>>iVer; if(iVer!=CVMHistLiteVER){ std::cout<<"Error in CVMHistLite::LoadAsObject: invalid object version"<<std::endl; return false; }
	if(!f->ReadNotEmptySymbol(c) || (c!=']')){ std::cout<<"CVMHistLite::LoadAsObject: ']' expected"<<std::endl; return false; }
	if(!f->ReadNotEmptySymbol(c) || (c!='[')){ std::cout<<"CVMHistLite::LoadAsObject: '[' expected"<<std::endl; return false; }
	*f>>szName; if(szName<0){ std::cout<<"CVMHistLite::LoadAsObject: invalid string size"<<std::endl; return false; }
	char* bufN=CVMUtil::strAlloc(szName+10); if(bufN==NULL){ std::cout<<"CVMHistLite::LoadAsObject: memory error"<<std::endl; return false; }
	if(!f->ReadNotEmptySymbol(c) || (c!='<')){ CVMUtil::strFree(bufN); std::cout<<"CVMHistLite::LoadAsObject: '<' expected"<<std::endl; return false; }
	if(!f->ReadString(bufN,szName+8,">]")){ CVMUtil::strFree(bufN); std::cout<<"CVMHistLite::LoadAsObject: '>]' expected"<<std::endl; return false; }
	//[debug] std::cout<<"bufN: "<<bufN<<std::endl;
	SetName(bufN); CVMUtil::strFree(bufN);
	if(!f->ReadNotEmptySymbol(c) || (c!='[')){ std::cout<<"CVMHistLite::LoadAsObject: '[' expected"<<std::endl; return false; }
	*f>>szLegend; if(szLegend<0){ std::cout<<"CVMHistLite::LoadAsObject: invalid string size"<<std::endl; return false; }
	char* bufL=CVMUtil::strAlloc(szLegend+10); if(bufL==NULL){ std::cout<<"CVMHistLite::LoadAsObject: memory error"<<std::endl; return false; }
	if(!f->ReadNotEmptySymbol(c) || (c!='<')){ CVMUtil::strFree(bufL); std::cout<<"CVMHistLite::LoadAsObject: '<' expected"<<std::endl; return false; }
	if(!f->ReadString(bufL,szLegend+8,">]")){ CVMUtil::strFree(bufL); std::cout<<"CVMHistLite::LoadAsObject: '>]' expected"<<std::endl; return false; }
	SetLegendText((szLegend>0)?bufL:NULL); CVMUtil::strFree(bufL);
	if(!f->ReadNotEmptySymbol(c) || (c!='S')){ std::cout<<"CVMHistLite::LoadAsObject: 'S' expected"<<std::endl; return false; }
	*f>>m_nBins>>m_nEvents>>m_SumValue>>m_SumValueAll>>m_NormalizeIntegral>>m_yMaxValue; //Wrong for binary file
	//[debug]std::cout<<"m_nBins="<<m_nBins<<", m_nEvents="<<m_nEvents<<", m_SumValue="<<m_SumValue<<", m_SumValueAll="<<m_SumValueAll<<", m_NormalizeIntegral="<<m_NormalizeIntegral<<", m_yMaxValue="<<m_yMaxValue<<std::endl;
	m_values=new double[m_nBins+2]; m_valsSqr=new double[m_nBins+2]; m_valSdSqr=new double[m_nBins+2]; m_margins=new double[m_nBins+1]; m_nBinEvents=new unsigned long[m_nBins+2];
	for(unsigned int i=0; i<(m_nBins+2); i++){
		*f>>m_values[i]>>m_valsSqr[i]>>m_valSdSqr[i]>>m_nBinEvents[i]; if(i<=m_nBins){*f>>m_margins[i];}
		//[debug]std::cout<<"m_values[i]="<<m_values[i]<<", m_valsSqr[i]="<<m_valsSqr[i]<<", m_valSdSqr[i]="<<m_valSdSqr[i]<<", m_nBinEvents[i]="<<m_nBinEvents[i]<<", m_margins[i]="<<m_margins[i]<<std::endl;
		//[debug]std::cout<<" File ready ["<<i<<"]= "<<f->IsReady()<<std::endl; if(!f->IsReady())break;
	}
	if(!f->ReadNotEmptySymbol(c) || (c!='E')){ std::cout<<"CVMHistLite::LoadAsObject: 'E' expected (read: "<<c<<")"<<std::endl; return false; }
	m_Ready=true;
	return m_Ready;
}

/////////////////////////////////////////////////////////////////////// CVMScatLite
#define VMScat_MAXEVTS	10000
CVMScatLite::CVMScatLite():IVMPlotData2D("Unknown") {
	m_valX = new double[VMScat_MAXEVTS];
	m_valY = new double[VMScat_MAXEVTS];
	m_valWgt = new double[VMScat_MAXEVTS];
	m_nEvents=0; m_bOverfilled=false; m_bAutoAverage=true; m_iAutoAverageNx=30; m_iAutoAverageNy=30;
	m_wgtMin=0.; m_wgtMax=0.; m_xMin=0.; m_xMax=0.; m_yMin=0.; m_yMax=0.; m_bExpandMargins=true;
}
CVMScatLite::CVMScatLite(const char* name, double xMin, double xMax, double yMin, double yMax, bool bAutoAverage):IVMPlotData2D(name) {
	m_valX = new double[VMScat_MAXEVTS];
	m_valY = new double[VMScat_MAXEVTS];
	m_valWgt = new double[VMScat_MAXEVTS];
	m_nEvents=0; m_bOverfilled=false; m_bAutoAverage=bAutoAverage; m_iAutoAverageNx=30; m_iAutoAverageNy=30;
	m_wgtMin=0.; m_wgtMax=0.; m_xMin=xMin; m_xMax=xMax; m_yMin=yMin; m_yMax=yMax; m_bExpandMargins=false;
}
CVMScatLite::CVMScatLite(const char* name, bool bAutoAverage):IVMPlotData2D(name) {
	m_valX = new double[VMScat_MAXEVTS];
	m_valY = new double[VMScat_MAXEVTS];
	m_valWgt = new double[VMScat_MAXEVTS];
	m_nEvents=0; m_bOverfilled=false; m_bAutoAverage=bAutoAverage; m_iAutoAverageNx=30; m_iAutoAverageNy=30;
	m_wgtMin=0.; m_wgtMax=0.; m_xMin=0.; m_xMax=0.; m_yMin=0.; m_yMax=0.; m_bExpandMargins=true;
}
CVMScatLite::~CVMScatLite() {
	if (m_valX!=NULL) delete[] m_valX;
	if (m_valY!=NULL) delete[] m_valY;
	if (m_valWgt!=NULL) delete[] m_valWgt;
}
int CVMScatLite::GetClassId() { return IVMPlotData2D::PL_SCAT; }
void CVMScatLite::Fill(double x, double y, double wgt) {
	if ((m_valX==NULL) || (m_valY==NULL) || (m_valWgt==NULL)) return;
	if (m_nEvents>=VMScat_MAXEVTS) {
		if(m_bAutoAverage) { Average(m_iAutoAverageNx,m_iAutoAverageNy); if(m_nEvents>=VMScat_MAXEVTS){m_bOverfilled=true;return;} }
		else{ m_bOverfilled=true; return; }
	}
	m_valX[m_nEvents]=x; m_valY[m_nEvents]=y; m_valWgt[m_nEvents]=wgt;
	if(m_bExpandMargins) {
		if(m_nEvents>0) { if(m_xMin>x)m_xMin=x; if(m_xMax<x)m_xMax=x; if(m_yMin>y)m_yMin=y; if(m_yMax<y)m_yMax=y; }
		else { m_xMin=x; m_xMax=x; m_yMin=y; m_yMax=y; }
	}
	if(m_nEvents>0){ if(m_wgtMin>wgt)m_wgtMin=wgt; if(m_wgtMax<wgt)m_wgtMax=wgt; }else{ m_wgtMin=wgt; m_wgtMax=wgt; }
	m_nEvents++;
}
bool CVMScatLite::GetValue(unsigned long i, double *x, double *y, double *wgt) {
	if ((i<m_nEvents) && (m_valX!=NULL) && (m_valY!=NULL) && (m_valWgt!=NULL)) {
		if(x!=NULL)*x=m_valX[i]; if(y!=NULL)*y=m_valY[i]; if(wgt!=NULL)*wgt=m_valWgt[i];
		return true;
	}
	return false;
}
void CVMScatLite::SetAutoAverage(bool bAverage, int nx, int ny) { m_bAutoAverage=bAverage; if(nx<=0)nx=30; if(ny<=0)ny=30; m_iAutoAverageNx=nx; m_iAutoAverageNy=ny; }
unsigned long CVMScatLite::GetNumEvents() { return m_nEvents; }
double CVMScatLite::GetMaxXValue() { return m_xMax; }
double CVMScatLite::GetMinXValue() { return m_xMin; }
double CVMScatLite::GetMaxYValue() { return m_yMax; }
double CVMScatLite::GetMinYValue() { return m_yMin; }
double CVMScatLite::GetMaxWgt() { return m_wgtMax; }
double CVMScatLite::GetMinWgt() { return m_wgtMin; }
bool CVMScatLite::IsEmpty() { return (m_nEvents<=0); }
bool CVMScatLite::IsOverfilled() { return m_bOverfilled; }
void CVMScatLite::Average(unsigned nx, unsigned ny) { // leave nx*ny points to reduce output size
	if ((nx<1) || (ny<1) || (((unsigned long)nx)*ny>VMScat_MAXEVTS)) return;
	if ((m_xMax<m_xMin) || (m_yMax<m_yMin)) return;
	double tmpX[VMScat_MAXEVTS]; double tmpY[VMScat_MAXEVTS]; double tmpW[VMScat_MAXEVTS]; long tmpN[VMScat_MAXEVTS]; unsigned long i,j;
	for(i=0;i<nx;i++){ for(j=0;j<ny;j++){
		tmpX[i*nx+j]=m_xMin+(m_xMax-m_xMin)*(0.5+i)/nx; tmpY[i*nx+j]=m_yMin+(m_yMax-m_yMin)*(0.5+j)/ny; tmpW[i*nx+j]=0.; tmpN[i*nx+j]=0;
	} }
	for(i=0;i<m_nEvents;i++) {
		double x=m_valX[i]; double y=m_valY[i];
		if((x<m_xMin)||(x>m_xMax)||(y<m_yMin)||(y>m_yMax)) continue; // filter outmargins
		unsigned ix=((m_xMax>m_xMin)?(int)((x-m_xMin)*nx/(m_xMax-m_xMin)):0); if(ix>=nx)ix=nx;
		unsigned iy=((m_yMax>m_yMin)?(int)((y-m_yMin)*ny/(m_yMax-m_yMin)):0); if(iy>=ny)iy=ny;
		tmpW[ix*nx+iy]+=m_valWgt[i]; tmpN[ix*nx+iy]++;
	}
	unsigned long nEvtMax=((unsigned long)nx)*ny;
	m_wgtMax=0; m_wgtMin=0; m_nEvents=0;
	for(i=0;i<nEvtMax;i++) {
		if (tmpN[i]==0) continue;
		m_valX[m_nEvents]=tmpX[i]; m_valY[m_nEvents]=tmpY[i]; double w=tmpW[i]; m_valWgt[m_nEvents]=w; m_nEvents++;
		if(i>0){ if(m_wgtMin>w)m_wgtMin=w; if(m_wgtMax<w)m_wgtMax=w; }else{ m_wgtMin=w; m_wgtMax=w; }
	}
}

/////////////////////////////////////////////////////////////////////// CVMFreeDrawing
#define CVMFreeDr_MAXEVTS	100000 // size: MAXEVTS*20
CVMFreeDrawing::CVMFreeDrawing(const char* name):IVMPlotData2D(name) {
	m_valX = new double[CVMFreeDr_MAXEVTS];
	m_valY = new double[CVMFreeDr_MAXEVTS];
	m_valType = new int[CVMFreeDr_MAXEVTS];
	m_xMin=0.; m_xMax=0.; m_yMin=0.; m_yMax=0.; m_nEvents=0; m_bOverfilled=false;
}
CVMFreeDrawing::~CVMFreeDrawing() {
	if (m_valX!=NULL) delete[] m_valX;
	if (m_valY!=NULL) delete[] m_valY;
	if (m_valType!=NULL) delete[] m_valType;
}
int CVMFreeDrawing::GetClassId() { return IVMPlotData2D::PL_FREEDRAW; }
void CVMFreeDrawing::MoveTo(double x, double y) { AddAction(1,x,y); }
void CVMFreeDrawing::LineTo(double x, double y) { AddAction(2,x,y); }
void CVMFreeDrawing::AddAction(int type, double x, double y) { //1(moveto),2(lineto)
	if ((m_valX==NULL) || (m_valY==NULL) || (m_valType==NULL)) return;
	if (m_nEvents>=CVMFreeDr_MAXEVTS) { m_bOverfilled=true; return; }
	// Ceck numbers
	// Same check to (CAlHepLib::IsNaN(val) || CAlHepLib::IsInf(val)) //TODO: Uniform this check!
	if ((x!=x) || ((((unsigned long*)&x)[1] & 0x7ff80000) == 0x7ff00000)) return;
	if ((y!=y) || ((((unsigned long*)&y)[1] & 0x7ff80000) == 0x7ff00000)) return;
	//
	m_valX[m_nEvents]=x; m_valY[m_nEvents]=y;
	m_valType[m_nEvents]=type; m_nEvents++;
	if(m_nEvents>0){if(m_xMin>x)m_xMin=x;if(m_xMax<x)m_xMax=x;if(m_yMin>y)m_yMin=y;if(m_yMax<y)m_yMax=y;}else{m_xMin=x;m_xMax=x;m_yMin=y;m_yMax=y;}
}
int CVMFreeDrawing::GetAction(unsigned long i, double *x, double *y) { //1(moveto),2(lineto)
	if ((i<m_nEvents) && (m_valX!=NULL) && (m_valY!=NULL) && (m_valType!=NULL)) {
		if(x!=NULL)*x=m_valX[i]; if(y!=NULL)*y=m_valY[i]; return m_valType[i];
	}
	return -1;
}
unsigned long CVMFreeDrawing::GetNumEvents() { return m_nEvents; }
double CVMFreeDrawing::GetMaxXValue() { return m_xMax; }
double CVMFreeDrawing::GetMinXValue() { return m_xMin; }
double CVMFreeDrawing::GetMaxYValue() { return m_yMax; }
double CVMFreeDrawing::GetMinYValue() { return m_yMin; }
bool CVMFreeDrawing::IsEmpty() { return (m_nEvents==0); }
bool CVMFreeDrawing::IsOverfilled() { return m_bOverfilled; }

/////////////////////////////////////////////////////////////////////// CVMHistLiteView
bool CVMHistLiteViewDefaults::m_bDrawAxisBox		= true;		//def: true
bool CVMHistLiteViewDefaults::m_bDrawLegend			= true;		//def: true
bool CVMHistLiteViewDefaults::m_bPlotBins			= false;	//def: false
bool CVMHistLiteViewDefaults::m_bDrawErrorMarks		= false;	//def: false
bool CVMHistLiteViewDefaults::m_bPrintNEvts			= false;	//def: false
bool CVMHistLiteViewDefaults::m_bPrintIntegral		= false;	//def: false
bool CVMHistLiteViewDefaults::m_bLegendOverlapped	= false;	//def: false (draw legend in overlapped box)
bool CVMHistLiteViewDefaults::m_bSaveViewWithEPS	= false;	//def: false (save draw 'CVMHistLiteView'-object together with SaveAsEPS() call)
bool CVMHistLiteViewDefaults::m_bThickGridLines		= true;
bool CVMHistLiteViewDefaults::m_bLablesAtCenter		= false;
//
CVMHistLiteView::CVMHistLiteView(IVMPlotData2D* plot) {
	//Warinig: 'Clear' and 'AddPlotToView' may NOT be virtual due to these calls
	//
	//--- Mandatory initializations before Clear() ---
	for(long i=0;i<VMHistLiteViewHMAX;i++){ m_plots[i]=NULL; }
	m_sXAxisLabel=NULL; m_sYAxisLabel=NULL; m_sPlotLabel=NULL;
	//
	Clear();
	if(plot!=NULL)AddPlotToView(plot); //add first plot
}
CVMHistLiteView::~CVMHistLiteView() {
	for(long i=0;i<VMHistLiteViewHMAX;i++){
		if((m_plots[i]!=NULL) && (!m_bPlotIsExternal[i])){ delete m_plots[i]; m_plots[i]=NULL; }
	}
}
void CVMHistLiteView::Clear() {
	int i;
	m_nPlots=0;
	m_xAxis=0; m_yAxis=0; m_xPosMax=1; m_yPosMax=1; m_bLogAxisX=false; m_bLogAxisY=false;
	m_xMinVal=0; m_xMaxVal=1; m_yMinVal=0; m_yMaxVal=1;
	m_bDrawVertGrid=false; m_bDrawHorGrid=false;
	//copy defaults
	m_bDrawAxisBox=CVMHistLiteViewDefaults::m_bDrawAxisBox;
	m_bDrawLegend=CVMHistLiteViewDefaults::m_bDrawLegend;
	m_bLegendOverlapped=CVMHistLiteViewDefaults::m_bLegendOverlapped;
	//[Old] m_bPlotBins=CVMHistLiteViewDefaults::m_bPlotBins;
	m_bDrawErrorMarks=CVMHistLiteViewDefaults::m_bDrawErrorMarks;
	m_bPrintNEvts=CVMHistLiteViewDefaults::m_bPrintNEvts;
	m_bPrintIntegral=CVMHistLiteViewDefaults::m_bPrintIntegral;
	m_bThickGridLines=CVMHistLiteViewDefaults::m_bThickGridLines;
	m_bLablesAtCenter=CVMHistLiteViewDefaults::m_bLablesAtCenter;
	//
	m_bHideEmptyHist=false;
	m_bShiftMergedPoints=true;
	m_xLevelLineAt=0.; m_yLevelLineAt=0.;
	m_iLegendPos=1; //1: RightTop, 2: RightBottom, 3: LeftTop, 4:LeftBottom
	m_bAxisXFactorInsideBox=false;
	m_bAxisYFactorInsideBox=false;
	//
	int styPlotDef=0; if(CVMHistLiteViewDefaults::m_bPlotBins)styPlotDef|=CVMHistLiteView::LN_DRAW_BINS;
	for(i=0;i<VMHistLiteViewHMAX;i++){
		if((m_plots[i]!=NULL) && (!m_bPlotIsExternal[i])){ delete m_plots[i]; m_plots[i]=NULL; }
		m_plots[i]=NULL; m_bPlotIsExternal[i]=true; m_clrPlot[i].Set(0.f,0.f,0.f); m_styPlot[i]=styPlotDef; m_LineWidth[i]=CVMOutput::LW_NORMAL; /*default style&colors*/
	}
	m_clrBcgr=CVMOutput::clrWhite;
	if (VMHistLiteViewHMAX>=5) { // modify default colors for first plots
		m_clrPlot[0].Set(0,0,1);
		m_clrPlot[1].Set(0,1,0); //m_clrPlot[1].Set(1,0,0);
		m_clrPlot[2].Set(1,0,0); //m_clrPlot[2].Set(0,1,0);
		m_clrPlot[3].Set(1,1,0);
		m_clrPlot[4].Set(1,0,1);
	}
	//
	m_RequestedYMargins[0]=0.; m_RequestedYMargins[1]=-1.; //Check for usage: if(m_RequestedYMargins[0]<m_RequestedYMargins[1])...
	m_RequestedXMargins[0]=0.; m_RequestedXMargins[1]=-1.; //Check for usage: if(m_RequestedXMargins[0]<m_RequestedXMargins[1])...
	if(m_sXAxisLabel!=NULL){ CVMUtil::strFree(m_sXAxisLabel); m_sXAxisLabel=NULL; }
	m_fYLabelYShift=0.; m_fXLabelYShift=0.;
	if(m_sYAxisLabel!=NULL){ CVMUtil::strFree(m_sYAxisLabel); m_sYAxisLabel=NULL; }
	if(m_sPlotLabel!=NULL){ CVMUtil::strFree(m_sPlotLabel); m_sPlotLabel=NULL; }
}
void CVMHistLiteView::AddPlotToView(IVMPlotData2D* plot, bool bKeepHistoInView) {
	if (m_nPlots<VMHistLiteViewHMAX) {
		if(plot!=NULL){
			m_plots[m_nPlots]=plot; m_bPlotIsExternal[m_nPlots]=(!bKeepHistoInView); m_nPlots++;
			if(m_nPlots==1){ // On first plot added...
				// Autodetect X-log-scale
				CVMHistLite* hist=NULL; //=='plot' here
				if ((m_nPlots>0) && (m_plots[0]!=NULL)) {
					if (m_plots[0]->GetClassId()==IVMPlotData2D::PL_HIST) { hist=((CVMHistLite*)m_plots[0]); }
				}
				if(hist!=NULL) {
					int N=hist->GetNBins();
					if (N>0) {
						double fRatio=0; double fRatio2=0; double wPrev=hist->GetBinXWidth(0); bool wOK=(wPrev>0);
						for(int i=1;((i<N) && wOK);i++) {
							double w=hist->GetBinXWidth(i); if(w<=0){wOK=false;break;} double r=w/wPrev; fRatio+=r; fRatio2+=r*r; wPrev=w;
						}
						if (wOK) {
							fRatio/=(N-1); fRatio2/=(N-1); double fSigma=fRatio2-fRatio*fRatio;
							if (fSigma<fRatio*fRatio/10000) { // check if bin widths are not random
								if (fRatio>1.0001) m_bLogAxisX=true;
								//Negative log-values not allowed: if ((fRatio<0.9999) || (fRatio>1.0001)) m_bLogAxisX=true;
							}
						}
					}
				}
				/////////////////
			}
		}
	} else {
		std::cout<<"ERROR in CVMHistLiteView::AddPlotToView(). Too many plots in view.";
	}
}
void CVMHistLiteView::RemovePlotFromView(int i) {
	if((i>=0)&&(i<m_nPlots)) {
		IVMPlotData2D* plot=m_plots[i]; if(!m_bPlotIsExternal[i])delete plot;
		//Shift plots list back
		for(int j=i;j<(m_nPlots-1);j++){ m_plots[j]=m_plots[j+1]; m_bPlotIsExternal[j]=m_bPlotIsExternal[j+1]; }
		m_nPlots--;
	} else {
		std::cout<<"ERROR in CVMHistLiteView::RemovePlotFromView(). Invalid plot ID.";
	}
}
const int CVMHistLiteView::LN_DENSE_DASH_FLAG	= 0x010;	// to be combined with CVMOutput::LINE_XXX
const int CVMHistLiteView::LN_MASK_BASIC_STYLE	= 0x00f;	// includes all CVMOutput::LINE_XXX but not local flags
const int CVMHistLiteView::LN_MASK_DASH_INFO	= 0x0ff;	// (for local use only) includes all dash info
const int CVMHistLiteView::LN_DRAW_BINS			= 0x100;
void CVMHistLiteView::GetHistStyle(int i, CVMRGB* clr, int* iLineType, float* fLineWidth) {
	if ((i>=0) && (i<VMHistLiteViewHMAX)) { // histogram may be not added yet
		if(clr!=NULL)*clr=m_clrPlot[i]; if(iLineType!=NULL)*iLineType=m_styPlot[i]; if(fLineWidth!=NULL)*fLineWidth=m_LineWidth[i];
	}
}
void CVMHistLiteView::SetHistStyle(int i, CVMRGB clr, int iLineType, float fLineWidth) {
	if ((i>=0) && (i<VMHistLiteViewHMAX)) { // histogram may be not added yet
		if(clr.IsValid())m_clrPlot[i]=clr;
		if(iLineType<0)iLineType=CVMOutput::LINE_SOLID;
		if(((iLineType & CVMHistLiteView::LN_DENSE_DASH_FLAG) == CVMHistLiteView::LN_DENSE_DASH_FLAG) &&
			((iLineType & CVMHistLiteView::LN_MASK_BASIC_STYLE) == CVMOutput::LINE_SOLID)) {
				iLineType |= CVMOutput::LINE_DASH; // set dash if 'LN_DENSE_DASH_FLAG' flag is set
		}
		m_styPlot[i]=iLineType;
		if(fLineWidth>=0.f)m_LineWidth[i]=fLineWidth;
	}
}
void CVMHistLiteView::SetBlackAndWhiteStyle() {
	m_clrBcgr=CVMOutput::clrWhite;
	for(long i=0;i<VMHistLiteViewHMAX;i++){
		m_clrPlot[i]=CVMOutput::clrBlack;
		int k=i%6;
		switch(k){
		case 0: m_styPlot[i]=CVMOutput::LINE_SOLID; m_LineWidth[i]=CVMOutput::LW_NORMAL; break;
		case 1: m_styPlot[i]=CVMOutput::LINE_DASH; m_LineWidth[i]=CVMOutput::LW_NORMAL; break;
		case 2: m_styPlot[i]=CVMOutput::LINE_DASHDOT; m_LineWidth[i]=CVMOutput::LW_NORMAL; break;
		case 3: m_styPlot[i]=CVMOutput::LINE_SOLID; m_LineWidth[i]=CVMOutput::LW_THIN; break;
		case 4: m_styPlot[i]=CVMOutput::LINE_DASH; m_LineWidth[i]=CVMOutput::LW_THIN; break;
		default: m_styPlot[i]=CVMOutput::LINE_DASHDOT; m_LineWidth[i]=CVMOutput::LW_THIN;
		}
	}
}
bool CVMHistLiteView::GetPlotBinsFlag(int idHist) {
	bool res=CVMHistLiteViewDefaults::m_bPlotBins;
	if((idHist>=0)&&(idHist<VMHistLiteViewHMAX)){ res=((m_styPlot[idHist]&CVMHistLiteView::LN_DRAW_BINS)==CVMHistLiteView::LN_DRAW_BINS); }
	return res;
}
void CVMHistLiteView::SetPlotBinsFlag(int idHist, bool bPlotBins) {
	if((idHist>=0)&&(idHist<VMHistLiteViewHMAX)){ if(bPlotBins){m_styPlot[idHist]|=CVMHistLiteView::LN_DRAW_BINS;}else{m_styPlot[idHist]^=CVMHistLiteView::LN_DRAW_BINS;} }
}
void CVMHistLiteView::SetHistLineWidth(int i, float line_width) {
	if ((i>=0) && (i<VMHistLiteViewHMAX)) { // histogram may be not added yet
		if(line_width>=0.f)m_LineWidth[i]=line_width;
	}
}
void CVMHistLiteView::SetPlotBins(bool bPlotBins) {
	for(int i=0;i<VMHistLiteViewHMAX;i++){ //include all future histos here!
		if(bPlotBins){
			m_styPlot[i]|=CVMHistLiteView::LN_DRAW_BINS;
		}else{
			if((m_styPlot[i] & CVMHistLiteView::LN_DRAW_BINS)==CVMHistLiteView::LN_DRAW_BINS){ m_styPlot[i]^=CVMHistLiteView::LN_DRAW_BINS; }
		}
	}
}
bool CVMHistLiteView::ContainBinPlots() { // if any plot is marked as 'bins-drawed'
	bool res=false; for(int i=0;i<m_nPlots;i++){ if((m_styPlot[i] & CVMHistLiteView::LN_DRAW_BINS)==CVMHistLiteView::LN_DRAW_BINS){ res=true; break; } } return res;
}
void CVMHistLiteView::SetLogX(bool bLog) { m_bLogAxisX=bLog; }
void CVMHistLiteView::SetLogY(bool bLog) { m_bLogAxisY=bLog; }
void CVMHistLiteView::SetXAxisLabel(const char* xLabel) {
	if(m_sXAxisLabel!=xLabel){ if(m_sXAxisLabel!=NULL){CVMUtil::strFree(m_sXAxisLabel);} m_sXAxisLabel=CVMUtil::strCpy(xLabel); }
}
void CVMHistLiteView::SetYAxisLabel(const char* yLabel) {
	if(m_sYAxisLabel!=yLabel){ if(m_sYAxisLabel!=NULL){CVMUtil::strFree(m_sYAxisLabel);} m_sYAxisLabel=CVMUtil::strCpy(yLabel); }
}
void CVMHistLiteView::SetPlotLabel(const char* sPlotLabel) {
	if(m_sPlotLabel!=sPlotLabel){ if(m_sPlotLabel!=NULL){CVMUtil::strFree(m_sPlotLabel);} m_sPlotLabel=CVMUtil::strCpy(sPlotLabel); }
}

bool CVMHistLiteView::SaveAsEPS(const char* fn) {
	if ((fn==NULL) || (m_nPlots<=0)) return false;
	CVMOutputEPS out(fn);
	//out.SetBoundingBox(91.5625, 321.9375, 3.1875, 145.5); // Mathematica default
	//out.SetBoundingBox(91.5625, 517.125, 3.1875, 266.125);
	//out.SetBoundingBox(0,500,0,300);
	out.Start();
	bool bWriteOK=false;
	if (out.IsOpen()) {
		bWriteOK=write_plot(&out);
		out.Finalize();
	}
	//Save View-object
	if(CVMHistLiteViewDefaults::m_bSaveViewWithEPS){ char buf[500]; sprintf(buf,"%s.%s",fn,VMHistLiteViewFileExt); SaveAsObject(buf); }
	return bWriteOK;
}
bool CVMHistLiteView::SaveAsNB(const char* fn) {
	if ((fn==NULL) || (m_nPlots<=0)) return false;
	CVMOutputMath out(fn); out.Start();
	bool bWriteOK=false;
	if (out.IsOpen()) {
		bWriteOK=write_plot(&out);
		if (bWriteOK) { // add legend
			char tmp_buf[300];
			for(int i=0; i<m_nPlots; i++) {
				if(m_plots[i]==NULL)continue;
				if(m_plots[i]->GetClassId()==IVMPlotData2D::PL_HIST) {
					sprintf(tmp_buf,"Histogram #%d: Events=%d, Max=%g, OutOfMargins=%d",i,((CVMHistLite*)m_plots[i])->GetNumEvents(),((CVMHistLite*)m_plots[i])->GetMaxYValue(),((CVMHistLite*)m_plots[i])->GetNumOutmargins());
					out.NewTextCell(); out.TextOut(0,0,tmp_buf);
				}
				out.NewTextCell(); out.TextOut(0,0,tmp_buf);
			}
		}
		out.Finalize();
	}
	//Save View-object
	if(CVMHistLiteViewDefaults::m_bSaveViewWithEPS){ char buf[500]; sprintf(buf,"%s.%s",fn,VMHistLiteViewFileExt); SaveAsObject(buf); }
	return bWriteOK;
}

bool CVMHistLiteView::SaveAsObject(const char* fn) {
	CVMFileOut f(fn,true); if(!f.IsReady())return false; long i; const char* sTmp;
	//
	f<<"["<<((long)CVMHistLiteVER)<<"]";
	f<<(long)m_nPlots<<"\n";
	for(i=0;i<m_nPlots;i++){ //or up to 'VMHistLiteViewHMAX'?
		f<<(long)(m_clrPlot[i].m_Red*100)<<" "<<(long)(m_clrPlot[i].m_Green*100)<<" "<<(long)(m_clrPlot[i].m_Blue*100)<<" "<<(long)(m_styPlot[i])<<" "<<(long)(m_LineWidth[i]*100)<<"\n";
	}
	f<<(long)(m_clrBcgr.m_Red*100)<<" "<<(long)(m_clrBcgr.m_Green*100)<<" "<<(long)(m_clrBcgr.m_Blue*100)<<"\n";

	f<<m_xAxis<<" "<<m_yAxis<<" "<<m_xPosMax<<" "<<m_yPosMax<<" "<<(long)(m_bLogAxisX?1:0)<<" "<<(long)(m_bLogAxisY?1:0)<<" "<<m_xMinVal<<" "<<m_xMaxVal<<" "<<m_yMinVal<<" "<<m_yMaxVal<<"\n";
	long iReplaceOfPlotBinsFlag=-1; //previous 'm_bPlotBins' flag ('0' or '1' is treated as ols value)
	f<<(long)(m_bDrawVertGrid?1:0)<<" "<<(long)(m_bDrawHorGrid?1:0)<<" "<<(long)(m_bDrawAxisBox?1:0)<<" "<<(long)(m_bDrawLegend?1:0)<<" "<<(long)(m_bLegendOverlapped?1:0)<<" "<<iReplaceOfPlotBinsFlag<<" "<<(long)(m_bDrawErrorMarks?1:0)<<" "<<(long)(m_bPrintNEvts?1:0)<<" "<<(long)(m_bPrintIntegral?1:0)<<"\n";
	f<<(long)(m_bHideEmptyHist?1:0)<<" "<<(long)(m_bShiftMergedPoints?1:0)<<" "<<m_xLevelLineAt<<" "<<m_yLevelLineAt<<" "<<(long)m_iLegendPos<<" "<<(long)(m_bAxisXFactorInsideBox?1:0)<<" "<<(long)(m_bAxisYFactorInsideBox?1:0)<<" ";
	f<<m_RequestedYMargins[0]<<" "<<m_RequestedYMargins[1]<<"\n";
	for(i=0;i<m_nPlots;i++){
		f<<i<<" "; if(m_plots[i]==NULL){ f<<"Z "; continue; }
		if(m_plots[i]->GetClassId()==IVMPlotData2D::PL_HIST){ f<<"H"; ((CVMHistLite*)m_plots[i])->SaveAsObject(&f); f<<" "; }else{ f<<"X "; }
	}
	f<<"A\n"; //'Added later' fields
	f<<m_RequestedXMargins[0]<<" "<<m_RequestedXMargins[1]<<"\n";
	sTmp=m_sXAxisLabel; if(sTmp==NULL)sTmp=""; f<<"["<<CVMUtil::strLen(sTmp)<<"<"<<sTmp<<">]";
	sTmp=m_sYAxisLabel; if(sTmp==NULL)sTmp=""; f<<"["<<CVMUtil::strLen(sTmp)<<"<"<<sTmp<<">]";
	f<<"E\n"; f.Close(); return true;

	//Removed fields:
	//	m_bPlotBins

	//Not saved fields:
	//	m_bThickGridLines, m_bLablesAtCenter
	//	m_sPlotLabel
	//	m_fYLabelYShift, m_fXLabelYShift
}
bool CVMHistLiteView::LoadAsObject(const char* fn) {
	CVMFileIn f(fn); if(!f.IsReady()){ std::cout<<"Can't read histogram file. File "<<fn<<" not found"<<std::endl; return false; }
	int errCode=LoadAsObjectImpl(f);
	if(errCode>0){
		std::cout<<"Error loading histogram file. Error code: "<<errCode<<"."<<std::endl;
	}
	return (errCode==0);
}
int CVMHistLiteView::LoadAsObjectImpl(CVMFileIn& f) {
	Clear();
	long i,iVer,iTmp1,iTmp2,iTmp3; char c;
	if(!f.ReadNotEmptySymbol(c) || c!='['){ return 1; }
	f>>iVer; if(iVer!=CVMHistLiteVER){ std::cout<<"Error in CVMHistLiteView::LoadAsObject: invalid object version"<<std::endl; return 2; }
	if(!f.ReadNotEmptySymbol(c) || c!=']')return 3;
	long nPlots=0;
	f>>nPlots; if((nPlots<0)||(nPlots>VMHistLiteViewHMAX))return 4;
	for(i=0;i<nPlots;i++){ //or up to 'VMHistLiteViewHMAX'?
		f>>iTmp1>>iTmp2>>iTmp3>>m_styPlot[i]; m_clrPlot[i].Set(((float)iTmp1)/100,((float)iTmp2)/100,((float)iTmp3)/100);
		f>>iTmp1; m_LineWidth[i]=((float)iTmp1)/100;
	}
	f>>iTmp1>>iTmp2>>iTmp3; m_clrBcgr.Set(((float)iTmp1)/100,((float)iTmp2)/100,((float)iTmp3)/100);
	f>>m_xAxis>>m_yAxis>>m_xPosMax>>m_yPosMax; f>>iTmp1;m_bLogAxisX=(iTmp1>0); f>>iTmp1;m_bLogAxisY=(iTmp1>0); f>>m_xMinVal>>m_xMaxVal>>m_yMinVal>>m_yMaxVal;
	f>>iTmp1;m_bDrawVertGrid=(iTmp1>0); f>>iTmp1;m_bDrawHorGrid=(iTmp1>0); f>>iTmp1;m_bDrawAxisBox=(iTmp1>0); f>>iTmp1;m_bDrawLegend=(iTmp1>0); f>>iTmp1;m_bLegendOverlapped=(iTmp1>0);
	f>>iTmp1; if((iTmp1==0)||(iTmp1==1))SetPlotBins(iTmp1>0); //<-- [Old:]m_bPlotBins=(iTmp1>0);
	f>>iTmp1;m_bDrawErrorMarks=(iTmp1>0); f>>iTmp1;m_bPrintNEvts=(iTmp1>0); f>>iTmp1;m_bPrintIntegral=(iTmp1>0);
	f>>iTmp1;m_bHideEmptyHist=(iTmp1>0); f>>iTmp1;m_bShiftMergedPoints=(iTmp1>0); f>>m_xLevelLineAt>>m_yLevelLineAt>>m_iLegendPos; f>>iTmp1;m_bAxisXFactorInsideBox=(iTmp1>0); f>>iTmp1;m_bAxisYFactorInsideBox=(iTmp1>0);
	f>>m_RequestedYMargins[0]>>m_RequestedYMargins[1];
	for(i=0;i<nPlots;i++){
		f>>iTmp1; if(iTmp1!=i)return false; if(!f.ReadNotEmptySymbol(c))return 5;
		if(c=='H'){
			CVMHistLite* hist=new CVMHistLite(); if(!hist->LoadAsObject(&f)){ delete hist; return 6; }
			AddPlotToView(hist,true);
		}
	}
	if(!f.ReadNotEmptySymbol(c))return 90;
	if(c=='A'){ //Load 'added later' fields
		f>>m_RequestedXMargins[0]>>m_RequestedXMargins[1];
		if(m_sXAxisLabel!=NULL){ CVMUtil::strFree(m_sXAxisLabel); m_sXAxisLabel=NULL; }
		if(m_sYAxisLabel!=NULL){ CVMUtil::strFree(m_sYAxisLabel); m_sYAxisLabel=NULL; }
		int iSz=-1;
		if((!f.ReadNotEmptySymbol(c)) || (c!='[')){ return 91; }
		f>>iSz; if(iSz<0){ return 91; }
		char* buf1=CVMUtil::strAlloc(iSz+10); if(buf1==NULL){ return 92; }
		if((!f.ReadNotEmptySymbol(c)) || (c!='<')){ CVMUtil::strFree(buf1); return 93; }
		if(!f.ReadString(buf1,iSz+8,">]")){ CVMUtil::strFree(buf1); return 94; }
		m_sXAxisLabel=((iSz>0)?(CVMUtil::strCpy(buf1)):NULL); CVMUtil::strFree(buf1); buf1=NULL;
		if((!f.ReadNotEmptySymbol(c)) || (c!='[')){ return 95; }
		f>>iSz; if(iSz<0){ return 96; }
		buf1=CVMUtil::strAlloc(iSz+10); if(buf1==NULL){ return 97; }
		if((!f.ReadNotEmptySymbol(c)) || (c!='<')){ CVMUtil::strFree(buf1); return 98; }
		if(!f.ReadString(buf1,iSz+8,">]")){ CVMUtil::strFree(buf1); return 99; }
		m_sYAxisLabel=((iSz>0)?(CVMUtil::strCpy(buf1)):NULL); CVMUtil::strFree(buf1); buf1=NULL;
		//
		if(!f.ReadNotEmptySymbol(c))return 198;
	}
	if(c!='E')return 199;
	return 0;
}

int CVMHistLiteView::UtilGetPower(double val, double* mantissa) { // format: n.mmm*10^pow, n[1,9]
	if(val<0)return UtilGetPower(-val);
	if(val==0.) { if(mantissa!=NULL)*mantissa=0; return 0; }
	int res=0;
	if (val<1) { for(;val<1;) { val*=10; res--; } }
	else { for(;val>=10;) { val/=10; res++; } }
	if(mantissa!=NULL)*mantissa=val;
	return res;
}
double CVMHistLiteView::GetRoundMark(double val_max, bool bAllow52, bool bGetUpper) {
	if (val_max<0) return -GetRoundMark(-val_max,bAllow52,bGetUpper);
	if (val_max==0) return 0;
	double res=1.;
	if (val_max<1) { for(;val_max<1;) { val_max*=10; res/=10; } }
	else { for(;val_max>=10;) { val_max/=10; res*=10; } }
	if (bGetUpper) {
		if (val_max>1) { // note: val_max may be ==1
			if ((bAllow52) && (val_max<=2)) { res*=2; }
			else if ((bAllow52) && (val_max<=5)) { res*=5; }
			else { res*=10; }
		}
	} else {
		if (bAllow52) {
			if(val_max>=5){ res*=5; }else if(val_max>=2){ res*=2; }
		}
	}
	return res;
}
long CVMHistLiteView::GetUpperRoundLong(long val, bool bOnly125) { //6..9->10, 11..14->15, 16..19->20, ...
	if (val<0) return -GetUpperRoundLong(-val,bOnly125);
	long res=val; long f=1;
	for(;;){
		if(val<=5){ res=val; break; }
		if(val<=10){ res=10; break; }
		if((!bOnly125)&&(val<=15)){ res=15; break; }
		if(val<=20){ res=20; break; }
		if((!bOnly125)&&(val<=25)){ res=25; break; }
		if((!bOnly125)&&(val<=40)){ res=40; break; }
		if(val<=50){ res=50; break; }
		f*=10; val=(val-val%10)/10+1;
	}
	res*=f; return res;
}
double CVMHistLiteView::GetTruncToStep(double val, double round_step) {
	if (val<0) return -GetTruncToStep(-val,round_step);
	if (round_step<=0.) return val;
	if (val<=round_step) return 0.;
	// Example: {val=123456,step=111}
	double valM=0; int valP=UtilGetPower(val, &valM); // {5,1.23456}
	double stepM=0; int stepP=UtilGetPower(round_step, &stepM); // {2,1.11}
	int i;

	double res=0.;
	if (valP>stepP) { // 5>2
		res=valM; for(i=stepP+1;i<valP;i++){res=res*10;} // {1.23456->123.456}
		int itmp=(int)res; if(itmp>res)itmp--; res=itmp; // {123.456->123.000}
		int pw=stepP+1; //=valP-(valP-(stepP+1));
		if(pw>=0){ for(i=0;i<pw;i++){res=res*10;} }else{ for(i=0;i<(-pw);i++){res=res/10;} } // {123.000->123000}
	}

	double sub=val-res; if(sub<0.)sub=0.; // // 0<sub<round_step*10 {456}
	for(i=0;i<10;i++){ if(round_step*(i+1)>sub)break; } // {i->4}
	res += round_step*i; // {123000+444}
	return res;
}
double CVMHistLiteView::GetXPos(double xVal) {
	double res=-1;
	if (m_bLogAxisX) {
		// Note: m_xMaxVal & m_xMinVal checks should be performed before
		if (xVal>0) {
			res=m_xAxis+(m_xPosMax-m_xAxis)*(log(xVal/m_xMinVal))/(log(m_xMaxVal/m_xMinVal));
		}
	} else {
		res=m_xAxis+(m_xPosMax-m_xAxis)*(xVal-m_xMinVal)/(m_xMaxVal-m_xMinVal);
	}
	// Old: if (res>m_xPosMax) { res=m_xPosMax; } // "this occurs in 10^-16 digits if xVal==m_xMaxVal"
	// No! This MUST occur if xVal>m_xMinVal
	// TODO: add 10^-16 overflow check if required
	return res;
}
double CVMHistLiteView::GetYPos(double yVal) {
	double res=-1;
	if (m_bLogAxisY) {
		// Note: m_yMaxVal & m_yMinVal checks should be performed before
		if (yVal>0) {
			res=m_yAxis+(m_yPosMax-m_yAxis)*(log(yVal/m_yMinVal))/(log(m_yMaxVal/m_yMinVal));
		}
	} else {
		res=m_yAxis+(m_yPosMax-m_yAxis)*(yVal-m_yMinVal)/(m_yMaxVal-m_yMinVal);
	}
	// Old: if (res>m_yPosMax) { res=m_yPosMax; } // "this occurs in 10^-16 digits if yVal==m_yMaxVal"
	// No! This MUST occur if yVal>m_yMinVal
	// TODO: add 10^-16 overflow check if required
	return res;
}

bool CVMHistLiteView::write_plot(CVMOutput* out) {
	if ((out==NULL) || (m_nPlots<=0)) return false;
	IVMPlotData2D* ptFirst=m_plots[0]; if(ptFirst==NULL)return false;
	double xMin=0; double xMax=1; double yMin=0; double yMax=1; out->GetMargins(&xMin,&xMax,&yMin,&yMax);
	char buf[500];
	int i;

	//debug
	//m_bDrawHorGrid=true;
	//m_bDrawVertGrid=true;
	if(m_clrBcgr!=CVMOutput::clrWhite)out->FillBackground(m_clrBcgr);
	bool bFitAllPlotsY=true;

	//Screen margins: m_xAxis, m_xPosMax, m_yAxis, m_yPosMax
	m_xAxis=xMin+(xMax-xMin)*0.1; m_xPosMax=xMax-(xMax-xMin)*0.02;
	m_yAxis=yMin+(yMax-yMin)*0.1; m_yPosMax=yMax-(yMax-yMin)*0.01;
	if(m_sXAxisLabel!=NULL){
		if(m_bLablesAtCenter){
			m_yAxis+=out->GetLetterHeight()*(1.1+m_fXLabelYShift);
		}else{
			int sz=CVMUtil::strLen(m_sXAxisLabel); if(sz>4)sz=3; //TODO: use correct size from 'out->TextOutAdv()'
			m_xPosMax-=sz*0.8*out->GetLetterWidth();
		}
	}
	if(m_sYAxisLabel!=NULL){
		if(m_bLablesAtCenter){
			m_xAxis+=out->GetLetterHeight()*(1.1+m_fYLabelYShift);
		}else{
			//TODO: use correct y-size from 'out->TextOutAdv()'
			m_yPosMax-=out->GetLetterHeight()*(1.2+m_fYLabelYShift);
		}
	}

	//Calculate plot margins (if not predefined)
	bool yValDefined=false; if(m_RequestedYMargins[0]<m_RequestedYMargins[1]){ m_yMinVal=m_RequestedYMargins[0]; m_yMaxVal=m_RequestedYMargins[1]; yValDefined=true; }
	bool xValDefined=false; if(m_RequestedXMargins[0]<m_RequestedXMargins[1]){ m_xMinVal=m_RequestedXMargins[0]; m_xMaxVal=m_RequestedXMargins[1]; xValDefined=true; }
	bool bFirstEmpty=ptFirst->IsEmpty();
	if(!xValDefined){ m_xMinVal=ptFirst->GetMinXValue(); m_xMaxVal=ptFirst->GetMaxXValue(); }
	if(!yValDefined){ m_yMinVal=ptFirst->GetMinYValue(); m_yMaxVal=ptFirst->GetMaxYValue(); }
	bool bMinYPlotEmpty=bFirstEmpty; long bMaxYPlotEmpty=bFirstEmpty;
	if (bFitAllPlotsY && (!yValDefined)) {
		for(i=1; i<m_nPlots; i++) {
			if(m_plots[i]==NULL)continue;
			bool bEmptyI=m_plots[i]->IsEmpty(); if(bEmptyI)continue;
			double iYMin=m_plots[i]->GetMinYValue(); if((iYMin<m_yMinVal)||(bMinYPlotEmpty)){m_yMinVal=iYMin;bMinYPlotEmpty=bEmptyI;}
			double iYMax=m_plots[i]->GetMaxYValue(); if((iYMax>m_yMaxVal)||(bMaxYPlotEmpty)){m_yMaxVal=iYMax;bMaxYPlotEmpty=bEmptyI;}
		}
	}
	if ((m_xMinVal>=m_xMaxVal) || (m_yMinVal>m_yMaxVal)) return false;
	if ((m_yMinVal==m_yMaxVal)&&(bMinYPlotEmpty)&&(bMaxYPlotEmpty)) { m_yMinVal=0; m_yMaxVal=1; } // Set [0,1] region for empty plot
	if (m_yMinVal==m_yMaxVal) { if(m_yMaxVal>0)m_yMinVal=0; else if (m_yMaxVal==0)m_yMaxVal=1; else m_yMaxVal=0; }
	if (!yValDefined) {
		if (m_bLogAxisY) {
			if((m_yMinVal<=0)&&(m_yMaxVal>0)){ m_yMinVal=m_yMaxVal/1000; }else{ if(m_yMaxVal<=0)m_bLogAxisY=false; }
		} else {
			if(m_yMinVal>0){ if((ContainBinPlots())||(m_yMinVal<m_yMaxVal*0.8))m_yMinVal=0; }
		}
	}
	if (m_bLogAxisX) {
		if((m_xMinVal<=0)&&(m_xMaxVal>0)){ m_xMinVal=m_xMaxVal/1000; }else{ if(m_xMaxVal<=0)m_bLogAxisX=false; }
	}

	//Axis
	out->SetLineWidth((!m_bThickGridLines)?CVMOutput::LW_THIN:CVMOutput::LW_NORMAL); //0.5,1.
	//[Old:] '(m_bDrawHorGrid || m_bDrawVertGrid)?' instead of '(!m_bThickGridLines)?'
	if (m_bDrawAxisBox) {
		out->MoveTo(m_xAxis,m_yAxis); out->LineTo(m_xAxis,m_yPosMax); out->LineTo(m_xPosMax,m_yPosMax);
		out->LineTo(m_xPosMax,m_yAxis); out->LineTo(m_xAxis,m_yAxis); //out->BlockEnd();
	} else {
		out->MoveTo(m_xAxis,m_yAxis-(m_yAxis-yMin)/2); out->LineTo(m_xAxis,yMax); //out->BlockEnd();
		out->MoveTo(m_xAxis-(m_xAxis-xMin)/2,m_yAxis); out->LineTo(xMax,m_yAxis); //out->BlockEnd();
	}

	//Axis subscriptions
	write_axis(out);

	//Draw additional 'level' line
	if (m_xLevelLineAt != 0.) {
		double xpos=GetXPos(m_xLevelLineAt);
		if((xpos>=m_xAxis)&&((xpos<=m_xPosMax))){
			out->SetColor(CVMRGB(0.4f,0.4f,0.4f)); out->SetLineWidth(CVMOutput::LW_THICK); out->MoveTo(xpos,m_yAxis); out->LineTo(xpos,m_yPosMax);
		}
	}
	if (m_yLevelLineAt != 0.) {
		double ypos=GetXPos(m_yLevelLineAt);
		if((ypos>=m_yAxis)&&((ypos<=m_yPosMax))){
			out->SetColor(CVMRGB(0.4f,0.4f,0.4f)); out->SetLineWidth(CVMOutput::LW_THICK); out->MoveTo(m_xAxis,ypos); out->LineTo(m_xPosMax,ypos);
		}
	}

	int iCurrLineType=0;
	bool bDashedExist=false; //affects legend style
	int nPlotsPrinted=0; // <m_nPlots if empty histos are drooped
	int iScatPlotNo=0; // count number of scatter-plots drawed
	for(i=0; i<m_nPlots; i++) {
		if (m_plots[i]==NULL) continue;
		out->SetLineWidth(m_LineWidth[i]);
		if ((m_bHideEmptyHist) && (m_plots[i]->IsEmpty())) continue;
		if((m_styPlot[i] & CVMHistLiteView::LN_MASK_DASH_INFO)!=iCurrLineType){
			iCurrLineType=(m_styPlot[i] & CVMHistLiteView::LN_MASK_DASH_INFO);
			bool bDense = ((iCurrLineType & CVMHistLiteView::LN_DENSE_DASH_FLAG) != 0);
			int iStyBasic = (iCurrLineType & CVMHistLiteView::LN_MASK_BASIC_STYLE);
			if(iStyBasic == CVMOutput::LINE_SOLID){ out->SetLineStyle(CVMOutput::LINE_SOLID); }
			else if(iStyBasic == CVMOutput::LINE_DASH){ out->SetLineStyle(CVMOutput::LINE_DASH,(bDense?1.3f:1.f)); bDashedExist=true; }
			else if(iStyBasic == CVMOutput::LINE_DOT){ out->SetLineStyle(CVMOutput::LINE_DOT,(bDense?1.3f:1.f)); bDashedExist=true; }
			else if(iStyBasic == CVMOutput::LINE_DASHDOT){ out->SetLineStyle(CVMOutput::LINE_DASHDOT,(bDense?1.3f:1.f)); bDashedExist=true; }
			else{ out->SetLineStyle(CVMOutput::LINE_SOLID); }
		}
		out->SetColor(m_clrPlot[i]);
		if(m_plots[i]->GetClassId()==IVMPlotData2D::PL_HIST) {
			write_plot_points(out,((CVMHistLite*)m_plots[i]),((m_styPlot[i]&CVMHistLiteView::LN_DRAW_BINS)==CVMHistLiteView::LN_DRAW_BINS));
		} else if(m_plots[i]->GetClassId()==IVMPlotData2D::PL_SCAT) {
			write_scat_points(out,((CVMScatLite*)m_plots[i]),iScatPlotNo); iScatPlotNo++;
		} else if(m_plots[i]->GetClassId()==IVMPlotData2D::PL_FREEDRAW) {
			write_fdraw_points(out,((CVMFreeDrawing*)m_plots[i]));
		} else {
			//
		}
		nPlotsPrinted++;
	}
	if(iCurrLineType!=0){out->SetLineStyle(CVMOutput::LINE_SOLID);iCurrLineType=0;}

	if (m_bDrawLegend) {
		//Change font if required
		float fnSz=out->GetFontSize(); bool bFontChanged=false; if(m_bPrintNEvts||m_bPrintIntegral){ out->SetFontSize(fnSz*0.7f); bFontChanged=true; }
		//
		//[m_iLegendPos] 1: RightTop, 2: RightBottom, 3: LeftTop, 4:LeftBottom
		if((m_iLegendPos<1)||(m_iLegendPos>4))m_iLegendPos=1; //Check parameter (default: 1)
		bool bLegTop=((m_iLegendPos==1)||(m_iLegendPos==3)); bool bLegLeft=((m_iLegendPos==3)||(m_iLegendPos==4));
		//Init legend position
		//Nt: 'yHigh' & 'xLeft' will be changed later for 'm_bLegendOverlapped' mode
		//Nt: use local 'lw' & 'lh' -- font may be changed here
		double lw=out->GetLetterWidth(); double lh=out->GetLetterHeight();
		double yHigh=((bLegTop)?(m_yPosMax-lh*0.5):(m_yAxis+lh*(1+nPlotsPrinted))); double xLeft=0.;
		int iPrintLoop=(m_bLegendOverlapped?0:1);
		for(; iPrintLoop<2; iPrintLoop++){ //loop #0 used for size estimation & box drawing
			int n=0; double yHeight=0.; double xWidth=0.;
			for(i=0; i<m_nPlots; i++) {
				if(m_plots[i]==NULL)continue;
				if ((m_bHideEmptyHist) && (m_plots[i]->IsEmpty())) continue;
				const char* hlabel=m_plots[i]->GetLegendTxt(); if(hlabel==NULL)continue; int l=0;
				if ((m_bPrintNEvts || m_bPrintIntegral) && (m_plots[i]->GetClassId()==IVMPlotData2D::PL_HIST)) {
					CVMHistLite* hist=((CVMHistLite*)m_plots[i]);
					if(m_bPrintNEvts && (!m_bPrintIntegral)){
						l=sprintf(buf,"%s %ld/%ld",hlabel,hist->GetNumEvents(),hist->GetNumOutmargins());
					}else if((!m_bPrintNEvts) && m_bPrintIntegral){
						l=sprintf(buf,"%s I=%g",hlabel,hist->GetIntegral());
					}else{ //(m_bPrintNEvts && m_bPrintIntegral)
						l=sprintf(buf,"%s I=%g, %ld/%ld",hlabel,hist->GetIntegral(),hist->GetNumEvents(),hist->GetNumOutmargins());
					}
					hlabel=buf;
				} else {
					for(l=0;l<100;l++) { if(*(hlabel+l)=='\0')break; } // inline length calc
				}
				if(xWidth<(lw*l))xWidth=(lw*l); yHeight+=lh*1.3;
				if(iPrintLoop==1){
					double xTxtMid=(m_bLegendOverlapped?(xLeft+lw+lw*l/2):((bLegLeft)?(m_xAxis+lw+lw*l/2):(m_xPosMax-lw-lw*l/2)));
					out->SetColor(m_clrPlot[i]); out->TextOut(xTxtMid,yHigh-lh*(1.3*n),hlabel);
					if(bDashedExist) { // draw piece of line
						if((m_styPlot[i] & CVMHistLiteView::LN_MASK_DASH_INFO)!=iCurrLineType){
							iCurrLineType=(m_styPlot[i] & CVMHistLiteView::LN_MASK_DASH_INFO);
							bool bDense = ((iCurrLineType & CVMHistLiteView::LN_DENSE_DASH_FLAG) != 0);
							int iStyBasic = (iCurrLineType & CVMHistLiteView::LN_MASK_BASIC_STYLE);
							if(iStyBasic == CVMOutput::LINE_SOLID){ out->SetLineStyle(CVMOutput::LINE_SOLID); }
							else if(iStyBasic == CVMOutput::LINE_DASH){ out->SetLineStyle(CVMOutput::LINE_DASH,(bDense?1.3f:1.f)); bDashedExist=true; }
							else if(iStyBasic == CVMOutput::LINE_DOT){ out->SetLineStyle(CVMOutput::LINE_DOT,(bDense?1.3f:1.f)); bDashedExist=true; }
							else if(iStyBasic == CVMOutput::LINE_DASHDOT){ out->SetLineStyle(CVMOutput::LINE_DASHDOT,(bDense?1.3f:1.f)); bDashedExist=true; }
							else{ out->SetLineStyle(CVMOutput::LINE_SOLID); }
						}
						out->MoveTo(m_xPosMax-lw*(3.5+l),yHigh-lh*(0.5+1.3*n)); out->LineTo(m_xPosMax-lw*(l+2),yHigh-lh*(0.5+1.3*n));
					}
				}
				n++;
			}
			if((iPrintLoop==0)&&(xWidth>0.)&&(yHeight>0.)){ //draw legend box
				if((!bFontChanged) && (xWidth>((m_xPosMax-m_xAxis)/2.))){ //Scale text if label is too long
					float f=0.7f; xWidth*=f; yHeight*=f; lw*=f; lh*=f; out->SetFontSize(fnSz*f); bFontChanged=true;
				}
				xWidth+=2.*lw; yHeight+=lh*0.4;
				yHigh=((bLegTop)?yMax:(yMin+yHeight))-lh*0.2; xLeft=((bLegLeft)?(xMin):(xMax-xWidth)); //update 'yHigh' & 'xLeft'
				double yBox1=((bLegTop)?yMax:(yMin+yHeight)); double yBox2=((bLegTop)?(yBox1-yHeight):yMin);
				double xBox1=xLeft; double xBox2=xLeft+xWidth;
				out->FillColor(xBox1,yBox1,xBox2,yBox2,m_clrBcgr);
				double w=out->GetLineWidth(); out->SetLineWidth(CVMOutput::LW_THIN); out->SetColor(CVMOutput::clrBlack); out->Rect(xBox1,yBox1,xBox2,yBox2); out->SetLineWidth(w);
			}
		}
		if(bFontChanged)out->SetFontSize(fnSz); //Restore font
		if(iCurrLineType!=0){out->SetLineStyle(CVMOutput::LINE_SOLID);iCurrLineType=0;}
	}

	//Draw axis labels (or move it to write_axis()?)
	//Params: m_bLablesAtCenter
	if(m_sXAxisLabel!=NULL){
		//parameter 'm_fXLabelYShift' is used at m_xAxis/m_xPosMax values set
		out->SetColor(CVMOutput::clrBlack); int taXOld,taYOld;
		if(m_bLablesAtCenter){
			taXOld=out->SetTextAlignX(CVMOutput::TA_CENTER); taYOld=out->SetTextAlignY(CVMOutput::TA_UP);
			out->TextOutAdv((m_xAxis+m_xPosMax)/2,out->GetLetterHeight()*1.2,m_sXAxisLabel,true);
			//
			//ver.2 [Overlap with numbers possible]
			//[ver.2] taXOld=out->SetTextAlignX(CVMOutput::TA_CENTER); taYOld=out->SetTextAlignY(CVMOutput::TA_BOTTOM);
			//[ver.2] out->TextOutAdv((m_xAxis+m_xPosMax)/2,out->GetLetterHeight()*0.1,m_sXAxisLabel,true);
		}else{
			taXOld=out->SetTextAlignX(CVMOutput::TA_RIGHT); taYOld=out->SetTextAlignY(CVMOutput::TA_UP);
			out->TextOutAdv(1.,m_yAxis-out->GetLetterHeight()*0.3,m_sXAxisLabel,true);
		}
		out->SetTextAlignX(taXOld); out->SetTextAlignY(taYOld);
	}
	if(m_sYAxisLabel!=NULL){
		//parameter 'm_fYLabelYShift' is used at m_yAxis/m_yPosMax values set
		out->SetColor(CVMOutput::clrBlack); int taXOld,taYOld;
		if(m_bLablesAtCenter){
			taXOld=out->SetTextAlignX(CVMOutput::TA_CENTER); taYOld=out->SetTextAlignY(CVMOutput::TA_UP);
			out->SetTextDirection(90);
			out->TextOutAdv(out->GetLetterHeight()*0.1,(m_yAxis+m_yPosMax)/2,m_sYAxisLabel,true);
			out->SetTextDirection(0);
		}else{
			taXOld=out->SetTextAlignX(CVMOutput::TA_LEFT); taYOld=out->SetTextAlignY(CVMOutput::TA_BOTTOM);
			out->TextOutAdv(0,m_yPosMax+out->GetLetterHeight()*(0.3+m_fYLabelYShift),m_sYAxisLabel,true);
		}
		out->SetTextAlignX(taXOld); out->SetTextAlignY(taYOld);
	}

	//Draw plot label
	if(m_sPlotLabel!=NULL){
		out->SetColor(CVMOutput::clrBlack); int taX,taY; double x,y;
		float fOld=out->SetFontSize(out->GetFontSize()*1.5f);
		double gap=out->GetLetterWidth()*0.6;
		switch(m_iLegendPos){ //1: RightTop, 2: RightBottom, 3: LeftTop, 4:LeftBottom
		case 1: taX=CVMOutput::TA_RIGHT; taY=CVMOutput::TA_UP; x=m_xPosMax-gap; y=m_yPosMax-gap; break;
		case 2: taX=CVMOutput::TA_RIGHT; taY=CVMOutput::TA_BOTTOM; x=m_xPosMax-gap; y=m_yAxis+gap; break;
		case 3: taX=CVMOutput::TA_LEFT; taY=CVMOutput::TA_UP; x=m_xAxis+gap; y=m_yPosMax-gap; break;
		case 4:
		default: taX=CVMOutput::TA_LEFT; taY=CVMOutput::TA_BOTTOM; x=m_xAxis+gap; y=m_yAxis+gap;
		}
		taX=out->SetTextAlignX(taX); taY=out->SetTextAlignY(taY);
		out->TextOutAdv(x,y,m_sPlotLabel,true);
		out->SetTextAlignX(taX); out->SetTextAlignY(taY); //restore
		out->SetFontSize(fOld);
	}

	//out->BlockEnd();
	return true;
}

void CVMHistLiteView::write_plot_points(CVMOutput* out, CVMHistLite* hist, bool bPlotBins) {
	bool bTruncOverlaps=(m_bDrawAxisBox && bPlotBins); //Nt: overlaps are forbidden if (m_bDrawAxisBox && bPlotBins)
	//
	if(hist==NULL)return;
	unsigned nBins = hist->GetNBins();
	bool bStartLine=true;
	double x_last=0; double y_last=-1; if(bPlotBins){ x_last=GetXPos(hist->GetBinXMargin(0)); if(x_last<m_xAxis)x_last=m_xAxis; y_last=m_yAxis; }
	bool bLoop=true;
	for(unsigned i=0; ((i<nBins) && bLoop); i++) {
		double x_curr = GetXPos(((bPlotBins)?hist->GetBinXMargin(i+1):hist->GetBinXCenter(i)));
		// x-margins MAY be different at current diagram ==> check it!
		if(x_curr<m_xAxis)continue;
		if(x_curr>m_xPosMax){ //Warning: (Right[last bin] ~~ m_xPosMax+eps) is possible
			if((!bPlotBins) || (x_last>=m_xPosMax))break;
			//Warning: (Right[last bin] ~~ m_xPosMax+eps) is possible ==> stop next loop only
			x_curr=m_xPosMax; bLoop=false;
		}
		if (hist->GetNumEventsBin(i)>0) {
			double vErr=0.; double vVal=hist->GetValue(i,&vErr); double y_curr=GetYPos(vVal);
			bool bPointInBox=((y_curr>=m_yAxis)&&(y_curr<=m_yPosMax));
			if (bPlotBins) {
				if(bTruncOverlaps && (!bPointInBox)){
					if(y_curr<m_yAxis){ y_curr=m_yAxis; }else if(y_curr>m_yPosMax){ y_curr=m_yPosMax; }
					if((!bStartLine) || (y_curr!=y_last)){ if(bStartLine)out->MoveTo(x_last,y_last); out->LineTo(x_last,y_curr); } //draw '|' between -Inf && + Inf
					bStartLine=true;
				}else{
					if(bStartLine){ out->MoveTo(x_last,y_last); bStartLine=false; }
					out->LineTo(x_last,y_curr); out->LineTo(x_curr,y_curr);
				}
				x_last=x_curr; y_last=y_curr;
			} else {
				if(bStartLine){ out->MoveTo(x_curr,y_curr); bStartLine=false; }else{ out->LineTo(x_curr,y_curr); }
			}
			if ((m_bDrawErrorMarks) && (vErr>0.)) {
				double y_ErrMin=GetYPos(vVal-vErr); double y_ErrMax=GetYPos(vVal+vErr); double x_BinC=((bPlotBins)?GetXPos(hist->GetBinXCenter(i)):x_curr);
				if(x_BinC<=m_xPosMax){ //'false' is possible in bPlotBins-mode if (Right[bin]>m_xPosMax)
					double w=out->GetLineWidth(); out->SetLineWidth(CVMOutput::LW_THIN); //0.25
					out->MoveTo(x_BinC,y_ErrMin); out->LineTo(x_BinC,y_ErrMax); out->SetLineWidth(w);
					out->MoveTo(x_curr,y_curr); //Move back to {x_curr,y_curr}
				}
			}
		} else if (bPlotBins) {
			if(bStartLine){ out->MoveTo(x_last,y_last); bStartLine=false; }
			if(y_last>m_yAxis)out->LineTo(x_last,m_yAxis);
			out->LineTo(x_curr,m_yAxis); x_last=x_curr; y_last=m_yAxis;
		}
	}
	if(bPlotBins){
		if(y_last>m_yAxis) {
			if(bStartLine){ out->MoveTo(x_last,y_last); bStartLine=false; }
			out->LineTo(x_last,m_yAxis);
		}
	}
}

void CVMHistLiteView::write_scat_points(CVMOutput* out, CVMScatLite* scat, int iOrder) {
	if(scat==NULL)return;
	unsigned long N=scat->GetNumEvents(); double wMax=scat->GetMaxWgt(); double wMin=scat->GetMinWgt();
	bool bUseWgt=(wMax>wMin); if(bUseWgt)wMin=0;
	double visMax=5.; double visMin=0.25; // visible size
	if(scat->IsOverfilled()) { out->TextOut((m_xAxis+m_xPosMax)/2,m_yAxis*0.4,"Overfilled"); }
	//else{ char buf[100]; int l=sprintf(buf,"Evts: %ld",N); out->TextOut((m_xAxis+m_xPosMax)/2,m_yAxis*0.4,buf,l); }
	for(unsigned long i=0; i<N; i++) {
		double x,y,w;
		if(!scat->GetValue(i,&x,&y,&w))continue;
		double x_curr = GetXPos(x); double y_curr = GetYPos(y);
		if((x_curr<m_xAxis) || (x_curr>m_xPosMax) || (y_curr<m_yAxis) || (y_curr>m_yPosMax)) continue;
		if ((m_bShiftMergedPoints) && (iOrder>0)) { // shift plot
			if (iOrder==1) { x_curr+=(m_xPosMax-m_xAxis)*0.01; }
			else if (iOrder==2) { y_curr+=(m_yPosMax-m_yAxis)*0.01; }
			else if (iOrder==3) { x_curr+=(m_xPosMax-m_xAxis)*0.01; y_curr+=(m_yPosMax-m_yAxis)*0.01; }
		}
		if(bUseWgt) { out->SetLineWidth(visMin+(visMax-visMin)*((w>=wMin)?(w-wMin):0.)/(wMax-wMin)); }
		out->MoveTo(x_curr,y_curr); out->LineTo(x_curr,y_curr);
	}
}

void CVMHistLiteView::write_fdraw_points(CVMOutput* out, CVMFreeDrawing* fdraw) {
	if(fdraw==NULL)return;
	unsigned long N=fdraw->GetNumEvents();
	if(fdraw->IsOverfilled()) { out->TextOut((m_xAxis+m_xPosMax)/2,m_yAxis*0.4,"Overfilled"); }
	//else{ char buf[100]; int l=sprintf(buf,"Evts: %ld",N); out->TextOut((m_xAxis+m_xPosMax)/2,m_yAxis*0.4,buf,l); }
	if(N>100)out->SetLineWidth(CVMOutput::LW_THIN); //small points
	for(unsigned long i=0; i<N; i++) {
		double x,y;
		int type=fdraw->GetAction(i,&x,&y); if(type<0)continue;
		double x_curr = GetXPos(x); double y_curr = GetYPos(y);
		//if((x_curr<m_xAxis) || (x_curr>m_xPosMax) || (y_curr<m_yAxis) || (y_curr>m_yPosMax)) continue;
		if(type==1) { out->MoveTo(x_curr,y_curr); } else if(type==2) { out->LineTo(x_curr,y_curr); }
	}
}

/*
*	_iArrMarkType: 1(text), 0(no text)
*	_iArrMarkLevel: 1(higest), 4+(lowest)
*/
bool CVMHistLiteView::calc_axis_log(double _MinVal, double _MaxVal, int nMarksMax,
		int nArrAlloc, int *_nMarks, double *_ArrMarks, double *_ArrMarkMantissa, int *_iArrMarkPow,
		int *_iArrMarkType, int *_iArrMarkLevel,
		double *_fComFactor, int *_iComFactorPow, bool *_bShowValuesAsExp)
{
	if ((_nMarks==NULL) || (_ArrMarks==NULL) || (_ArrMarkMantissa==NULL) || (_iArrMarkPow==NULL) ||
		(_iArrMarkType==NULL) || (_iArrMarkLevel==NULL) || 
		(_fComFactor==NULL) || (_iComFactorPow==NULL) || (_bShowValuesAsExp==NULL)) return false;
	if ((nMarksMax>nArrAlloc) || (nArrAlloc<100)) return false;
	if ((_MinVal<=0) || (_MinVal>=_MaxVal)) return false;

	// extract common power
	int iPowExtr=UtilGetPower(_MaxVal); if(iPowExtr!=UtilGetPower(_MinVal)){iPowExtr=0;} //separate similar power only
	if((iPowExtr>=-1)&&(iPowExtr<=2)) { iPowExtr=0; }
	*_iComFactorPow=iPowExtr;
	*_fComFactor=vmPow(10,iPowExtr);//asdf

	// calc marks values
	double _MarkMin=GetRoundMark(_MinVal,false,true); int iPowMin=((_MarkMin>0)?(int)(log(_MarkMin)/log((double)10)*1.001):0);
	double _MarkMax=GetRoundMark(_MaxVal,false,false); int iPowMax=((_MarkMax>0)?(int)(log(_MarkMax)/log((double)10)*1.001):0);
	int nMainMarks=iPowMax-iPowMin+1;
	int iPowPerMark=1; if(nMainMarks>nMarksMax)iPowPerMark=1+(nMainMarks/nMarksMax);
	if(iPowPerMark>5){ iPowPerMark=GetUpperRoundLong(iPowPerMark-1,false); } //round 'iPowPerMark': 6->5,7->10,
	*_bShowValuesAsExp=true; if((iPowExtr!=0) || ((_MarkMax<=1000)&&(_MarkMin>=0.01))) *_bShowValuesAsExp=false;

	// decimal marks values
	bool bShowDecSubMarks=((iPowPerMark==1) && (nMainMarks<5));
	bool bShowDecTextMarks=(bShowDecSubMarks && (_MarkMin>_MarkMax)); // no basic text marks

	// fill mark arrays
	double _ValCurr=_MarkMin; int iMarkPow=iPowMin; if(_ValCurr>_MinVal){ _ValCurr/=10; iMarkPow--; }
	*_nMarks=0;
	for (unsigned i=0; i<1000; i++) {// 1000 to avoid infinity loops
		if(i>0)iMarkPow++;
		if (bShowDecSubMarks) {
			for (int iSub=1; iSub<=9; iSub++) { // sub-marks before current mark
				if (*_nMarks>=nArrAlloc) break;
				double _SubVal=_ValCurr*0.1*iSub; if (_SubVal<_MinVal) continue; if (_SubVal>_MaxVal) break;
				_ArrMarks[*_nMarks]=_SubVal; _ArrMarkMantissa[*_nMarks]=iSub; _iArrMarkPow[*_nMarks]=iMarkPow-1;
				_iArrMarkType[*_nMarks]=((bShowDecTextMarks)?1:0); _iArrMarkLevel[*_nMarks]=3;
				(*_nMarks)++;
			}
		}
		if (*_nMarks>=nArrAlloc) break;
		if ((_ValCurr>=_MinVal) && (_ValCurr<=_MaxVal)) {
			_ArrMarks[*_nMarks]=_ValCurr; _ArrMarkMantissa[*_nMarks]=1; _iArrMarkPow[*_nMarks]=iMarkPow;
			//[Old]if(i%iPowPerMark == 0){
			if(abs(iMarkPow)%iPowPerMark == 0){
				_iArrMarkType[*_nMarks]=1; _iArrMarkLevel[*_nMarks]=1; // text mark
			}else{
				_iArrMarkType[*_nMarks]=0; _iArrMarkLevel[*_nMarks]=2; // label without text
			}
			(*_nMarks)++;
		}
		if (_ValCurr>_MaxVal) break;
		_ValCurr=_ValCurr*10;
	}
	return true;
}

bool CVMHistLiteView::calc_axis_lin(double _MinVal, double _MaxVal, int nMarksMin,
		int nArrAlloc, int *_nMarks, double *_ArrMarks, double *_ArrMarkMantissa, int *_iArrMarkPow,
		int *_iArrMarkType, int *_iArrMarkLevel,
		double *_fComFactor, int *_iComFactorPow, bool *_bShowValuesAsExp)
{
	if ((_nMarks==NULL) || (_ArrMarks==NULL) || (_ArrMarkMantissa==NULL) || (_iArrMarkPow==NULL) || (_iArrMarkType==NULL) || (_iArrMarkLevel==NULL) || 
		(_fComFactor==NULL) || (_iComFactorPow==NULL) || (_bShowValuesAsExp==NULL)) return false;
	if ((nMarksMin>nArrAlloc) || (nArrAlloc<100)) return false;
	if (_MinVal==_MaxVal) return false;
	if(nMarksMin<2)nMarksMin=2;

	// extract common power
	double _MaxValAbs=fabs(_MaxVal); if(_MaxValAbs<fabs(_MinVal))_MaxValAbs=fabs(_MinVal); //to get max. power
	int iPowExtr=UtilGetPower(_MaxValAbs); if((iPowExtr>=-1)&&(iPowExtr<=2)) { iPowExtr=0; }
	*_iComFactorPow=iPowExtr; *_fComFactor=vmPow(10,iPowExtr);
	*_bShowValuesAsExp=false; // ==> power & mantissa arrays are unused

	double _ActMinVal=_MinVal; double _ActMaxVal=_MaxVal; if(_MinVal>_MaxVal){_ActMinVal=_MaxVal;_ActMaxVal=_MinVal;}
	double _ValStep=GetRoundMark((_ActMaxVal-_ActMinVal)/(nMarksMin-1),true,false);					// visible values
	double _ValCurr=GetTruncToStep(_ActMinVal,_ValStep); if(_ValCurr<_ActMinVal)_ValCurr+=_ValStep;	// visible values
	if(_ValStep<=0.)return false;
	/*
	double _ValCurr=GetRoundMark(_ActMinVal+_ValStep,true,false);					// visible values
	if((_ValCurr-_ValStep)>=_ActMinVal)_ValCurr-=_ValStep;
	*/

	bool bShowDecSubMarks=((int)((_ActMaxVal-_ActMinVal)/_ValStep)<7);	//7 or less marks are estimated

	// fill mark arrays
	*_nMarks=0; int tmpPow=0; double tmpMant=0;
	for (int i=0;i<(nMarksMin*3);i++) { // allows extra marks if _ValStep is decreased by truncation
	//for (int i=0; i<1000; i++) {	// 1000 to avoid infinity loops
		if(bShowDecSubMarks){
			for(int iSub=1; iSub<=9; iSub++) { // sub-marks before current mark
				if (*_nMarks>=nArrAlloc) break;
				double _SubVal=_ValCurr-_ValStep+0.1*iSub*_ValStep; if (_SubVal<=_ActMinVal)continue; if(_SubVal>=_ActMaxVal)break;
				if(fabs(_SubVal/_ValStep)<0.000001)_SubVal=0.; // truncate {-2,-1,10^-17,1,2} to {-2,-1,0,1,2}
				tmpPow=UtilGetPower(_SubVal,&tmpMant); tmpMant=((int)(tmpMant+0.001)); // truncate 10^-16 artifacts
				_ArrMarks[*_nMarks]=_SubVal; _ArrMarkMantissa[*_nMarks]=tmpMant; _iArrMarkPow[*_nMarks]=tmpPow;
				_iArrMarkType[*_nMarks]=0; _iArrMarkLevel[*_nMarks]=((iSub==5)?2:3);
				(*_nMarks)++;
			}
		}
		if(_ValCurr>_ActMaxVal)break;
		if (_ValCurr>=_ActMinVal) {
			if(*_nMarks>=nArrAlloc)break;
			if(fabs(_ValCurr/_ValStep)<0.000001)_ValCurr=0.; // truncate {-2,-1,10^-17,1,2} to {-2,-1,0,1,2}
			tmpPow=UtilGetPower(_ValCurr,&tmpMant); tmpMant=((int)(tmpMant+0.001)); // truncate 10^-16 artifacts
			_ArrMarks[*_nMarks]=_ValCurr; _ArrMarkMantissa[*_nMarks]=tmpMant; _iArrMarkPow[*_nMarks]=tmpPow;
			_iArrMarkType[*_nMarks]=1; _iArrMarkLevel[*_nMarks]=1; //((tmpMant==1.)?1:2);
			(*_nMarks)++;
		}
		_ValCurr+=_ValStep;
	}
	return true;
}

void CVMHistLiteView::write_axis(CVMOutput* out) {
	double xMin=0; double xMax=1; double yMin=0; double yMax=1; out->GetMargins(&xMin,&xMax,&yMin,&yMax);
	int i; char buf[200]; char buf2[200]; int l,l2;
	bool xTxtC=true; // text x-align: center
	double lw=out->GetLetterWidth(); double lh=out->GetLetterHeight(); // average letter size
	float fnSz=out->GetFontSize();
	float fExpSize=0.7f;
	float fPowShift=0.5f; //=0.7f if (fExpSize==1.f)
	int arrLHsize=3; // size of markLx,markHx,markLy,markHy
	double markLx[3]={0.1,0.05,0.}; double markHx[3]={0.15,0.1,0.05}; //Axis mark 'relative' size
	double markLy[3]={0.1,0.05,0.}; double markHy[3]={0.1,0.07,0.03}; //[Values ignored!] Axis mark 'relative' size
	for(i=0;i<arrLHsize;i++){ //Scale values to absolute
		//double xMarkScale=(yMax-m_yAxis)*0.115; //Scale to image size. Note: x-Marks are 'vertical', || Oy.
		//[Old:]
		double xMarkScale=(m_yAxis-yMin); //Scale to border size. Note: x-Marks are 'vertical', || Oy.
		markLx[i]*=xMarkScale; markHx[i]*=xMarkScale;
		markLy[i]=markLx[i]; markHy[i]=markHx[i]; //Prev: markLy[i]*=(m_xAxis-xMin); markHy[i]*=(m_xAxis-xMin);
		if(m_bDrawAxisBox){ markLx[i]=0.; markLy[i]=0.; } //Remove out-of-box tales
	}
	bool bZeroPrinted=false; //to avoid 0/0 at axis crossing

	for(int idAxis=0; idAxis<2; idAxis++) {
		double _MinVal,_MaxVal; double _MinMarksGap,_MinTextMarksGap; int nMarksMax,nMarksMin; bool bLogAxis; //settings
		if (idAxis==0) { //Ox
			_MinVal=m_xMinVal; _MaxVal=m_xMaxVal; nMarksMax=5; nMarksMin=4; bLogAxis=m_bLogAxisX;
			_MinMarksGap=((m_xPosMax-m_xAxis)/400); //Prev: /250
			_MinTextMarksGap=((m_xPosMax-m_xAxis)/15);
		} else { //Oy
			_MinVal=m_yMinVal; _MaxVal=m_yMaxVal; nMarksMax=8; nMarksMin=5; bLogAxis=m_bLogAxisY;
			_MinMarksGap=((m_yPosMax-m_yAxis)/300); //Prev: /200
			_MinTextMarksGap=((m_yPosMax-m_yAxis)/15);
		}
		int nMarks=0; double fComFactor=0; int iComFactorPow=0; bool bShowValuesAsExp=true; bool bAxisOK=false;
		if (bLogAxis) bAxisOK=calc_axis_log(_MinVal,_MaxVal,nMarksMax,500,&nMarks,m_TmpMarks,m_TmpMarkMantissa,m_TmpMarkPows,m_TmpMarkType,m_TmpMarkLevel,&fComFactor,&iComFactorPow,&bShowValuesAsExp);
		else bAxisOK=calc_axis_lin(_MinVal,_MaxVal,nMarksMin,500,&nMarks,m_TmpMarks,m_TmpMarkMantissa,m_TmpMarkPows,m_TmpMarkType,m_TmpMarkLevel,&fComFactor,&iComFactorPow,&bShowValuesAsExp);
		if (!bAxisOK) continue;
		double posLastMark=-1; int lastLev=-1; int iLevSkip=100; double posLastTextMark=-1; int lastTxtLev=-1; int iTxtLevSkip=100;
		int iMarkLevMin=arrLHsize; int iMarkLevMax=-1; for(i=0;i<nMarks;i++){ int iLev=m_TmpMarkLevel[i]; if(iMarkLevMin>iLev)iMarkLevMin=iLev; if(iMarkLevMax<iLev)iMarkLevMax=iLev; }
		for (i=0; i<nMarks; i++) {
			double val=m_TmpMarks[i]; double vMant=m_TmpMarkMantissa[i]; int iPow=m_TmpMarkPows[i]-iComFactorPow; int iLev=m_TmpMarkLevel[i];
			double pos=((idAxis==0)?GetXPos(val):GetYPos(val));
			if (m_TmpMarkType[i]>0) { // text mark
				if ((iLev<iTxtLevSkip) && ((pos>=posLastTextMark+_MinTextMarksGap) || (iLev>lastTxtLev))) {
					bool bDrawText=true; if((val==0.)&&(_MinVal==0.)){ if(bZeroPrinted)bDrawText=false; bZeroPrinted=true; }
					if(bDrawText) {
						if ((bShowValuesAsExp) && (iPow!=0)) {
							l=((vMant==1.)?sprintf(buf,"10"):sprintf(buf,"%.0f*10",vMant)); l2=sprintf(buf2,"%d",iPow);
							if (idAxis==0) {
								out->TextOut(pos-lw*fExpSize*l2/2-((xTxtC)?(0):(lw*l/2)),m_yAxis-lh*(0.4+fPowShift),buf);
								out->SetFontSize(fnSz*fExpSize);
								out->TextOut(pos+lw*l/2-((xTxtC)?(0):(lw*fExpSize*l2/2)),m_yAxis-lh*0.4,buf2);
								out->SetFontSize(fnSz);
							} else {
								out->TextOut(m_xAxis-lw/2-lw*fExpSize*l2-((xTxtC)?(lw*l/2):(lw*l)),pos-lh*0.,buf);
								out->SetFontSize(fnSz*fExpSize);
								out->TextOut(m_xAxis-lw/2-((xTxtC)?(lw*fExpSize*l2/2):(lw*fExpSize*l2)),pos+lh*fPowShift,buf2);
								out->SetFontSize(fnSz);
							}
						} else {
							if(bShowValuesAsExp){ l=sprintf(buf,"%.0f",vMant); }else{ l=sprintf(buf,"%g",val/fComFactor); }
							if(idAxis==0){ out->TextOut(pos-lw/2-((xTxtC)?(0):(lw*l/2)),m_yAxis-lh*0.4,buf); }
							else{ out->TextOut(m_xAxis-lw/2-((xTxtC)?(lw*l/2):(lw*l)),pos+lh/2,buf); }
						}
						posLastTextMark=pos; lastTxtLev=iLev; iTxtLevSkip=100;
					}
					if ((idAxis==0) && m_bDrawVertGrid) {
						double wOld=out->GetLineWidth(); out->SetLineWidth(CVMOutput::LW_THIN/2.);
						out->SetLineStyle(CVMOutput::LINE_DASH); out->MoveTo(pos,m_yAxis); out->LineTo(pos,m_yPosMax); out->SetLineStyle(CVMOutput::LINE_SOLID);
						out->SetLineWidth(wOld);
					}
					else if((idAxis==1) && m_bDrawHorGrid) {
						double wOld=out->GetLineWidth(); out->SetLineWidth(CVMOutput::LW_THIN/2.);
						out->SetLineStyle(CVMOutput::LINE_DASH); out->MoveTo(m_xAxis,pos); out->LineTo(m_xPosMax,pos); out->SetLineStyle(CVMOutput::LINE_SOLID);
						out->SetLineWidth(wOld);
					}
				} else if ((bLogAxis) && (iLev<iTxtLevSkip)) { iTxtLevSkip=iLev; }
			}
			bool bDrawMark=((iLev<iLevSkip) && ((pos>=posLastMark+_MinMarksGap) || (iLev>lastLev)));
			if ((bDrawMark) && (bLogAxis) && (i<nMarks-1)) { if ((m_TmpMarkLevel[i+1]<iLev) && (pos+_MinMarksGap>((idAxis==0)?GetXPos(m_TmpMarks[i+1]):GetYPos(m_TmpMarks[i+1])))) bDrawMark=false; }
			if (bDrawMark) {
				int iMarkID=iLev-1; if(iMarkID<0)iMarkID=0; if(iMarkID>=arrLHsize)iMarkID=arrLHsize-1;
				//[Next:] Correction for L-1 non-text marks if no L-2 marks exist: L-1 --> L-2.
				if((m_TmpMarkType[i]<=0)&&(iMarkLevMax<arrLHsize)&&(iMarkID==(arrLHsize-2))&&(iMarkID>0))iMarkID=arrLHsize-1;
				if(idAxis==0){
					out->MoveTo(pos,m_yAxis-markLx[iMarkID]); out->LineTo(pos,m_yAxis+markHx[iMarkID]);
					if(m_bDrawAxisBox){ out->MoveTo(pos,m_yPosMax+markLx[iMarkID]); out->LineTo(pos,m_yPosMax-markHx[iMarkID]); }
				}else {
					out->MoveTo(m_xAxis-markLy[iMarkID],pos); out->LineTo(m_xAxis+markHy[iMarkID],pos);
					if(m_bDrawAxisBox){ out->MoveTo(m_xPosMax+markLy[iMarkID],pos); out->LineTo(m_xPosMax-markHy[iMarkID],pos); }
				}
				//out->BlockEnd();
				posLastMark=pos; lastLev=iLev; iLevSkip=100;
			} else if ((bLogAxis) && (iLev<iLevSkip)) { iLevSkip=iLev; }
		}
		if (iComFactorPow != 0) {
			l=sprintf(buf,"*10"); l2=sprintf(buf2,"%d",iComFactorPow); //Sinle line: l=sprintf(buf,"*10^%d",iComFactorPow);
			if (idAxis==0) {
				//Single line: out->TextOut(m_xPosMax-lw/2-((xTxtC)?(lw*l/2):(lw*l)),m_yAxis-lh*1.5,buf);
				int xAlign=out->SetTextAlignX(CVMOutput::TA_RIGHT);
				if(m_bAxisXFactorInsideBox){
					out->TextOut(m_xPosMax-lw*(0.5+l2)*fExpSize,m_yAxis+lh*2.,buf,l);
					out->SetFontSize(fnSz*fExpSize); out->TextOut(m_xPosMax-lw*0.5*fExpSize,m_yAxis+lh*(2.+fPowShift),buf2,l2); out->SetFontSize(fnSz);
				} else {
					out->TextOut(xMax-lw*l2*fExpSize,m_yAxis-lh*1.5,buf,l);
					out->SetFontSize(fnSz*fExpSize); out->TextOut(xMax,m_yAxis-lh*(1.5-fPowShift),buf2,l2); out->SetFontSize(fnSz);
				}
				out->SetTextAlignX(xAlign);
			} else {
				if(m_bAxisYFactorInsideBox){
					//Single line: out->TextOut(m_xAxis+((xTxtC)?(lw+lw*l/2):lw),m_yPosMax-lh/2,buf);
					int yAlign=out->SetTextAlignY(CVMOutput::TA_UP); int xAlign=out->SetTextAlignX(CVMOutput::TA_LEFT);
					out->TextOut(m_xAxis+lw*1.,m_yPosMax-lh*(0.3+fPowShift),buf,l);
					out->SetFontSize(fnSz*fExpSize); out->TextOut(m_xAxis+lw*(1.+l),m_yPosMax-lh*0.3,buf2,l2); out->SetFontSize(fnSz);
					out->SetTextAlignY(yAlign); out->SetTextAlignX(xAlign);
				}else{
					int yAlign=out->SetTextAlignY(CVMOutput::TA_UP); int xAlign=out->SetTextAlignX(CVMOutput::TA_RIGHT);
					//y-mark: Max(posLastTextMark+lh*0.5+lh*1.1, yMax-lh*fPowShift)
					double yFactorPos=(yMax-lh*fPowShift); if(yFactorPos<posLastTextMark+lh*1.8)yFactorPos=posLastTextMark+lh*1.8;
					out->TextOut(m_xAxis-lw*l2*fExpSize,yFactorPos,buf,l);
					out->SetFontSize(fnSz*fExpSize); out->TextOut(m_xAxis,yFactorPos+lh*fPowShift,buf2,l2); out->SetFontSize(fnSz);
					out->SetTextAlignY(yAlign); out->SetTextAlignX(xAlign);
				}
			}
		}
		//debug if ((idAxis==0) && (nMarks>=2)) { l=sprintf(buf,"%g",m_TmpMarks[nMarks-1]-m_TmpMarks[0]); out->TextOut(m_xPosMax-lw*20-((xTxtC)?(lw*l/2):(lw*l)),m_yAxis-lh*1.5,buf); }
	}
}

#endif //#ifdef VM_HEP_LIB_FULL
