/*
*	VMJetReco.cpp
*/
#include <math.h>
#include <iostream>
#include <fstream>

#include "VMLib/include/vmlib.h"
#ifdef VM_HEP_LIB_FULL
/*
#include "VMOutput.h"
#include "VMHistLite.h"
#include "VMGen.h"
#include "HEPEvtLite.h" //if not included in "VMJetDurham.h" file
#include "VMJetReco.h"
#include "SysSettings.h" // another path: "../../HEPEvtLite/src/SysSettings.h"
*/

#define CVMJet_2PI	6.283185307179586476
#define CVMJet_PI	3.141592653589793238
//////////////////////////////////////////////////////////////////////////////////////	CVMJetAlgorithm::CVMJetRec
// Internal marks for m_idJetForParticle[] array
#define CVMJetPart_Undetected	-1
#define CVMJetPart_Isolated		-2
#define CVMJetPart_CuttedJet	-3
//
CVMJetAlgorithm::CVMJetRec::CVMJetRec() { m_bUseFullRecord=true; }
CVMJetAlgorithm::CVMJetRec::CVMJetRec(const CVMJetAlgorithm::CVMJetRec& jet) {
	m_Mom=jet.m_Mom; m_nParticles=jet.m_nParticles;
	m_bUseFullRecord=true; if(m_bUseFullRecord){ m_Eta=jet.m_Eta; m_Et=jet.m_Et; m_Phi=jet.m_Phi; }
}
void CVMJetAlgorithm::CVMJetRec::Set(const CHEPParticleLite& part) {
	m_Mom=part.m_Mom; m_nParticles=1;
	if(m_bUseFullRecord) {
		m_Phi=DBLtR(part.m_Phi); //m_Phi=m_Mom.GetPhi();
		if(m_Phi<0.)m_Phi+=CVMJet_2PI; if(m_Phi>=CVMJet_2PI){m_Phi-=CVMJet_2PI;} //To [0,2Pi]
		m_Eta=DBLtR(part.m_Eta); //m_Eta=m_Mom.GetEta();
		double pt=DBLtR(part.m_Pt); if(pt<=0.)pt=DBLtR(m_Mom.GetPt()); m_Et=DBLtR(m_Mom.m_E*pt/part.m_Mom.GetPAbs());
	}
}
CVMJetAlgorithm::CVMJetRec& CVMJetAlgorithm::CVMJetRec::operator=(const CVMJetAlgorithm::CVMJetRec& jet) {
	m_Mom=jet.m_Mom; m_nParticles=jet.m_nParticles;
	if(m_bUseFullRecord){ m_Eta=jet.m_Eta; m_Et=jet.m_Et; m_Phi=jet.m_Phi; }
	return *this;
}
CVMJetAlgorithm::CVMJetRec& CVMJetAlgorithm::CVMJetRec::operator+=(const CVMJetAlgorithm::CVMJetRec& jet) {
	m_nParticles+=jet.m_nParticles;
	if(m_bUseFullRecord) {
		bool bAsFastJet=false;
		if(bAsFastJet) {
			double wi,wj;
			// Note: different schemes of weight selection may be here
			wi=DBLtR(m_Mom.GetPt()); wj=DBLtR(jet.m_Mom.GetPt()); //wi*=wi; wj*=wj;
			//wi=m_Et; wj=jet.m_Et;
			double wij=wi+wj;
			if(wij>0.) {
				// Fixate '-2Pi/2' shift: abs(_Phi_j-_Phi_i) MUST be <=Pi
				double _Phi_j=jet.m_Phi; while(_Phi_j>(m_Phi+CVMJet_PI)){_Phi_j-=CVMJet_2PI;} while(_Phi_j<=(m_Phi-CVMJet_PI)){_Phi_j+=CVMJet_2PI;}
				m_Phi=(m_Phi*wi+_Phi_j*wj)/wij;
				while(m_Phi>=CVMJet_2PI){m_Phi-=CVMJet_2PI;} while(m_Phi<0){m_Phi+=CVMJet_2PI;} //Normalize [0,2Pi)
				//while(m_Phi>CVMJet_PI){m_Phi-=CVMJet_2PI;} while(m_Phi<=-CVMJet_PI){m_Phi+=CVMJet_2PI;} //Normalize (-Pi,Pi]
				
				double pt_ij=DBLtR(m_Mom.GetPt() + jet.m_Mom.GetPt());
				//double pt_ij=m_Et + jet.m_Et;
				m_Eta=(m_Eta*wi+jet.m_Eta*wj)/wij;
				m_Mom.Set(pt_ij*cosh(m_Eta),pt_ij*cos(m_Phi),pt_ij*sin(m_Phi),pt_ij*sinh(m_Eta));
				m_Et=m_Et+jet.m_Et; // unused(?) here
			} else {
				m_Mom.Set(0.,0.,0.,0.); m_Eta=0.; m_Phi=0.; m_Et=0.;
			}
		} else {
			m_Mom+=jet.m_Mom;
			double wi,wj;
			// Note: different schemes of weight selection may be here
			wi=m_Et; wj=jet.m_Et;
			double wij=wi+wj;
			if(wij>0.) {
				// Fixate '-2Pi/2' shift: abs(_Phi_j-_Phi_i) MUST be <=Pi
				double _Phi_i=m_Phi; double _Phi_j=jet.m_Phi; while(_Phi_j>(_Phi_i+CVMJet_PI)){_Phi_j-=CVMJet_2PI;} while(_Phi_j<=(_Phi_i-CVMJet_PI)){_Phi_j+=CVMJet_2PI;}
				m_Phi=(_Phi_i*wi+_Phi_j*wj)/wij;
				while(m_Phi>=CVMJet_2PI){m_Phi-=CVMJet_2PI;} while(m_Phi<0){m_Phi+=CVMJet_2PI;} //Normalize [0,2Pi)
				//while(m_Phi>CVMJet_PI){m_Phi-=CVMJet_2PI;} while(m_Phi<=-CVMJet_PI){m_Phi+=CVMJet_2PI;} //Normalize (-Pi,Pi]

				//Fast combining (from kT algorithm)
				m_Et=m_Et+jet.m_Et;
				m_Eta=(m_Eta*wi+jet.m_Eta*wj)/wij;
			} else {
				m_Mom.Set(0.,0.,0.,0.); m_Eta=0.; m_Phi=0.; m_Et=0.;
			}
		}
	} else {
		m_Mom+=jet.m_Mom;
	}
	return *this;
}

//////////////////////////////////////////////////////////////////////////////////////	CVMJetAlgorithm::CVMParticlesSelection
CVMJetAlgorithm::CVMParticleFilter::CVMParticleFilter() {
	m_PMaxEta=-1.; m_PMinMomAbs=-1.; m_PMinPT=-1.; m_PNoLeptons=false; m_UseMarked=0; //m_PMinMomAbs=0.05;
}

//////////////////////////////////////////////////////////////////////////////////////	CVMJetAlgorithm
CVMJetAlgorithm::CVMJetAlgorithm(bool bUse2JetsCache) {
	m_evt=NULL; m_bSortFinalJets=false; m_MinJetPt=0.; m_bKeepPart2JetInfo=false; m_bMomOnlyUsed=false;
	m_Cache2Jets=((bUse2JetsCache)?(new double[NJETS_MAX*NJETS_MAX]):NULL); m_bUse2jCache=(m_Cache2Jets!=NULL);
	m_nJets=0; m_nFixedJets=0;
}
CVMJetAlgorithm::~CVMJetAlgorithm() {
	if(m_Cache2Jets!=NULL){ delete[] m_Cache2Jets; m_Cache2Jets=NULL; }
}
bool CVMJetAlgorithm::Run(CHEPEvtLite* evt) {
	m_evt=evt; if(m_evt==NULL)return false;

	//Adjust parameters
	if(m_MinJetPt>0.)m_bSortFinalJets=true;

	// Particle to '1-particle-jets' selection
	SelectParticlesToJets();

	// Fill 2-jets resolutions cache
	if(m_Cache2Jets!=NULL){
		for(int i=0;i<m_nJets-1;i++){ for(int j=i+1;j<m_nJets;j++){ m_Cache2Jets[i*NJETS_MAX+j]=Calc2JetsRes(i,j); } }
	}

	return true;
}
void CVMJetAlgorithm::SelectParticlesToJets() { //initial jets filling
	// Select all stable paricles + not neutrinos + use 'm_PartFilter' filter
	m_nJets=0; m_SumEnergy=0.; m_nParticlesUsed=0; m_nFixedJets=0;
	if(m_evt==NULL)return;
	for(long i=0; i<m_evt->m_nParticles; i++) {
		CHEPParticleLite* part=&m_evt->m_Particles[i]; m_idJetForParticle[i]=CVMJetPart_Undetected;

		int idPAbs=abs(part->m_PID);
		bool bDetectable = ((part->m_Status == 1) &&				// stable
			(idPAbs != 12) && (idPAbs != 14) && (idPAbs != 16) &&	// nu
			(part->m_Pt > m_PartFilter.m_PMinPT));					// |Pt|>m_PMinPT {GeV}
		if ((bDetectable) && (m_PartFilter.m_PMinMomAbs>0)) {		// |P|>m_PMinMomAbs {GeV}
			if(part->m_Mom.GetPSqr()<m_PartFilter.m_PMinMomAbs*m_PartFilter.m_PMinMomAbs)bDetectable=false;
		}
		if ((bDetectable) && (m_PartFilter.m_UseMarked>0)) {		// 'marked'
			if (part->m_Mark != m_PartFilter.m_UseMarked) {
				m_idJetForParticle[i]=CVMJetPart_Isolated;			//Note: consider 'marked' as 'isolated' particle
				bDetectable=false;
			}
		}
		if ((bDetectable) && (m_PartFilter.m_PMaxEta>0)) {			// Eta < m_PMaxEta
			if (part->m_Eta > m_PartFilter.m_PMaxEta)bDetectable=false;
		}
		if ((bDetectable) && (m_PartFilter.m_PNoLeptons)) {			// no leptons (consider as isolated)
			if ((idPAbs==11) || (idPAbs==13) || (idPAbs==15)) {
				m_idJetForParticle[i]=CVMJetPart_Isolated;
				bDetectable=false;
			}
		}
		if (bDetectable) {
			//std::cout<<"Detectable: "<<CAlHepLib::GetParticleTitle(part->m_PID)<<std::endl;
			if (m_nJets<NJETS_MAX) {
				m_Jets[m_nJets].m_bUseFullRecord=(!m_bMomOnlyUsed);
				m_Jets[m_nJets].Set(*part);
				m_SumEnergy+=DBLtR(part->m_Mom.m_E); //used by Durham algorithm
				if(m_bKeepPart2JetInfo)m_idJetForParticle[i]=m_nJets;
				m_nJets++;
			} else { std::cerr<<"ERROR: INCREASE NJETS_MAX VALUE!!!"<<std::endl; break; }
		} else {
			//std::cout<<"NOT Detectable: "<<CAlHepLib::GetParticleTitle(part->m_PID)<<std::endl;
		}
	}
	m_nParticlesUsed=m_nJets;
}
void CVMJetAlgorithm::CombineJets(int i, int j) { // Combine 2 jets
	if(i==j)return;
	int iMin=((i>j)?j:i); int iMax=((i>j)?i:j);
	m_Jets[iMin]+=m_Jets[iMax];
	if(m_bKeepPart2JetInfo){ int s=m_evt->m_nParticles; for(int n=0;n<s;n++){if(m_idJetForParticle[n]==iMax)m_idJetForParticle[n]=iMin;} }
	if(iMax<(m_nJets-1)){ // Exchange iMax<->m_nJets-1
		m_Jets[iMax]=m_Jets[m_nJets-1];
		if(m_bKeepPart2JetInfo){ int s=m_evt->m_nParticles; int iLast=(m_nJets-1); for(int n=0;n<s;n++){if(m_idJetForParticle[n]==iLast)m_idJetForParticle[n]=iMax;} }
	}

	if(m_bUse2jCache){ // Update 2-jets cache values (the only n>=m_nFixedJets values are important)
		// Exchange iMax<->m_nJets-1 are exchanged
		// TODO: or better to keep & shift sorted data?
		int n;
		if (iMax<(m_nJets-1)) {
			for(n=m_nFixedJets;n<iMax;n++) { m_Cache2Jets[n*NJETS_MAX+iMax]=m_Cache2Jets[n*NJETS_MAX+m_nJets-1]; }
			for(n=iMax+1;n<(m_nJets-1);n++) { m_Cache2Jets[iMax*NJETS_MAX+n]=m_Cache2Jets[n*NJETS_MAX+m_nJets-1]; }
		}
		// Shift for sorted data (erroneous if iRepl<->iLast are exchanged)
		/////for(i=0;i<iPairMax;i++) { for(j=iPairMax+1;j<m_nJets;j++) { m_Cache2Jets[i*NJETS_MAX+j-1]=m_Cache2Jets[i*NJETS_MAX+j]; } }
		/////for(i=iPairMax+1;i<(m_nJets-1);i++) { for(j=i+1;j<m_nJets;j++){ m_Cache2Jets[(i-1)*NJETS_MAX+(j-1)]=m_Cache2Jets[i*NJETS_MAX+j]; } }
		// Update cache for 'iMin' jet
		for(n=m_nFixedJets;n<iMin;n++) { m_Cache2Jets[n*NJETS_MAX+iMin]=Calc2JetsRes(n,iMin); }
		for(n=iMin+1;n<(m_nJets-1);n++) { m_Cache2Jets[iMin*NJETS_MAX+n]=Calc2JetsRes(iMin,n); }
	}
	m_nJets--;
}
void CVMJetAlgorithm::FixateJet(int i) {
	if(i>m_nFixedJets) {
		//Don't use SwapJets(m_nFixedJets,i) here!
		CVMJetRec tmpJet=m_Jets[i]; m_Jets[i]=m_Jets[m_nFixedJets]; m_Jets[m_nFixedJets]=tmpJet;
		if(m_bKeepPart2JetInfo){ int s=m_evt->m_nParticles; for(int n=0;n<s;n++){ if(m_idJetForParticle[n]==i){m_idJetForParticle[n]=m_nFixedJets;}else if(m_idJetForParticle[n]==m_nFixedJets){m_idJetForParticle[n]=i;} } }

		// Update cache values (the only n>=(m_nFixedJets+1) values are important)
		if(m_bUse2jCache){
			int n;
			for(n=m_nFixedJets+1;n<i;n++) { m_Cache2Jets[n*NJETS_MAX+i]=m_Cache2Jets[m_nFixedJets*NJETS_MAX+n]; }
			for(n=i+1;n<m_nJets;n++) { m_Cache2Jets[i*NJETS_MAX+n]=m_Cache2Jets[m_nFixedJets*NJETS_MAX+n]; }
		}
	}
	m_nFixedJets++;
}
void CVMJetAlgorithm::SwapJets(int i, int j) {
	if(i!=j){
		CVMJetRec tmpJet=m_Jets[i]; m_Jets[i]=m_Jets[j]; m_Jets[j]=tmpJet;
		if(m_bKeepPart2JetInfo){ int s=m_evt->m_nParticles; for(int n=0;n<s;n++){ if(m_idJetForParticle[n]==i){m_idJetForParticle[n]=j;}else if(m_idJetForParticle[n]==j){m_idJetForParticle[n]=i;} } }

		// Update cache values (the only n>=m_nFixedJets values are important)
		// WARNING: erroneuos values may occur when swapping(i<fixed,j>fixed)
		if(m_bUse2jCache){
			int iMin=((i>j)?j:i); int iMax=((i>j)?i:j);
			if(m_nFixedJets<iMax) {
				int n;
				for(n=m_nFixedJets;n<iMin;n++){ double tmp=m_Cache2Jets[n*NJETS_MAX+iMin]; m_Cache2Jets[n*NJETS_MAX+iMin]=m_Cache2Jets[n*NJETS_MAX+iMax]; m_Cache2Jets[n*NJETS_MAX+iMax]=tmp; }
				for(n=iMin+1;n<iMax;n++){ double tmp=m_Cache2Jets[n*NJETS_MAX+iMax]; m_Cache2Jets[n*NJETS_MAX+iMax]=m_Cache2Jets[iMin*NJETS_MAX+n]; m_Cache2Jets[iMin*NJETS_MAX+n]=tmp; }
				for(n=iMax+1;n<m_nJets;n++){ double tmp=m_Cache2Jets[iMin*NJETS_MAX+n]; m_Cache2Jets[iMin*NJETS_MAX+n]=m_Cache2Jets[iMax*NJETS_MAX+n]; m_Cache2Jets[iMax*NJETS_MAX+n]=tmp; }
			}
		}
	}
}
void CVMJetAlgorithm::SortFinalJets() { // by Pt
	//Note: To be called if jets are not sorted by algorythm automatically
	//TODO: cache pt values + invoke fastsort here for (m_nJets>10)
	//WARNING: It's strongly recommended to call after fixating all the jets only!
	if(m_nJets<0)return;
	int nJetsOrig=m_nJets;
	if(m_nFixedJets<m_nJets) { std::cerr<<"WARNING: CVMJetAlgorithm::SortFinalJets() call before fixating the jets!"<<std::endl; }
	//std::cout<<"Sort()"<<std::endl;
	for(int i=0;i<(m_nJets-1);i++) {
		double ptMax=DBLtR(m_Jets[i].m_Mom.GetPt()); int n_max=i;
		for(int j=i+1;j<m_nJets;j++){ double ptJ=DBLtR(m_Jets[j].m_Mom.GetPt()); if(ptJ>ptMax){ ptMax=ptJ; n_max=j; } }
		//std::cout<<i<<". Pt Max="<<ptMax<<std::endl;
		if (ptMax<m_MinJetPt) { m_nJets=i; break; }
		if(n_max>i)SwapJets(n_max,i);
	}
	
	if((m_bKeepPart2JetInfo) && (nJetsOrig>m_nJets)) {
		int s=m_evt->m_nParticles; for(int n=0;n<s;n++){ if(m_idJetForParticle[n]>=m_nJets){m_idJetForParticle[n]=CVMJetPart_CuttedJet;} }
	}
}
int CVMJetAlgorithm::CalcJetsNumber(double minJetPt, int minNParts) {
	int n=0;
	for(int i=0;i<m_nJets;i++) {
		double jetPt=DBLtR(m_Jets[i].m_Mom.GetPt()); int nParts=m_Jets[i].m_nParticles;
		if ((jetPt>=minJetPt) && (nParts>=minNParts)) n++;
	}
	return n;
}
void CVMJetAlgorithm::FillJetEvent(CHEPEvtLite* evtJets) { //Fill 'evt' with jets as particles (e.g. for further save&load)
	if(evtJets==NULL)return;
	evtJets->m_nParticles=m_nJets;
	for(int i=0;i<m_nJets;i++) {
		evtJets->m_Particles[i].m_Status=1; evtJets->m_Particles[i].m_PID=0; //0 as 'jet' ID
		evtJets->m_Particles[i].m_Pt=Jet(i).m_Mom.GetPt();
		evtJets->m_Particles[i].m_Mom=Jet(i).m_Mom; evtJets->m_Particles[i].m_Mark=Jet(i).m_nParticles;
		evtJets->m_Particles[i].m_Phi=Jet(i).m_Phi;
		evtJets->m_Particles[i].m_Eta=Jet(i).m_Eta; evtJets->m_Particles[i].m_Rapidity=Jet(i).m_Eta;
		evtJets->m_Particles[i].m_Mother2_f0=Jet(i).m_nParticles; //Debug
	}
}
void CVMJetAlgorithm::DrawEvent(const char* title, bool bDrawUnusedParticles) { //including jets
	if(m_evt==NULL)return;
#ifdef Decl_CHEPEvtLiteView
	char buf[200]; if(title==NULL)title="Evt&Jets";
	CHEPEvtLite* evtJ=new CHEPEvtLite(); //create dynamically to save stack
	FillJetEvent(evtJ);
	int i;
	//Settings
	bool bDrawUndetectables=bDrawUnusedParticles;
	bool bDrawIsolatedParticles=bDrawUnusedParticles; // 'leptons'& 'marked' particles
	bool bDrawCuttedJetParticles=bDrawUnusedParticles; // particles involved into 'cutted'(soft) jets
	bool bDrawNeutrinos=bDrawUnusedParticles;
	// tests
	bDrawUndetectables=true;
	bDrawIsolatedParticles=false;
	bDrawCuttedJetParticles=false;
	bDrawNeutrinos=false;
	//save original marks
	int _PartMarks[EVTL_MAX_PART]; for(i=0; i<m_evt->m_nParticles; i++){ _PartMarks[i]=m_evt->m_Particles[i].m_Mark; }
	//set drawing marks
	int nJets=evtJ->m_nParticles;
	for(i=0; i<m_evt->m_nParticles; i++){
		int iJetId=m_idJetForParticle[i]; int iJetMark;
		if(iJetId>=0) {
			iJetMark=iJetId+1;
		} else {
			iJetMark=-1; //set negative mark
			if(bDrawUndetectables && (iJetId==CVMJetPart_Undetected)){
				std::cout<<"Undet: "; m_evt->m_Particles[i].Log(false);
				iJetMark=0;
			}
			if(bDrawIsolatedParticles && (iJetId==CVMJetPart_Isolated)){ iJetMark=0; }
			if(bDrawCuttedJetParticles && (iJetId==CVMJetPart_CuttedJet)){ iJetMark=0; }
			if(bDrawNeutrinos && (iJetMark<0)) {
				int pid=m_evt->m_Particles[i].m_PID; int idPAbs=((pid>=0)?pid:-pid); if((idPAbs==12)||(idPAbs==14)||(idPAbs==16))iJetMark=0;
			}
		}
		m_evt->m_Particles[i].m_Mark=iJetMark;
	}
	for(i=0; i<evtJ->m_nParticles; i++){ evtJ->m_Particles[i].m_Mark=i+1; }
	{
		sprintf(buf,"%s_xy.eps",title); CHEPEvtLiteView view(buf,CHEPEvtLiteView::MODE_XY);
		view.m_xValMax=10; view.m_yValMax=view.m_xValMax; //Debug values. Insert 'average' here or remove line.
		view.m_bColorTrackByPID=false; view.m_bDashedTracksByPID[0]=true; view.m_bSkipNegativeMarks=true;
		view.AddEventToView(m_evt); view.AddEventToView(evtJ); view.SaveView();
	}
	{
		sprintf(buf,"%s_yz.eps",title); CHEPEvtLiteView view(buf,CHEPEvtLiteView::MODE_YZ);
		view.m_bColorTrackByPID=false; view.m_bDashedTracksByPID[0]=true; view.m_bSkipNegativeMarks=true;
		view.AddEventToView(m_evt); view.AddEventToView(evtJ); view.SaveView();
	}
	delete evtJ;
	//restore original marks
	for(i=0; i<m_evt->m_nParticles; i++){ m_evt->m_Particles[i].m_Mark=_PartMarks[i]; }
#else
	std::cerr<<"Function CVMJetAlgorithm::DrawEvent() is unavailable."<<std::endl;
#endif //Decl_CHEPEvtLiteView
}

//////////////////////////////////////////////////////////////////////////////////////	CVMJetKT
CVMJetKT::CVMJetKT(bool bUseCache) : CVMJetAlgorithm(bUseCache) {
	m_R=0.7;
}
CVMJetKT::~CVMJetKT() {}
bool CVMJetKT::Run(CHEPEvtLite* evt) {
	if(!CVMJetAlgorithm::Run(evt)) return false;
	int i,j; // local variables

	// kT jet clustering algorithm (according to hep-ph/9807532v1)
	while(m_nFixedJets<(m_nJets-1)){ // at least 2 non-fixed jets
		// Find jets pair with minimal resolution 'd_ij'
		int iPairMin=m_nFixedJets; int iPairMax=m_nFixedJets+1; double d2MinSqr=Get2JetsRes(iPairMin,iPairMax);
		for(i=m_nFixedJets; i<m_nJets-1; i++) {
			for(j=i+1; j<m_nJets; j++) { // j>i
				double dSqr=Get2JetsRes(i,j); if(dSqr<d2MinSqr){ d2MinSqr=dSqr; iPairMin=i; iPairMax=j; }
			}
		}

		// Find jets pair with minimal 'd_i'
		int iMin=m_nFixedJets; double dMin=CalcJetRes(iMin);
		for(i=iMin+1; i<m_nJets; i++) { double d=CalcJetRes(i); if(d<dMin){ dMin=d; iMin=i; } }

		// Compare 'd_ij' to 'd_i'
		if(d2MinSqr<=(dMin*dMin)) {
			CombineJets(iPairMin,iPairMax); // Combine 2 jets
		} else {
			FixateJet(iMin); // Store the jet as 'fixed'
		}
    }

	m_nFixedJets=m_nJets; // Fixate all jets for faster further sorting

	// Sort jets if required
	if(m_bSortFinalJets){
		//TODO: check is the following list 'inversion' makes further sorting faster or not.
		//for(i=0;i<(m_nJets/2);i++){ int j=m_nJets-i;if(i<j)SwapJets(i,j); }	//Inverse jets order
		SortFinalJets();													//The additional sorting is required
	}
	return true;
}
double CVMJetKT::Calc2JetsRes(int i, int j) { // returns resolution squared (d_ij)^2
	double dEta=Jet(i).m_Eta-Jet(j).m_Eta;
	double dPhi=Jet(i).m_Phi-Jet(j).m_Phi;
	while(dPhi>CVMJet_PI){dPhi-=CVMJet_2PI;} while(dPhi<=-CVMJet_PI){dPhi+=CVMJet_2PI;}
	double Et_i=Jet(i).m_Et; double Et_j=Jet(j).m_Et; double Emin=((Et_i<Et_j)?Et_i:Et_j);
	//double Rij=sqrt(dEta*dEta+dPhi*dPhi);
	//return Emin*sqrt(dEta*dEta+dPhi*dPhi); //kT jets resolution
	return Emin*Emin*(dEta*dEta+dPhi*dPhi); //kT jets resolution squared (faster)
}
double CVMJetKT::CalcJetRes(int i) {
	return Jet(i).m_Et*m_R;
}

//////////////////////////////////////////////////////////////////////////////////////	CVMJetDurham
CVMJetDurham::CVMJetDurham(bool bUseYCache) : CVMJetAlgorithm(bUseYCache) {
	m_nJetsRequired=1; m_yCut=-1.; // default parameters
	m_nMinPartInJet=1; m_HardJetMinMom=0.; // additional parameters (unused by default)
	m_bMomOnlyUsed=true; // faster operation
}
CVMJetDurham::~CVMJetDurham() {}
/*
*	Based on implementation by A.Raspereza
*/
bool CVMJetDurham::Run(CHEPEvtLite* evt) {
	if(!CVMJetAlgorithm::Run(evt)) return false;
	// Notes:
	// 1. m_evt is already set to evt, particles selected to 1-particle-jets, m_nParticlesUsed & m_SumEnergy filled.
	if (m_nJets<2) return true;
	if (m_SumEnergy<=0.)return false;

	// adjust parameters
	if (m_nMinPartInJet<=0) m_nMinPartInJet=1;
	if (m_nJetsRequired<1) m_nJetsRequired=1;
	bool bCalcHardJets=(m_HardJetMinMom > 0.);
	bool bNeedFillFinals=(m_yCut>0);

	// local variables
	int nJetsMult=m_nJets;	// (additional) number of multiparticle jets (n particles>=m_nMinPartInJet)
	int nJetsHard=m_nJets;	// (additional) number of 'hard' jets (P>m_HardJetMinMom)
	int i,j;
	//'Copy of jets to save as finals' removed due to 'jet-to-particle' info must also be saved
	//Old://int nyCutJets=-1;					// copied to 'm_nJets' at the final
	//Old://CVMJetRec yCutJets[NJETS_MAX];	// copied to 'm_Jets' at the final

	// Clear output varuables
	for (i=0; i<m_nParticlesUsed; i++) {
		m_yResN[i]=0; m_nMultiPartJets[i]=0; m_nHardJets[i]=0;
		if ((bCalcHardJets) && (Jet(i).m_Mom.GetPSqr()>m_HardJetMinMom*m_HardJetMinMom))nJetsHard++;
	}

	// Durham jet clustering algorithm
	while (m_nJets>=(m_nJetsRequired+1)) {	// minimum 2 jets here
		// Find jets pair with minimal 'y'
		int iPairMin=0; int iPairMax=1; double yResMin=Get2JetsRes(iPairMin,iPairMax);
		for(i=0; i<m_nJets-1; i++) {
			for(j=i+1; j<m_nJets; j++) { // j>i
				double yRes=Get2JetsRes(i,j); if(yRes<yResMin){ yResMin=yRes; iPairMin=i; iPairMax=j; }
			}
		}
		// Save final jets array copy (when y reaches the cut)
		if(bNeedFillFinals){
			if(yResMin>m_yCut) {
				//Removed copying due to 'jet-to-particle' info must also be saved
				//Old: nyCutJets=m_nJets; for(j=0;j<nyCutJets;j++){ yCutJets[j]=m_Jets[j]; } bNeedFillFinals=false;
				break;
			}
		}
		// Save additional output
		m_yResN[m_nJets-1]=yResMin; // Minimal resolution of (N+1) jets (filled from [1] to [m_nParticlesUsed-1])
		m_nMultiPartJets[m_nJets-1]=nJetsMult; // Number of multiparticle jets
		m_nHardJets[m_nJets-1]=nJetsHard;

		// Combine 2 closest jets
		if(Jet(iPairMin).m_nParticles>=m_nMinPartInJet)nJetsMult--; if(Jet(iPairMax).m_nParticles>=m_nMinPartInJet)nJetsMult--;
		if(bCalcHardJets){ if(Jet(iPairMin).m_Mom.GetPSqr()>=m_HardJetMinMom*m_HardJetMinMom)nJetsHard--; if(Jet(iPairMax).m_Mom.GetPSqr()>=m_HardJetMinMom*m_HardJetMinMom)nJetsHard--; }
		CombineJets(iPairMin,iPairMax);
		if(Jet(iPairMin).m_nParticles>=m_nMinPartInJet)nJetsMult++; if(bCalcHardJets){ if(Jet(iPairMin).m_Mom.GetPAbs()>=m_HardJetMinMom*m_HardJetMinMom)nJetsHard++; }
    }

	// Removed copying due to 'jet-to-particle' info must also be saved
	//Old: //Save copied jets (nyCutJets, yCutJets) as final set
	//Old: if(nyCutJets>=0){ m_nJets=nyCutJets; for(i=0;i<nyCutJets;i++){ m_Jets[i]=yCutJets[i]; } }

	m_nFixedJets=m_nJets; // Fixate all jets for faster further sorting

	// Sort jets if required
	if(m_bSortFinalJets)SortFinalJets();
	return true;
}
int CVMJetDurham::GetNJetsRes(double yCut) { // get nJets from m_yResN[] array
	int nJets=-1; // error value
	// Note: m_yResN[i] is min-y between (i+1) jets ==> m_yResN[0] is always unfilled
	if (m_yResN[1]>yCut) {
		//for(int i=m_nParticlesUsed-1; i>0; i--) { if(m_yResN[i]>yCut){ nJets=i+1; break; } }
		nJets=m_nParticlesUsed; // default result for too small yCut
		for(int i=2;i<m_nParticlesUsed;i++){ if(m_yResN[i]<yCut){nJets=i;break;} } //same to previous line but faster
	} else {
		nJets=((m_nParticlesUsed>0)?1:0);
	}
	return nJets;
}
double CVMJetDurham::Calc2JetsRes(int i, int j) { // Durham Algorithm yres
	const CVM4Vec* p1=&Jet(i).m_Mom; const CVM4Vec* p2=&Jet(j).m_Mom;
	double E1Sqr=DBLtR(p1->m_E*p1->m_E); double E2Sqr=DBLtR(p2->m_E*p2->m_E);
	double emin2 = ((E1Sqr>E2Sqr)?E2Sqr:E1Sqr);	// Durham Algorithm
	// double emin2 = p1->m_E*p2->m_E;			// Jade Algorithm
	return 2.*emin2*(1.-DBLtR(p1->GetAngleCos3Vec(*p2)))/(m_SumEnergy*m_SumEnergy);
}
//double m_yResN[NJETS_MAX];		// Minimal resolution of (N+1) jets (filled from [1] to [m_nParticlesUsed-1])
//int m_nMultiPartJets[NJETS_MAX];// Number of jets with Npart>m_nMinPartInJet
//int m_nHardJets[NJETS_MAX];		// Number of 'hard' jets (P>m_HardJetMinMom)

/*
CVMJetDurham::CVMJetDurham() {
	m_evt=NULL; m_nEventsNJETS_MAX=0; m_yCache = new double[NJETS_MAX*NJETS_MAX];
	// default parameters
	m_nMinJets=2; m_nMinPartInJet=1; m_PMaxEta=-1; m_PMinMomAbs=0.05; m_PMinPT=-1; m_PNoLeptons=false; m_bSortFinalJets=false;
	m_HardJetMom=0.; // minimal 'hard jet' momentum
}

CVMJetDurham::~CVMJetDurham() {
	delete[] m_yCache;
}

bool CVMJetDurham::run(CHEPEvtLite* evt, double yCut) {
	m_evt=evt;
	if (m_evt==NULL) return false;
	bool bRes=RunDurhamImpl(yCut);
	return bRes;
}

int CVMJetDurham::GetNJetsRes(double yCut) { // get nJets from m_yResN[] array
	int nJets=-1; // error value
	// Note: m_yResN[i] is min-y between (i+1) jets ==> m_yResN[0] is always unfilled
	if (m_yResN[1]>yCut) {
		//for(int i=m_nDetectables-1; i>0; i--) { if(m_yResN[i]>yCut){ nJets=i+1; break; } }
		nJets=m_nDetectables; // default result for too small yCut
		for(int i=2;i<m_nDetectables;i++){ if(m_yResN[i]<yCut){nJets=i;break;} } //same to previous line but faster
	} else {
		nJets=((m_nDetectables>0)?1:0);
	}
	return nJets;
}

double CVMJetDurham::CalcYRes(CVM4Vec* p1, CVM4Vec* p2, double evs) { // Durham Algorithm yres
	double E1Sqr=p1->m_E*p1->m_E; double E2Sqr=p2->m_E*p2->m_E;
	double emin2 = ((E1Sqr>E2Sqr)?E2Sqr:E1Sqr);	// Durham Algorithm
	//  double emin2 = p1->m_E*p2->m_E;			// Jade Algorithm
	return 2.*emin2*(1.-p1->GetAngleCos3Vec(*p2))/(evs*evs);
}

// Based on implementation by A.Raspereza
bool CVMJetDurham::RunDurhamImpl(double yCut) {
	//std::cout<<"RunDurhamImpl() call. N particles = "<<m_evt->m_nParticles<<std::endl;

	// misc. settings
	bool bUseCache=true;
	m_bFinalJetsFilled=false;

	// adjust parameters
	if (m_nMinPartInJet<=0) m_nMinPartInJet=1;
	if ((bUseCache) && (m_yCache==NULL)) bUseCache=false;
	if (m_nMinJets<2) m_nMinJets=2;
	bool bCalcHardJets=(m_HardJetMom > 0.);

	// buffer variables
	int nPartJet[NJETS_MAX];
	CVM4Vec pJet[NJETS_MAX];
	int nJets=0;		// number of jets
	int nJetsMult=0;	// number of multiparticle jets (n particles>=m_nMinPartInJet)
	int nJetsHard=0;	// number of 'hard' jets (P>m_HardJetMom)

	int i,j;
	m_Evis = 0.0;	// visible energy
	// readin all stable paricles 2 b used for clustering
	for (i=0; i<m_evt->m_nParticles; i++) {
		CHEPParticleLite* part=&m_evt->m_Particles[i];
		double ptot = part->m_Mom.GetPAbs(); int idPAbs=abs(part->m_PID);
		//double cosTheta = fabs(part->m_Mom.m_Pz/ptot); //vm: why abs?
		bool bDetectable = ((part->m_Status == 1) &&				// stable
			(idPAbs != 12) && (idPAbs != 14) && (idPAbs != 16) &&	// nu
			(ptot > m_PMinMomAbs) &&								// |P|>m_PMinMomAbs {GeV}
			(part->m_Pt > m_PMinPT));								// |Pt|>m_PMinPT {GeV}
		if ((m_PMaxEta>0) && (bDetectable)) {						// Eta < m_PMaxEta
			if (part->m_Eta > m_PMaxEta) bDetectable=false;
		}
		if ((m_PNoLeptons) && (bDetectable)) {
			if ((idPAbs==11) || (idPAbs==13) || (idPAbs==15)) bDetectable=false;
		}
		if (bDetectable) {
			if (nJets<NJETS_MAX) {
				m_Evis+=part->m_Mom.m_E;
				nPartJet[nJets]=1;
				pJet[nJets]=part->m_Mom;
				m_yResN[nJets]=0; m_nMultiPartJets[nJets]=0; // clear values
				m_nHardJets[nJets]=0;
				if ((bCalcHardJets) && (ptot>m_HardJetMom)) nJetsHard++;
				nJets++;
			} else {
				m_nEventsNJETS_MAX++; break;
			}
		}
	}
	if(m_nMinPartInJet==1)nJetsMult=nJets; // always
	if(!bCalcHardJets)nJetsHard=nJets; // always
	m_nDetectables = nJets;

	bool bNeedFillFinals=(yCut>0);
	//std::cout<<"m_nDetectables: "<<m_nDetectables<<std::endl;

	if (bUseCache) { // fill cache
		for(i=0;i<nJets-1;i++){ for(j=i+1;j<nJets;j++){ m_yCache[i*NJETS_MAX+j]=CalcYRes(&pJet[i],&pJet[j],m_Evis); } }
	}

	// Jet clustering
	double yres_min,yres;
	int iParticle,jParticle,iReplace,iNew;
	while (nJets>=m_nMinJets) {	// minimum 2 jets here
		iParticle=0; jParticle=1;
		if (bUseCache) {
			yres_min=m_yCache[0*NJETS_MAX+1];
		} else {
			yres_min=CalcYRes(&pJet[0],&pJet[1],m_Evis);
		}
		for(i=0; i<nJets-1; i++) {
			for(j=i+1; j<nJets; j++) {
				if (bUseCache) {
					yres=m_yCache[i*NJETS_MAX+j];
				} else {
					yres=CalcYRes(&pJet[i],&pJet[j],m_Evis);
				}
				if (yres<yres_min) { yres_min=yres; iParticle=i; jParticle=j; }
			}
		}
		//std::cout<<"nJ="<<nJets<<", iMin="<<iParticle<<", jMin="<<jParticle<<", yMin="<<yres_min<<std::endl;
		//std::cout<<"y="<<CalcYRes(&pJet[iParticle],&pJet[jParticle],m_Evis)<<std::endl;

		if (bNeedFillFinals) {
			// if(log(yres_min)>yCutLn) {
			if(yres_min>yCut) {	// fill array of final jets when y reaches the cut
				m_nJetFnl = nJets;
				//std::cout<<"N jets resolved: "<<m_nJetFnl <<std::endl;
				for(j=0; j<nJets; j++){
					m_nPartJetFnl[j]=nPartJet[j];
					m_pJetFnl[j]=pJet[j];
				}
				bNeedFillFinals=false; m_bFinalJetsFilled=true;
			}
		}
		m_yResN[nJets-1]=yres_min; // Minimal resolution of (N+1) jets (filled from [1] to [m_nDetectables-1])
		m_nMultiPartJets[nJets-1]=nJetsMult; // Number of multiparticle jets
		m_nHardJets[nJets-1]=nJetsHard;

		// Combine 2 closest jets
		iNew     = ((iParticle>jParticle)?jParticle:iParticle); // min
		iReplace = ((iParticle>jParticle)?iParticle:jParticle); // max
		if(nPartJet[iNew]>=m_nMinPartInJet)nJetsMult--; if(nPartJet[iReplace]>=m_nMinPartInJet)nJetsMult--;
		if(bCalcHardJets){ if(pJet[iNew].GetPAbs()>m_HardJetMom)nJetsHard--; if(pJet[iReplace].GetPAbs()>m_HardJetMom)nJetsHard--; }
		pJet[iNew]=pJet[iParticle]+pJet[jParticle];
		nPartJet[iNew]=nPartJet[iNew]+nPartJet[iReplace];
		if(nPartJet[iNew]>=m_nMinPartInJet)nJetsMult++;
		if(bCalcHardJets){ if(pJet[iNew].GetPAbs()>m_HardJetMom)nJetsHard++; }
		if (iReplace<(nJets-1)) {
			pJet[iReplace]=pJet[nJets-1];
			nPartJet[iReplace]=nPartJet[nJets-1];
		}
		if (bUseCache) { // update cache
			// iReplace<->nJets-1 are exchanged
			if (iReplace<(nJets-1)) {
				for(i=0;i<iReplace;i++) { m_yCache[i*NJETS_MAX+iReplace]=m_yCache[i*NJETS_MAX+nJets-1]; }
				for(i=iReplace+1;i<(nJets-1);i++) { m_yCache[iReplace*NJETS_MAX+i]=m_yCache[i*NJETS_MAX+nJets-1]; }
			}
			// Shift sorted data (erroneous if iRepl<->iLast are exchanged)
			//for(i=0;i<iReplace;i++) { for(j=iReplace+1;j<nJets;j++) { m_yCache[i*NJETS_MAX+j-1]=m_yCache[i*NJETS_MAX+j]; } }
			//for(i=iReplace+1;i<(nJets-1);i++) { for(j=i+1;j<nJets;j++){ m_yCache[(i-1)*NJETS_MAX+(j-1)]=m_yCache[i*NJETS_MAX+j]; } }

			j=iNew;
			for(i=0;i<j;i++) {
				m_yCache[i*NJETS_MAX+j]=CalcYRes(&pJet[i],&pJet[j],m_Evis);
			}
			i=iNew;
			for(j=i+1;j<(nJets-1);j++) {
				m_yCache[i*NJETS_MAX+j]=CalcYRes(&pJet[i],&pJet[j],m_Evis);
			}
		}
		nJets--;
    }

	if(m_bSortFinalJets)SortFinalJets();
	return true;
}

void CVMJetDurham::SortFinalJets() { // by Pt
	if (!m_bFinalJetsFilled) return;

	//TODO: cache pt values + invoke fastsort here for (m_nJetFnl>10)
	CVM4Vec vecTmp; int nPartTmp;
	for(int i=0;i<(m_nJetFnl-1);i++) {
		double ptMax=m_pJetFnl[i].GetPt(); int n_max=i;
		for(int j=i+1;j<m_nJetFnl;j++) {
			double ptJ=m_pJetFnl[j].GetPt(); if (ptJ>ptMax) { ptMax=ptJ; n_max=j; }
		}
		if (n_max>i) {
			vecTmp=m_pJetFnl[n_max]; m_pJetFnl[n_max]=m_pJetFnl[i]; m_pJetFnl[i]=vecTmp;
			nPartTmp=m_nPartJetFnl[n_max]; m_nPartJetFnl[n_max]=m_nPartJetFnl[i]; m_nPartJetFnl[i]=nPartTmp;
		}
	}
}
*/

#endif //#ifdef VM_HEP_LIB_FULL
