////////////////////////////////////////////////////////
//	This file is a part of ALHEP project
//	Author: V.Makarenko (makarenko@hep.by)
//	See details at http://www.hep.by/alhep
////////////////////////////////////////////////////////

// VMOutput.cpp: implementation of the CVMOutput class.
//
//////////////////////////////////////////////////////////////////////

#include <iostream>
#include <fstream>
#include <time.h>
#include <math.h>
//#include <iomanip> for std::cout<<std::setprecision(7)<<std::setw(8);

#include "VMLib/include/vmlib.h"

#ifdef VM_HEP_LIB_FULL
//#include "VMOutput.h"
//#include "SysSettings.h"

/////////////////////////////////////////////////////////////////////// CVMRGB
CVMRGB::CVMRGB() { m_Red=0; m_Green=0; m_Blue=0; }
CVMRGB::CVMRGB(float r, float g, float b) { m_Red=r; m_Green=g; m_Blue=b; }
CVMRGB::CVMRGB(const CVMRGB& c) { m_Red=c.m_Red; m_Green=c.m_Green; m_Blue=c.m_Blue; }
CVMRGB &CVMRGB::operator=(const CVMRGB& c) { m_Red=c.m_Red; m_Green=c.m_Green; m_Blue=c.m_Blue; return *this; }
bool CVMRGB::operator==(const CVMRGB& c) const { return ((m_Red==c.m_Red)&&(m_Green==c.m_Green)&&(m_Blue==c.m_Blue)); }
bool CVMRGB::operator!=(const CVMRGB& c) const { return ((m_Red!=c.m_Red)||(m_Green!=c.m_Green)||(m_Blue!=c.m_Blue)); }
CVMRGB& CVMRGB::operator+=(const CVMRGB& c) { // MIX with other color
	m_Red=(m_Red+c.m_Red)/2; m_Green=(m_Green+c.m_Green)/2; m_Blue=(m_Blue+c.m_Blue)/2; return *this;
}
CVMRGB CVMRGB::operator+(const CVMRGB& c) const { // MIX with other color. Note: "+=" is faster
	return CVMRGB((m_Red+c.m_Red)/2,(m_Green+c.m_Green)/2,(m_Blue+c.m_Blue)/2);
}
void CVMRGB::Set(float r, float g, float b) { m_Red=r; m_Green=g; m_Blue=b; }

/////////////////////////////////////////////////////////////////////// CVMOutput
bool CVMOutputSettings::m_bSaveOutputTime	= true; //Def: true, use 'false' for simplier comparison of plots
//
int CVMOutput::FF_COURIER	= 0;
int CVMOutput::FF_SYMBOL	= 1;
int CVMOutput::FF_ARIAL		= 2;
const char* CVMOutput::GetFontFamilyID(int iFontFamily) {
	const char* res = "Unknown";
	if(iFontFamily == CVMOutput::FF_COURIER) { res="Courier"; }
	if(iFontFamily == CVMOutput::FF_SYMBOL) { res="Symbol"; }
	if(iFontFamily == CVMOutput::FF_ARIAL) { res="Arial"; }
	return res;
}
//
float CVMOutput::LW_THIN	= 0.25f;
float CVMOutput::LW_NORMAL	= 0.75f;
float CVMOutput::LW_THICK	= 1.25f;
//
const int CVMOutput::TA_LEFT	= -1;
const int CVMOutput::TA_BOTTOM	= -1;	//==TA_LEFT
const int CVMOutput::TA_CENTER	= 0;
const int CVMOutput::TA_RIGHT	= 1;
const int CVMOutput::TA_UP		= 1;	//==TA_RIGHT
//
const int CVMOutput::LINE_SOLID		= 0;
const int CVMOutput::LINE_DASH		= 1;
const int CVMOutput::LINE_DOT		= 2;
const int CVMOutput::LINE_DASHDOT	= 3;
const int CVMOutput::LINE_OTHER		= 4; // set for specific styles (e.g. dots with non-default rate)
//
CVMOutput::CVMOutput() {
	m_bStarted=false; m_bFinalized=false;
	m_aspect_ratio=0.61803; m_xMin=0; m_xMax=1; m_yMin=0; m_yMax=m_aspect_ratio; m_xCurr=0.; m_yCurr=0.;
	m_fPad2LogicalCoordScale=1.; //to be updated after the pad size is defined
	m_clrDraw.Set(0,0,0);
	m_clrBackground.Set(1,1,1);
	m_dLineWidth=CVMOutput::LW_NORMAL; m_bLineWidthChanged=false;
	m_FontSize=10.f; m_bFontSizeChanged=false;
	//TODO: should we start from 'm_bXXXChanged=true' and exclude default settings from file headers?
	m_TextAlignX=CVMOutput::TA_CENTER; //default
	m_TextAlignY=CVMOutput::TA_UP; //default
	m_fTxtAngle=0.f;
	m_bFontFamilyChanged=false;
	m_iFontFamily=CVMOutput::FF_COURIER;
	m_fBlockRotationAngle=0.f;
	m_bUseMathTextMacros=true;
	//m_bUseMathTextMacros=false;
	//[Unused: 's' is written BEFORE next output]m_bBlockEnded=false; //'stroke' is just written (for EPS-output)
}
CVMOutput::~CVMOutput() {}
void CVMOutput::Start() { m_bStarted=true; }
void CVMOutput::Finalize() { m_bFinalized=true; }
void CVMOutput::Rect(double x1, double y1, double x2, double y2, bool bFill) {
	if(bFill)FillColor(x1,y1,x2,y2,m_clrBackground);
	MoveTo(x1,y1);LineTo(x1,y2);LineTo(x2,y2);LineTo(x2,y1);LineTo(x1,y1);
}
void CVMOutput::Arc(double x, double y, double r, double phi1, double phi2) { return; } //'not implemented'
float CVMOutput::SetFontSize(float sz) { float res=m_FontSize; if(m_FontSize!=sz){ m_FontSize=sz; m_bFontSizeChanged=true; } return res; }
void CVMOutput::SetFontFamily(int iFontFamily) { if(m_iFontFamily!=iFontFamily)m_bFontFamilyChanged=true;  m_iFontFamily=iFontFamily; }
double CVMOutput::GetFontSizeScaled() {
	double res=m_FontSize*m_fPad2LogicalCoordScale;
	if(m_iFontFamily != CVMOutput::FF_COURIER){ //scale other fonts to the size of 'CVMOutput::FF_COURIER'
		if(m_iFontFamily == CVMOutput::FF_SYMBOL)res*=0.85; //Symbol
		//else if(m_iFontFamily == CVMOutput::FF_ARIAL)res*=0.95; //Arial
		else if(m_iFontFamily == CVMOutput::FF_ARIAL)res*=0.9; //Arial
	}
	return res;
}
double CVMOutput::GetFontWidthFactor() {
	double res=1.0;
	if(m_iFontFamily != CVMOutput::FF_COURIER){ //scale other fonts to the size of 'CVMOutput::FF_COURIER'
		if(m_iFontFamily == CVMOutput::FF_SYMBOL)res=1./0.85; //Symbol
		else if(m_iFontFamily == CVMOutput::FF_ARIAL)res=0.7; //Arial
	}
	return res;
}
int CVMOutput::SetTextAlignX(int xAlign) { int iRes=m_TextAlignX; m_TextAlignX=xAlign; return iRes; }
int CVMOutput::SetTextAlignY(int yAlign) { int iRes=m_TextAlignY; m_TextAlignY=yAlign; return iRes; }
float CVMOutput::SetTextDirection(float fTxtAngle) { float angPrev=m_fTxtAngle; for(;fTxtAngle<0.f;){fTxtAngle+=360.f;} for(;fTxtAngle>=360.f;){fTxtAngle-=360.f;} if(m_fTxtAngle!=fTxtAngle){m_fTxtAngle=fTxtAngle;/*m_bTxtAngleChanged=true;*/} return angPrev; }
bool CVMOutput::UseMathTextMacros(bool bUse) {
	bool bRes=m_bUseMathTextMacros; m_bUseMathTextMacros=bUse;
	m_bFontFamilyChanged=true; //set 're-declare font' flag
	return bRes;
}
//
void CVMOutput::StartBlockRotation(double x, double y, float fBlockAngle) { for(;fBlockAngle<0.f;){fBlockAngle+=360.f;} for(;fBlockAngle>=360.f;){fBlockAngle-=360.f;} m_fBlockRotationAngle=fBlockAngle; }
void CVMOutput::StopBlockRotation() { m_fBlockRotationAngle=0.f; }
//
void CVMOutput::SetLineWidth(double w) { if(m_dLineWidth!=w){ m_dLineWidth=w; m_bLineWidthChanged=true; } }
//
void CVMOutput::SetLineStyle(int lStyle, float fRate) {
	if(lStyle == CVMOutput::LINE_OTHER) { lStyle=CVMOutput::LINE_SOLID; }
	if(fRate<=0.f)fRate=1.f;

	if(lStyle == CVMOutput::LINE_SOLID) {
		SetLineDash(0,NULL);
	}else if(lStyle == CVMOutput::LINE_DASH) {
		double szDash=(m_xMax-m_xMin)/100/fRate;
		double szArr[2]={szDash,szDash}; SetLineDash(2,szArr);
	}else if(lStyle == CVMOutput::LINE_DOT) {
		double szDot=(m_xMax-m_xMin)/2000;
		double szGap=szDot*20/fRate;
		double szArr[2]={szDot,szGap}; SetLineDash(2,szArr);
		//[ 0.0005 0.005 ] 0 setdash
	}else if(lStyle == CVMOutput::LINE_DASHDOT) {
		double szDot=(m_xMax-m_xMin)/2000;
		double szGap=szDot*20/fRate;
		double szDash=szGap;
		double szArr[4]={szDot,szGap,szDash,szGap}; SetLineDash(4,szArr);
	}
}
CVMRGB CVMOutput::SetColorBackground(CVMRGB rgb) { CVMRGB res=m_clrBackground; m_clrBackground=rgb; return res; }
void CVMOutput::FillBackground(CVMRGB rgb) { FillColor(m_xMin,m_yMin,m_xMax,m_yMax,rgb); }
void CVMOutput::SetColor(CVMRGB rgb) { m_clrDraw=rgb; }
void CVMOutput::FillContour(CVMRGB rgb) {
	//Not suported
	std::cout<<"Warning: CVMOutput::FillContour() method is not supported."<<std::endl;
}
void CVMOutput::FillContour() { FillContour(m_clrBackground); }
bool CVMOutput::IsOpen() { return (m_bStarted && (!m_bFinalized)); }
void CVMOutput::GetMargins(double* xMin, double* xMax, double* yMin, double* yMax) {
	if(xMin!=NULL)*xMin=m_xMin; if(xMax!=NULL)*xMax=m_xMax; if(yMin!=NULL)*yMin=m_yMin; if(yMax!=NULL)*yMax=m_yMax;
}
void CVMOutput::SetMargins(double xMin, double xMax, double yMin, double yMax) {
	if ((xMax>xMin) && (yMax>yMin)) {
		m_xMin=xMin; m_xMax=xMax, m_yMin=yMin; m_yMax=yMax; m_aspect_ratio=(m_yMax-m_yMin)/(m_xMax-m_xMin);
	}
}
/*
*	bAllowYFit: allow or not y-correction.
*	'true' may lead to 'dancing' lines after consecutive TextOutAdv() calls.
*	------ TODO: Move 'out::TextOutAdv' to TeXParser class !!! -----
*	Note: works in '!m_bUseMathTextMacros' mode only, elswhere some parts of string are NOT on the single line.
*/
void CVMOutput::TextOutAdv(double x, double y, const char* text, bool bAllowYFit) { //Parse TeX line (^,_,\gamma etc.)
	// ------ TODO: Move 'out::TextOutAdv' to TeXParser class !!! -----
	if(text==NULL)return;
	// cycle#1: len calc, cycle#1: printing
	bool bError=false;
	double lw=GetLetterWidth(); double lh=GetLetterHeight(); float fSz=GetFontSize();
	int iLenMax=1000;
	char* buf=new char[iLenMax+2]; if(buf==NULL)return;
	int xAlignPrev=m_TextAlignX; m_TextAlignX=CVMOutput::TA_LEFT;
	int yAlignPrev=m_TextAlignY; m_TextAlignY=CVMOutput::TA_BOTTOM;
	double xPos=x; double yPos=y;
	double yBoxMin=yPos; double yBoxMax=yPos+lh;
	// No 'return' further!
	bool bUseMathTextMacrosPrev=m_bUseMathTextMacros; if(bUseMathTextMacrosPrev)UseMathTextMacros(false);
	float fAngle=m_fTxtAngle; if(fAngle!=0.f){ StartBlockRotation(x,y,fAngle); m_fTxtAngle=0.f; } //Rotate whole block
	for(int iStep=0;iStep<2;iStep++){
		if(bError)break;
		if(iStep>0){
			double xLen=xPos-x; xPos=x; //xPos was updated during previous step
			if(xAlignPrev==CVMOutput::TA_RIGHT){ xPos-=xLen; }else if(xAlignPrev==CVMOutput::TA_CENTER){ xPos-=xLen/2.; }
			yPos=y;
			if(bAllowYFit){
				if(yAlignPrev==CVMOutput::TA_UP){ // 'y'->'yBoxMax'
					yPos=y-(yBoxMax-y);
				}else if(yAlignPrev==CVMOutput::TA_CENTER){ // 'y'->'(yBoxMin+yBoxMax)/2'
					yPos=y-((yBoxMax+yBoxMin)/2.-y);
				}else if(yAlignPrev==CVMOutput::TA_BOTTOM){ // 'y'->'yBoxMin'
					yPos=y+(y-yBoxMin);
				}
			}else{
				if(yAlignPrev==CVMOutput::TA_UP){ yPos-=lh; }else if(yAlignPrev==CVMOutput::TA_CENTER){ yPos-=lh/2.; }
			}
		}
		//float f=1.f; double h=0.; //font factor, height offset
		int nDepthMax=100; int iDepth;
		float f[100]; double h[100];/*font factor, height offset*/ for(iDepth=0;iDepth<nDepthMax;iDepth++){ f[iDepth]=1.f; h[iDepth]=0.; }
		int nBrackets[100]; for(iDepth=0;iDepth<nDepthMax;iDepth++){ nBrackets[iDepth]=0; }
		bool bBufIsCmd=false;
		char sCmd[100]; sCmd[0]='\0'; int lCmd=-1; //to save TeX commands
		float fLast=1.f;
		int i=0; char c; iDepth=0;
		for(;;){
			//TeX-string parsing
			if(bError)break;
			if(i>(iLenMax-3)){ std::cout<<"[ERROR][CVMOutput::TextOutAdv] String too long"<<std::endl; bError=true; break; }
			int iBuf=0; bool bSubItemEnd=false;
			for(;;){
				if((iBuf>(iLenMax-3)) || (i>(iLenMax-3)))break;
				c=*(text+i);
				if(c=='\0'){ i++; break; }
				if(c=='^'){ i++; break; }
				if(c=='_'){ i++; break; }
				if(c=='{'){ if(iBuf>0)break; i++; nBrackets[iDepth]++; continue; }
				if(c=='}'){
					nBrackets[iDepth]--;
					if(nBrackets[iDepth]<0){ std::cout<<"Error TeX string parsing: unexpected '}'"<<std::endl; nBrackets[iDepth]=0; }
					i++;
					if((iDepth>0) && (nBrackets[iDepth]<=0))bSubItemEnd=true;
					break;
				}
				if(c=='\\'){
					if(iBuf>0)break; //and don't skip '\'-symbol
					i++; bBufIsCmd=true;
					for(;;){
						if((iBuf>(iLenMax-3)) || (i>(iLenMax-3)))break;
						c=*(text+i);
						//if((CVMUtil::bIsSymbol(c))||(CVMUtil::bIsNumber(c))){ i++; *(buf+iBuf)=c; iBuf++; continue; }
						if((c>='a')&&(c<='z')){ i++; *(buf+iBuf)=c; iBuf++; continue; }
						if(c==' '){i++;} //don't skip symbol if it isn't space
						break;
					}
					if((iDepth>0)&&(nBrackets[iDepth]<=0)){ bSubItemEnd=true; }
					break;
				}
				//if(c==' ')... //skip spaces here?
				i++; *(buf+iBuf)=c; iBuf++;
				if((iDepth>0) && (nBrackets[iDepth]<=0)){ bSubItemEnd=true; break; }
			}
			*(buf+iBuf)='\0';
			//std::cout<<"buf: "<<buf<<", iBuf: "<<iBuf<<", bBufIsCmd="<<bBufIsCmd<<", bSubItemEnd="<<bSubItemEnd<<std::endl;
			if(bBufIsCmd){
				bool bGreek=false;
				if(CVMUtil::strEqu(buf,"alpha")){ iBuf=sprintf(buf,"a"); bGreek=true; }
				else if(CVMUtil::strEqu(buf,"beta")){ iBuf=sprintf(buf,"b"); bGreek=true; }
				else if(CVMUtil::strEqu(buf,"gamma")){ iBuf=sprintf(buf,"g"); bGreek=true; }
				else if(CVMUtil::strEqu(buf,"delta")){ iBuf=sprintf(buf,"d"); bGreek=true; }
				else if(CVMUtil::strEqu(buf,"theta")){ iBuf=sprintf(buf,"q"); bGreek=true; }
				else if(CVMUtil::strEqu(buf,"Theta")){ iBuf=sprintf(buf,"Q"); bGreek=true; }
				else if(CVMUtil::strEqu(buf,"sigma")){ iBuf=sprintf(buf,"s"); bGreek=true; }
				else if(CVMUtil::strEqu(buf,"mu")){ iBuf=sprintf(buf,"m"); bGreek=true; }
				else if(CVMUtil::strEqu(buf,"nu")){ iBuf=sprintf(buf,"n"); bGreek=true; }
				else if(CVMUtil::strEqu(buf,"pm")){ //0xb1=='', but '-' in Courier font
					*((unsigned char*)buf)=0xb1; iBuf=1; *(buf+iBuf)='\0'; bGreek=true;
				}
				if(bGreek){
					if(iStep>0){
						if(f[iDepth]!=fLast){ SetFontSize(fSz*f[iDepth]); fLast=f[iDepth]; }
						int ffPrev=m_iFontFamily; if(ffPrev!=CVMOutput::FF_SYMBOL)SetFontFamily(CVMOutput::FF_SYMBOL);
						TextOut(xPos,yPos+h[iDepth]*lh,buf,iBuf);
						if(ffPrev!=CVMOutput::FF_SYMBOL)SetFontFamily(ffPrev);
					}
					if(yBoxMax<(yPos+h[iDepth]*lh+lh*f[iDepth]))yBoxMax=yPos+h[iDepth]*lh+lh*f[iDepth];
					//[currently unused] if(yBoxMin>(yPos+h[iDepth]*lh))yBoxMin=(yPos+h[iDepth]*lh)
					xPos+=iBuf*lw*f[iDepth];
					//
					if(bSubItemEnd && (iDepth>0))iDepth--; if(c=='\0')break; continue;
				}
				//else ignore command '\' symbol
			}
			if(iBuf>0){
				if(iStep>0){
					if(f[iDepth]!=fLast){ SetFontSize(fSz*f[iDepth]); fLast=f[iDepth]; }
					TextOut(xPos,yPos+h[iDepth]*lh,buf,iBuf);
				}
				if(yBoxMax<(yPos+h[iDepth]*lh+lh*f[iDepth]))yBoxMax=yPos+h[iDepth]*lh+lh*f[iDepth];
				//[currently unused] if(yBoxMin>(yPos+h[iDepth]*lh))yBoxMin=(yPos+h[iDepth]*lh)
				xPos+=iBuf*lw*f[iDepth];
			}
			if(c=='^'){ //Is it still correct to use 'c' here?
				if(iDepth<(nDepthMax-1)){
					h[iDepth+1]=h[iDepth]+0.6*f[iDepth]; f[iDepth+1]=f[iDepth]*0.7f; iDepth++;
				}
			}
			if(c=='_'){
				if(iDepth<(nDepthMax-1)){
					h[iDepth+1]=h[iDepth]-0.2*f[iDepth]; f[iDepth+1]=f[iDepth]*0.7f; iDepth++;
				}
			}
			if(bSubItemEnd && (iDepth>0))iDepth--; if(c=='\0')break;
		}
		if(iStep>0){ if(fLast!=1.f)SetFontSize(fSz); }
	}
	m_TextAlignX=xAlignPrev; m_TextAlignY=yAlignPrev;
	if(fAngle!=0.f){ StopBlockRotation(); m_fTxtAngle=fAngle; } //Restore 'm_fTxtAngle'
	if(bUseMathTextMacrosPrev)UseMathTextMacros(true);
	delete[] buf;
}
//
const CVMRGB CVMOutput::clrRed(1,0,0); // = CVMRGB(1,0,0)
const CVMRGB CVMOutput::clrBlue(0,0,1); // = CVMRGB(0,0,1);
const CVMRGB CVMOutput::clrGreen(0,1,0); // = CVMRGB(0,1,0);
const CVMRGB CVMOutput::clrWhite(1,1,1); // = CVMRGB(1,1,1);
const CVMRGB CVMOutput::clrBlack(0,0,0); // = CVMRGB(0,0,0);
const CVMRGB CVMOutput::clrYellow(1,1,0); // = CVMRGB(1,1,0);
const CVMRGB CVMOutput::clrMagenta(1,0,1); // = CVMRGB(1,0,1);
const CVMRGB CVMOutput::clrCyan(0,1,1); // = CVMRGB(0,1,1);
const CVMRGB CVMOutput::clrGrey(0.75f,0.75f,0.75f); // = CVMRGB(0.75f,0.75f,0.75f);
const CVMRGB CVMOutput::clrGreyDark(0.5f,0.5f,0.5f); // = CVMRGB(0.5f,0.5f,0.5f)
const CVMRGB CVMOutput::clrGreyLight(0.90f,0.90f,0.90f); // = CVMRGB(0.90f,0.90f,0.90f);
const CVMRGB CVMOutput::clrGreyLightLight(0.95f,0.95f,0.95f); // = CVMRGB(0.95f,0.95f,0.95f);
const CVMRGB CVMOutput::clrOrange(1,0.65f,0); // = CVMRGB(1,0.65f,0);
const CVMRGB CVMOutput::clrLawnGreen(0.49f,0.98f,0); // = CVMRGB(0.49f,0.98f,0); //'salat'
const CVMRGB CVMOutput::clrKhaki(0.9f,0.82f,0.58f); // = CVMRGB(0.9f,0.82f,0.58f);
const CVMRGB CVMOutput::clrBrown(0.58f,0.25f,0.11f); // = CVMRGB(0.58f,0.25f,0.11f);
const CVMRGB CVMOutput::clrPink(1,0.4f,0.7f); // = CVMRGB(1,0.4f,0.7f); //Original: (1,0.75,0.8)
const CVMRGB CVMOutput::clrViolet(0.93f,0.5f,0.93f); // = CVMRGB(0.93f,0.5f,0.93f);
const CVMRGB CVMOutput::clrYellowGreen(0.6f,0.8f,0.2f); // = CVMRGB(0.6f,0.8f,0.2f);
const CVMRGB CVMOutput::clrDefault(-1.f,-1.f,-1.f); //(-1.,-1.,-1.)

/////////////////////////////////////////////////////////////////////// CVMFileOutput
#define  CVMFILEOUT_BUFMIN	100000
CVMFileOutput::CVMFileOutput(const char* fn, bool bKeepOpened) : CVMOutput() {
	if(fn==NULL)fn="NULL"; // copy filename
	int i; int lfn=-1; for(i=0;i<255;i++) { if (*(fn+i)=='\0'){lfn=i;break;} } if(lfn<=0){ fn="Invalid Name"; lfn=12; }
	m_fn=new char[lfn+2]; for(i=0;i<=lfn;i++){*(m_fn+i)=*(fn+i);}
	m_pOfs=NULL; m_buf=NULL; m_lBufLen=0; m_lBufPos=0;
	m_bKeepOpened=bKeepOpened;
}
CVMFileOutput::~CVMFileOutput() {
	if (m_fn!=NULL) delete[]m_fn;
	if (m_pOfs!=NULL) { if(m_pOfs->is_open())m_pOfs->close(); delete m_pOfs; }
	if (m_buf!=NULL) delete[]m_buf;
}
void CVMFileOutput::SetUseBuffer(long lBufLen) {
	if (IsOpen()) return; // too late to change settings
	if (m_buf!=NULL) { delete[]m_buf; m_buf=NULL; }
	m_lBufLen=lBufLen;
	if (m_lBufLen<CVMFILEOUT_BUFMIN)m_lBufLen=CVMFILEOUT_BUFMIN;
	m_buf=new char[m_lBufLen+2];
	m_lBufPos=0; if(m_buf!=NULL) *(m_buf+m_lBufPos)='\0';
}
void CVMFileOutput::Start() {
	CVMOutput::Start();
	if (m_buf==NULL) {
		if (m_bKeepOpened) {
			m_pOfs=new std::ofstream(m_fn,std::ios::out|std::ios::trunc);
		} else {
			std::ofstream ofs(m_fn,std::ios::out|std::ios::trunc); if(ofs){ofs.close();} //truncate file only
		}
	}
}
void CVMFileOutput::Finalize() {
	if (m_buf==NULL) {
		if (m_bKeepOpened) { m_pOfs->close(); delete m_pOfs; m_pOfs=NULL; }
	} else {
		std::ofstream ofs(m_fn,std::ios::out|std::ios::trunc);
		if (ofs) {
			ofs<<m_buf; //ofs.write(m_buf,m_lBufPos+1);
			ofs.close();
		}
	}
	CVMOutput::Finalize();
}
void CVMFileOutput::Write(const char* str) {
	if (str==NULL) return;
	if (m_buf==NULL) {
		if (m_bKeepOpened) {
			if (m_pOfs!=NULL) {
				if(m_pOfs->is_open()) { *m_pOfs<<str; }
			}
		} else {
			std::ofstream ofs(m_fn,std::ios::out|std::ios::app);
			if (ofs) { ofs<<str; /*ofs.write(str,len);*/ ofs.close(); }
		}
	} else {
		m_lBufPos+=sprintf((m_buf+m_lBufPos),"%s",str); // "%s",str -> to retain '%'-symbols
	}
}
std::ofstream* CVMFileOutput::GetStreamPtr() {
	if ((m_buf==NULL) || (!m_bKeepOpened)) return NULL;
	return m_pOfs;
}

/////////////////////////////////////////////////////////////////////// CVMOutputMath
CVMOutputMath::CVMOutputMath(const char* fn) : CVMFileOutput(fn,true) {
	m_idCurrCell=-1;
	m_nCells=0;
	m_fPad2LogicalCoordScale=1./230.375; //Mathematica default size (w:230.375, h:142.3125)
}
CVMOutputMath::~CVMOutputMath() {}
void CVMOutputMath::Start() {
	CVMFileOutput::Start();
	// write nb-header
	Write("(*NotebookFileLineBreakTest\nNotebookFileLineBreakTest*)\n");
	Write("(*NotebookOptionsPosition[     17786,        448]*)\n");
	Write("(*NotebookOutlinePosition[     18441,        471]*)\n");
	Write("(*  CellTagsIndexPosition[     18397,        467]*)\n");
	Write("(*WindowFrame->Normal*)\n\n\n");
	Write("Notebook[{\n");
	//Write("Cell[CellGroupData[{\n");
	m_idCurrCell=-1; m_nCells=0;
	// new cell will be started manually or automatically at first drawing call
}
void CVMOutputMath::Finalize() {
	FinalizeCell();
	// write nb-postfix
	Write("},\nFrontEndVersion->\"4.0 for Microsoft Windows\",\nScreenRectangle->{{0, 1280}, {0, 727}},\nWindowSize->{496, 631},\nWindowMargins->{{18, Automatic}, {Automatic, 23}}\n]\n");
	Write("\n(*NotebookFileOutline\nNotebook[{\n}\n]\n*)\n");
	CVMFileOutput::Finalize();
}
void CVMOutputMath::MoveTo(double x, double y) {
	PrepareCell();
	if(m_idCurrCell==0) { m_xCurr=x; m_yCurr=y; sprintf(m_TmpBuf,"%g %g m\n",x,y); Write(m_TmpBuf); }
}
void CVMOutputMath::LineTo(double x, double y) {
	PrepareCell();
	if(m_idCurrCell==0) { m_xCurr=x; m_yCurr=y; sprintf(m_TmpBuf,"%g %g L\n",x,y); Write(m_TmpBuf); }
}
void CVMOutputMath::Arc(double x, double y, double r, double phi1, double phi2) {
	PrepareCell();
	if(m_idCurrCell==0) {
		Write("s\n"); //Note: required 'stroke' command before, otherwise draws Line(Old pos-->New pos)
		sprintf(m_TmpBuf,"%g %g %g %g %g arc\n",x,y,r,phi1,phi2); Write(m_TmpBuf);
	}
}
void CVMOutputMath::TextOut(double x, double y, const char* text, int len) {
	PrepareCell();
	if(m_idCurrCell==0) {
		bool bRotate=(m_fTxtAngle!=0.f);
		if((!m_bUseMathTextMacros) || (bRotate) || (m_fBlockRotationAngle!=0.f) || (m_iFontFamily!=CVMOutput::FF_COURIER)){
			//TODO: copy implementation from CVMOutputEPS::TextOut() after it's finished
			if(bRotate)StartBlockRotation(x,y,m_fTxtAngle);
			//Note: text is aligned Left+Bottom below ==> fix x,y
			if(len<0)len=CVMUtil::strLen(text);
			if(m_TextAlignX!=-1){ x-=GetLetterWidth()/2.*(m_TextAlignX+1)*len; }
			if(m_TextAlignY!=-1){ y-=GetLetterHeight()/2.*(m_TextAlignY+1); }
			sprintf(m_TmpBuf,"%g %g m\n",x,y); Write(m_TmpBuf);
			//if(m_bFontFamilyChanged){ //reset font
			const char* cFontFamily=CVMOutput::GetFontFamilyID(m_iFontFamily);
			sprintf(m_TmpBuf,"/%s findfont %g scalefont\n",cFontFamily,GetFontSizeScaled()); Write(m_TmpBuf);
			sprintf(m_TmpBuf,"setfont\n"); Write(m_TmpBuf);
			if(m_bFontFamilyChanged)m_bFontFamilyChanged=false;
			//}
			if(len>0){ sprintf(m_TmpBuf,"("); memcpy(m_TmpBuf+1,text,len); sprintf(m_TmpBuf+1+len,") show\n"); }
			else{ sprintf(m_TmpBuf,"(%s) show\n",text); }
			Write(m_TmpBuf);
			if(bRotate)StopBlockRotation();
		}else{
			if(m_bFontFamilyChanged){ //reset font
				sprintf(m_TmpBuf,"/Courier findfont %g scalefont setfont\n",(double)GetFontSize()); Write(m_TmpBuf);
				m_bFontFamilyChanged=false; //m_bFontSizeChanged=false;
			}
			if(len>0){ sprintf(m_TmpBuf,"[("); memcpy(m_TmpBuf+2,text,len); sprintf(m_TmpBuf+2+len,")] %g %g %ld %ld Mshowa\n",x,y,m_TextAlignX,m_TextAlignY); }
			else{ sprintf(m_TmpBuf,"[(%s)] %g %g %ld %ld Mshowa\n",text,x,y,m_TextAlignX,m_TextAlignY); }
			Write(m_TmpBuf);
		}
	} else if(m_idCurrCell==1) {
		if(len>0){ memcpy(m_TmpBuf,text,len); *(m_TmpBuf+len)='\0'; Write(m_TmpBuf); }
		else{ Write(text); }
	}
}
void CVMOutputMath::StartBlockRotation(double x, double y, float fBlockAngle) {
	if(fBlockAngle!=0.f){
		CVMOutput::StartBlockRotation(x,y,fBlockAngle);
		sprintf(m_TmpBuf,"gsave %g %g translate %g rotate %g %g translate\n",x,y,(double)fBlockAngle,-x,-y); Write(m_TmpBuf);
	}
}
void CVMOutputMath::StopBlockRotation() {
	if(m_fBlockRotationAngle!=0.f){
		CVMOutput::StopBlockRotation();
		sprintf(m_TmpBuf,"grestore\n"); Write(m_TmpBuf);
	}
}
double CVMOutputMath::GetLetterWidth() { return 0.026*GetFontSize()*GetFontWidthFactor()/10; }
double CVMOutputMath::GetLetterHeight() { return 0.03*GetFontSize()/10; }
void CVMOutputMath::SetLineDash(int nElements, double* len) {
	PrepareCell();
	if(m_idCurrCell==0) {
		Write("s\n"); // show previous
		if(nElements>0){ if(len==NULL){ nElements=0; }else if(len[0]<0.){ nElements=0; } } //set line solid
		if(nElements>=2) {
			int lBuf=sprintf(m_TmpBuf,"[ ");
			for(int i=0;i<nElements;i++){ double lI=len[i]; if(lI<0.)lI=len[0]; lBuf+=sprintf(m_TmpBuf+lBuf,"%.2g ",lI); }
			lBuf+=sprintf(m_TmpBuf+lBuf,"] 0 setdash\n"); Write(m_TmpBuf);
		}else{ sprintf(m_TmpBuf,"[ ] 0 setdash\n"); Write(m_TmpBuf); }
		//
		//[Old]if (l1>=0) { if(l2<0)l2=l1; sprintf(m_TmpBuf,"[ %.2g %.2g ] 0 setdash\n",l1,l2); Write(m_TmpBuf); }
		//[Old]else { sprintf(m_TmpBuf,"[ ] 0 setdash\n"); Write(m_TmpBuf); }
	}
}
void CVMOutputMath::SetColor(CVMRGB rgb) {
	PrepareCell();
	if(m_idCurrCell==0) {
		Write("s\n"); // show previous
		if(m_clrDraw!=rgb){
			CVMOutput::SetColor(rgb);
			sprintf(m_TmpBuf,"%.4g %.4g %.4g r\n",(double)rgb.m_Red,(double)rgb.m_Green,(double)rgb.m_Blue); Write(m_TmpBuf);
		}
	}
}
void CVMOutputMath::FillColor(double x1, double y1, double x2, double y2, CVMRGB rgb) {
	PrepareCell();
	if(m_idCurrCell==0) {
		Write("s\n"); // show previous
		//set color
		sprintf(m_TmpBuf,"%.4g %.4g %.4g r\n",(double)rgb.m_Red,(double)rgb.m_Green,(double)rgb.m_Blue); Write(m_TmpBuf);
		sprintf(m_TmpBuf,"%.6g %.6g m\n%.6g %.6g L\n%.6g %.6g L\n%.6g %.6g L\nfill\n",x1,y1,x1,y2,x2,y2,x2,y1); Write(m_TmpBuf);
		//restore color
		sprintf(m_TmpBuf,"%.4g %.4g %.4g r\n",(double)m_clrDraw.m_Red,(double)m_clrDraw.m_Green,(double)m_clrDraw.m_Blue); Write(m_TmpBuf);
	}
}
void CVMOutputMath::FillContour(CVMRGB rgb) {
	if(m_idCurrCell==0) {
		sprintf(m_TmpBuf,"%.4g %.4g %.4g r\nfill\n",(double)rgb.m_Red,(double)rgb.m_Green,(double)rgb.m_Blue); Write(m_TmpBuf);
		//restore color
		sprintf(m_TmpBuf,"%.4g %.4g %.4g r\n",(double)m_clrDraw.m_Red,(double)m_clrDraw.m_Green,(double)m_clrDraw.m_Blue); Write(m_TmpBuf);
	}
}
void CVMOutputMath::BlockEnd() {
	if(m_idCurrCell<0)NewPlotCell();
	if(m_idCurrCell==0){ Write("s\n"); }
}
void CVMOutputMath::PrepareCell() { //create/update cell before output
	if(m_idCurrCell<0)NewPlotCell();

	if(m_bLineWidthChanged) { //Set new line width
		if(m_idCurrCell==0) {
			Write("s\n"); // show previous
			sprintf(m_TmpBuf,"%g Mabswid\n",m_dLineWidth); Write(m_TmpBuf); m_bLineWidthChanged=false;
		}
	}
	if(m_bFontSizeChanged) { //Set new font size
		if(m_idCurrCell==0) {
			Write("s\n"); // show previous
			sprintf(m_TmpBuf,"/Courier findfont %g scalefont setfont\n",m_FontSize); Write(m_TmpBuf); m_bFontSizeChanged=false;
		}
	}
}
void CVMOutputMath::NewPlotCell() {
	if(m_idCurrCell>=0)FinalizeCell();
	if (m_nCells>0) Write(",\n");

	// write graphics cell header
	Write("\nCell[GraphicsData[\"PostScript\", \"\\<\\\n");

	// write postscript header
	Write("%!\n%%Creator: VM\n%AspectRatio: "); sprintf(m_TmpBuf,"%g \n",m_aspect_ratio); Write(m_TmpBuf);
	Write("MathPictureStart\n/Mabs {\nMgmatrix idtransform\nMtmatrix dtransform\n} bind def\n/Mabsadd { Mabs\n3 -1 roll add\n3 1 roll add\nexch } bind def\n");
	Write("%% Graphics\n");
	//fonts used
	Write("%%IncludeResource: font Courier\n%%IncludeFont: Courier\n");
	Write("%%IncludeResource: font Symbol\n%%IncludeFont: Symbol\n");
	Write("/Courier findfont 10  scalefont  setfont\n");
	Write("% Scaling calculations\n");
	Write("0.461538 0.153846 0.0147151 0.0255915 [\n");
	Write("[.15385 .00222 -9 -9 ]\n[.15385 .00222 9 0 ]\n[.30769 .00222 -9 -9 ]\n[.30769 .00222 9 0 ]\n");
	Write("[.46154 .00222 -6 -9 ]\n[.46154 .00222 6 0 ]\n[.61538 .00222 -6 -9 ]\n[.61538 .00222 6 0 ]\n");
	Write("[.76923 .00222 -6 -9 ]\n[.76923 .00222 6 0 ]\n[.92308 .00222 -6 -9 ]\n[.92308 .00222 6 0 ]\n");
	Write("[-0.0125 .14267 -6 -4.5 ]\n[-0.0125 .14267 0 4.5 ]\n[-0.0125 .27063 -12 -4.5 ]\n[-0.0125 .27063 0 4.5 ]\n");
	Write("[-0.0125 .39859 -12 -4.5 ]\n[-0.0125 .39859 0 4.5 ]\n[-0.0125 .52654 -12 -4.5 ]\n[-0.0125 .52654 0 4.5 ]\n");
	Write("[ 0 0 0 0 ]\n");
	sprintf(m_TmpBuf,"[ 1 %.6f 0 0 ]\n",m_aspect_ratio); Write(m_TmpBuf);
	Write("] MathScale\n");
	Write("% Start of Graphics\n");
	Write("1 setlinecap\n1 setlinejoin\nnewpath\n");
	Write("0 g\n");
	sprintf(m_TmpBuf,"%g Mabswid\n",m_dLineWidth); Write(m_TmpBuf); m_bLineWidthChanged=false; //Set line width

	//Write("closepath\nclip\nnewpath\n0 g\n.5 Mabswid\n"); // initially was after exis drawing
	m_idCurrCell=0; m_nCells++;
}
void CVMOutputMath::NewTextCell() {
	if(m_idCurrCell>=0)FinalizeCell();
	if (m_nCells>0) Write(",\n");
	Write("\nCell[BoxData[\\(");
	m_idCurrCell=1; m_nCells++;
}
void CVMOutputMath::FinalizeCell() {
	if (m_idCurrCell==0) {
		Write("s\n"); // show previous
		// write postscript postfix
		Write("% End of Graphics\nMathPictureEnd\n");
		// write graphics cell postfix
		Write("\\\n\\>\"], \"Graphics\"");
		//Write(",\nImageSize->{288, 177.938},\nImageMargins->{{43, 0}, {0, 0}},\nImageRegion->{{0, 1}, {0, 1}}");
		Write("]\n");
	} else if (m_idCurrCell==1) {
		Write("\\)], \"Text\"]");
	}
	m_idCurrCell=-1;
}

/////////////////////////////////////////////////////////////////////// CVMOutputEPS
CVMOutputEPS::CVMOutputEPS(const char* fn) : CVMFileOutput(fn,true) {
	//m_xPadMin=91.5625; m_xPadMax=321.9375; m_yPadMin=3.1875; m_yPadMax=145.5; //Mathematica default (w:230.375, h:142.3125)
	m_xPadMin=0.; m_xPadMax=0.; m_yPadMin=0.; m_yPadMax=1.; //unused numbers
	SetBoundingBox(0.,230.375,0.); //SetBoundingBox(0.,230.375,0.,142.3125);
}
CVMOutputEPS::~CVMOutputEPS() {}
void CVMOutputEPS::GetBoundingBox(double* xBoxMin, double* xBoxMax, double* yBoxMin, double* yBoxMax) {
	if(xBoxMin!=NULL)*xBoxMin=m_xPadMin; if(xBoxMax!=NULL)*xBoxMax=m_xPadMax;
	if(yBoxMin!=NULL)*yBoxMin=m_yPadMin; if(yBoxMax!=NULL)*yBoxMax=m_yPadMax;
}
void CVMOutputEPS::SetBoundingBox(double xBoxMin, double xBoxMax, double yBoxMin) { //scaled to current aspect ratio
	if ((xBoxMin>=0) && (yBoxMin>=0) && (xBoxMax>xBoxMin)) {
		m_xPadMin=xBoxMin; m_xPadMax=xBoxMax; m_yPadMin=yBoxMin; m_yPadMax=m_yPadMin+(m_xPadMax-m_xPadMin)*m_aspect_ratio;
		m_fPad2LogicalCoordScale=1./(xBoxMax-xBoxMin);
	}
}
void CVMOutputEPS::SetBoundingBox(double xBoxMin, double xBoxMax, double yBoxMin, double yBoxMax) { //change aspect ratio
	if ((xBoxMin>=0) && (yBoxMin>=0) && (xBoxMax>xBoxMin) && (yBoxMax>yBoxMin)) {
		m_xPadMin=xBoxMin; m_xPadMax=xBoxMax; m_yPadMin=yBoxMin; m_yPadMax=yBoxMax;
		SetMargins(0.,1.,0.,1.*(yBoxMax-yBoxMin)/(xBoxMax-xBoxMin)); //Set 'logical' margins to fit new aspect ratio
		m_fPad2LogicalCoordScale=1./(xBoxMax-xBoxMin);
	}
}
void CVMOutputEPS::Start() {
	CVMFileOutput::Start();

	// write eps-prefix
	Write("%!PS-Adobe-2.0 EPSF-1.2\n");
	int ixPadMin=(int)m_xPadMin; int iyPadMin=(int)m_yPadMin; int ixPadMax=(int)m_xPadMax; int iyPadMax=(int)m_yPadMax;
	//Adjust 'truncations' to:
	//[Old] iMin<=fMin, iMax>=fMax ==> 'Empty lines' may appear at top/right sides in GSView.
	//if(ixPadMin>m_xPadMin)ixPadMin--; if(iyPadMin>m_yPadMin)iyPadMin--;
	//if(ixPadMax<m_xPadMax)ixPadMax++; if(iyPadMax<m_yPadMax)iyPadMax++;
	//[New] iMin>=fMin, iMax<=fMax
	if(ixPadMin<m_xPadMin)ixPadMin++; if(iyPadMin<m_yPadMin)iyPadMin++;
	if(ixPadMax>m_xPadMax)ixPadMax--; if(iyPadMax>m_yPadMax)iyPadMax--;
	if(ixPadMax<=ixPadMin)ixPadMax=ixPadMin+1; if(iyPadMax<=iyPadMin)iyPadMax=iyPadMin+1; //Fix possible 0-width/height
	//
	sprintf(m_TmpBuf,"%%%%BoundingBox: %d %d %d %d\n",ixPadMin,iyPadMin,ixPadMax,iyPadMax); Write(m_TmpBuf);
	sprintf(m_TmpBuf,"%%%%HiResBoundingBox: %.6g %.6g %.6g %.6g\n",m_xPadMin,m_yPadMin,m_xPadMax,m_yPadMax); Write(m_TmpBuf);
	Write("%%Creator: CVMHistLiteView by V.M.\n");
	if(CVMOutputSettings::m_bSaveOutputTime) {
		time_t t_curr; time(&t_curr); tm tm_curr=*localtime(&t_curr);
		char* wDays[7]={"Sun","Mon","Tue","Wen","Thu","Fri","Sat"};
		char* wMons[12]={"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
		sprintf(m_TmpBuf,"%%%%CreationDate: %s %s %02d %02d:%02d:%02d %d\n",wDays[tm_curr.tm_wday],wMons[tm_curr.tm_mon],
			tm_curr.tm_mday,tm_curr.tm_hour,tm_curr.tm_min,tm_curr.tm_sec,tm_curr.tm_year+1900); Write(m_TmpBuf);
	}
	Write("%%Title: CVMHistLiteView plot\n");
	Write("%%DocumentNeededResources: font Courier\n");
	Write("%%DocumentSuppliedResources:\n");
	Write("%%DocumentNeededFonts: Courier\n");
	Write("%%DocumentSuppliedFonts:\n");
	Write("%%DocumentFonts: Courier\n");
	Write("%%EndComments\n");
	Write("/MISOfy\n{\n");
	Write("\t/newfontname exch def\n\t/oldfontname exch def\n\n\toldfontname findfont\n");
	Write("\tdup length dict begin\n");
	Write("\t\t{1 index /FID ne {def} {pop pop} ifelse} forall\n");
	Write("\t\t/Encoding ISOLatin1Encoding def\n");
	Write("\t\tcurrentdict\n");
	Write("\tend\n\n");
	Write("\tnewfontname exch definefont pop\n");
	Write("} def\n\n");
	sprintf(m_TmpBuf,"0 %.6g translate 1 -1 scale\n",m_yPadMin+m_yPadMax); Write(m_TmpBuf); //Why yMix+yMax? Mathematica bug?
	Write("gsave\n150 dict begin\n/Mfixwid true def\n/Mrot 0 def\n");
	Write("/Mpstart {\n    MathPictureStart\n} bind def\n");
	Write("/Mpend {\n    MathPictureEnd\n} bind def\n");
	Write("/Mscale {\n    0 1 0 1\n    5 -1 roll\n    MathScale\n} bind def\n");
	Write("/Plain\t/Courier findfont def\n");
	Write("/Bold\t/Courier-Bold findfont def\n");
	Write("/Italic /Courier-Oblique findfont def\n");
	Write("/MathPictureStart {\n\t/Mimatrix\n\t matrix currentmatrix\n\tdef\n\tgsave\n\tnewpath\n\tMleft\n\tMbottom\n\ttranslate\n\t/Mtmatrix\n\tmatrix currentmatrix\n\tdef\n\tPlain\n\tMfontsize scalefont\n\tsetfont\n\t0 setgray\n\t0 setlinewidth\n} bind def\n");
	Write("/MathPictureEnd {\n\tgrestore\n} bind def\n");
	Write("/MathSubStart {\n\tMomatrix\n\tMgmatrix Mtmatrix\n\tMleft Mbottom\n\tMwidth Mheight\n\t9 -2 roll\n\tmoveto\n\tMtmatrix setmatrix\n\tcurrentpoint\n\tMgmatrix setmatrix\n\t11 -2 roll\n");
	Write("\tmoveto\n\tMtmatrix setmatrix\n\tcurrentpoint\n\t2 copy translate\n\t/Mtmatrix matrix currentmatrix def\n\t/Mleft 0 def\n\t/Mbottom 0 def\n\t3 -1 roll\n\texch sub\n\t/Mheight exch def\n\tsub\n\t/Mwidth exch def\n} bind def\n");
	Write("/MathSubEnd {\n\t/Mheight exch def\n\t/Mwidth exch def\n\t/Mbottom exch def\n\t/Mleft exch def\n\t/Mtmatrix exch def\n\tdup setmatrix\n\t/Mgmatrix exch def\n\t/Momatrix exch def\n} bind def\n");
	Write("/Mdot {\n\tmoveto\n\t0 0 rlineto\n\tstroke\n} bind def\n");
	Write("/Mtetra {\n\tmoveto\n\tlineto\n\tlineto\n\tlineto\n\tfill\n} bind def\n");
	Write("/Metetra {\n\tmoveto\n\tlineto\n\tlineto\n\tlineto\n\tclosepath\n\tgsave\n\tfill\n\tgrestore\n\t0 setgray\n\tstroke\n} bind def\n");
	Write("/Mistroke {\n\tflattenpath\n\t0 0 0\n\t{\n\t4 2 roll\n\tpop pop\n\t}\n\t{\n\t4 -1 roll\n\t2 index\n\tsub dup mul\n\t4 -1 roll\n\t2 index\n\tsub dup mul\n\tadd sqrt\n\t4 -1 roll\n\tadd\n\t3 1 roll\n");
	Write("\t}\n\t{\n\tstop\n\t}\n\t{\n\tstop\n\t}\n\tpathforall\n\tpop pop\n\tcurrentpoint\n\tstroke\n\tmoveto\n\tcurrentdash\n\t3 -1 roll\n\tadd\n\tsetdash\n} bind def\n");
	Write("/Mfstroke {\n\tstroke\n\tcurrentdash\n\tpop 0\n\tsetdash\n} bind def\n");
	Write("/Mrotsboxa {\n\tgsave\n\tdup\n\t/Mrot\n\texch def\n\tMrotcheck\n\tMtmatrix\n\tdup\n\tsetmatrix\n\t7 1 roll\n\t4 index\n\t4 index\n");
	Write("\ttranslate\n\trotate\n\t3 index\n\t-1 mul\n\t3 index\n\t-1 mul\n\ttranslate\n\t/Mtmatrix\n\tmatrix\n\tcurrentmatrix\n\tdef\n\tgrestore\n\tMsboxa\n\t3  -1 roll\n\t/Mtmatrix\n\texch def\n\t/Mrot\n\t0 def\n} bind def\n");
	Write("/Msboxa {\n\tnewpath\n\t5 -1 roll\n\tMvboxa\n\tpop\n\tMboxout\n\t6 -1 roll\n\t5 -1 roll\n\t4 -1 roll\n\tMsboxa1\n\t5 -3 roll\n\tMsboxa1\n\tMboxrot\n\t[\n\t7 -2 roll\n\t2 copy\n\t[\n\t3 1 roll\n\t10 -1 roll\n\t9 -1 roll\n\t]\n\t6 1 roll\n\t5 -2 roll\n\t]\n} bind def\n");
	Write("/Msboxa1 {\n\tsub\n\t2 div\n\tdup\n\t2 index\n\t1 add\n\tmul\n\t3 -1 roll\n\t-1 add\n\t3 -1 roll\n\tmul\n} bind def\n");
	Write("/Mvboxa {\n\tMfixwid\n\t{\n\tMvboxa1\n\t}\n\t{\n\tdup\n\tMwidthcal\n\t0 exch\n\t{\n\tadd\n\t}\n\tforall\n\texch\n\tMvboxa1\n\t4 index\n\t7 -1 roll\n\tadd\n\t4 -1 roll\n\tpop\n\t3 1 roll\n\t}\n\tifelse\n} bind def\n");
	Write("/Mvboxa1 {\n\tgsave\n\tnewpath\n\t[ true\n\t3 -1 roll\n\t{\n\tMbbox\n\t5 -1 roll\n\t{\n\t0\n\t5 1 roll\n\t}\n\t{\n\t7 -1 roll\n\texch sub\n\t(m) stringwidth pop\n\t.3 mul\n\tsub\n\t7 1 roll\n\t6 -1 roll\n\t4 -1 roll\n");
	Write("\tMmin\n\t3 -1 roll\n\t5 index\n\tadd\n\t5 -1 roll\n\t4 -1 roll\n\tMmax\n\t4 -1 roll\n\t}\n\tifelse\n\tfalse\n\t}\n\tforall\n\t{ stop } if\n\tcounttomark\n\t1 add\n\t4 roll\n\t]\n\tgrestore\n} bind def\n");
	Write("/Mbbox {\n\t0 0 moveto\n\tfalse charpath\n\tflattenpath\n\tpathbbox\n\tnewpath\n} bind def\n");
	Write("/Mmin {\n\t2 copy\n\tgt\n\t{ exch } if\n\tpop\n} bind def\n");
	Write("/Mmax {\n\t2 copy\n\tlt\n\t{ exch } if\n\tpop\n} bind def\n");
	Write("/Mrotshowa {\n\tdup\n\t/Mrot\n\texch def\n\tMrotcheck\n\tMtmatrix\n\tdup\n\tsetmatrix\n\t7 1 roll\n\t4 index\n\t4 index\n\ttranslate\n\trotate\n\t3 index\n\t-1 mul\n\t3 index\n\t-1 mul\n\ttranslate\n\t/Mtmatrix\n\tmatrix\n\tcurrentmatrix\n\tdef\n\tMgmatrix setmatrix\n\tMshowa\n\t/Mtmatrix\n\texch def\n\t/Mrot 0 def\n} bind def\n");
	Write("/Mshowa {\n\t4 -2 roll\n\tmoveto\n\t2 index\n\tMtmatrix setmatrix\n\tMvboxa\n\t7 1 roll\n\tMboxout\n\t6 -1 roll\n\t5 -1 roll\n\t4 -1 roll\n\tMshowa1\n\t4 1 roll\n\tMshowa1\n\trmoveto\n\tcurrentpoint\n\tMfixwid\n\t{\n\tMshowax\n\t}\n\t{\n\tMshoway\n\t}\n\tifelse\n\tpop pop pop pop\n\tMgmatrix setmatrix\n} bind def\n");
	Write("/Mshowax {\n\t0 1\n\t4 index length\n\t-1 add\n\t{\n\t2 index\n\t4 index\n\t2 index\n\tget\n\t3 index\n\tadd\n\tmoveto\n\t4 index\n\texch get\n\tshow\n\t} for\n} bind def\n");
	Write("/Mshoway {\n\t3 index\n\tMwidthcal\n\t5 1 roll\n\t0 1\n\t4 index length\n\t-1 add\n\t{\n\t2 index\n\t4 index\n\t2 index\n\tget\n\t3 index\n\tadd\n\tmoveto\n\t4 index\n\texch get\n");
	Write("\t[\n\t6 index\n\taload\n\tlength\n\t2 add\n\t-1 roll\n\t{\n\tpop\n\tStrform\n\tstringwidth\n\tpop\n\tneg\n\texch\n\tadd\n\t0 rmoveto\n\t}\n\texch\n\tkshow\n\tcleartomark\n\t} for\n\tpop\n} bind def\n");
	Write("/Mwidthcal {\n\t[\n\texch\n\t{\n\tMwidthcal1\n\t}\n\tforall\n\t]\n\t[\n\texch\n\tdup\n\tMaxlen\n\t-1 add\n\t0 1\n\t3 -1 roll\n\t{\n\t[\n\texch\n\t2 index\n\t{\n\t1 index\n\tMget\n\texch\n\t}\n\tforall\n\tpop\n\tMaxget\n\texch\n\t}\n\tfor\n\tpop\n\t]\n\tMreva\n} bind def\n");
	Write("/Mreva {\n\t[\n\texch\n\taload\n\tlength\n\t-1 1\n\t{1 roll}\n\tfor\n\t]\n} bind def\n");
	Write("/Mget {\n\t1 index\n\tlength\n\t-1 add\n\t1 index\n\tge\n\t{\n\tget\n\t}\n\t{\n\tpop pop\n\t0\n\t}\n\tifelse\n} bind def\n");
	Write("/Maxlen {\n\t[\n\texch\n\t{\n\tlength\n\t}\n\tforall\n\tMaxget\n} bind def\n");
	Write("/Maxget {\n\tcounttomark\n\t-1 add\n\t1 1\n\t3 -1 roll\n\t{\n\tpop\n\tMmax\n\t}\n\tfor\n\texch\n\tpop\n} bind def\n");
	Write("/Mwidthcal1 {\n\t[\n\texch\n\t{\n\tStrform\n\tstringwidth\n\tpop\n\t}\n\tforall\n\t]\n} bind def\n");
	Write("/Strform {\n\t/tem (x) def\n\ttem 0\n\t3 -1 roll\n\tput\n\ttem\n} bind def\n");
	Write("/Mshowa1 {\n\t2 copy\n\tadd\n\t4 1 roll\n\tsub\n\tmul\n\tsub\n\t-2 div\n} bind def\n");
	Write("/MathScale {\n\tMwidth\n\tMheight\n\tMlp\n\ttranslate\n\tscale\n\t/yscale exch def\n\t/ybias exch def\n\t/xscale exch def\n\t/xbias exch def\n\t/Momatrix\n\txscale yscale matrix scale\n\txbias ybias matrix translate\n\tmatrix concatmatrix def\n\t/Mgmatrix\n\tmatrix currentmatrix\n\tdef\n} bind def\n");
	Write("/Mlp {\n\t3 copy\n\tMlpfirst\n\t{\n\tMnodistort\n\t{\n\tMmin\n\tdup\n\t} if\n\t4 index\n\t2 index\n\t2 index\n\tMlprun\n\t11 index\n\t11 -1 roll\n\t10 -4 roll\n\tMlp1\n\t8 index\n\t9 -5 roll\n\tMlp1\n\t4 -1 roll\n\tand\n\t{ exit } if\n\t3 -1 roll\n\tpop pop\n\t} loop\n\texch\n\t3 1 roll\n\t7 -3 roll\n\tpop pop pop\n} bind def\n");
	Write("/Mlpfirst {\n\t3 -1 roll\n\tdup length\n\t2 copy\n\t-2 add\n\tget\n\taload\n\tpop pop pop\n\t4 -2 roll\n\t-1 add\n\tget\n\taload\n\tpop pop pop\n\t6 -1 roll\n\t3 -1 roll\n\t5 -1 roll\n\tsub\n\tdup /MsaveAx exch def\n\tdiv\n\t4 1 roll\n\texch sub\n\tdup /MsaveAy exch def\n\tdiv\n} bind def\n");
	Write("/Mlprun {\n\t2 copy\n\t4 index\n\t0 get\n\tdup\n\t4 1 roll\n\tMlprun1\n\t3 copy\n\t8 -2 roll\n\t9 -1 roll\n\t{\n\t3 copy\n\tMlprun1\n\t3 copy\n\t11 -3 roll\n\t/gt Mlpminmax\n\t8 3 roll\n\t11 -3 roll\n");
	Write("\t/lt Mlpminmax\n\t8 3 roll\n\t} forall\n\tpop pop pop pop\n\t3 1 roll\n\tpop pop\n\taload pop\n\t5 -1 roll\n\taload pop\n\texch\n\t6 -1 roll\n\tMlprun2\n\t8 2 roll\n\t4 -1 roll\n\tMlprun2\n\t6 2 roll\n\t3 -1 roll\n\tMlprun2\n\t4 2 roll\n\texch\n\tMlprun2\n\t6 2 roll\n} bind def\n");
	Write("/Mlprun1 {\n\taload pop\n\texch\n\t6 -1 roll\n\t5 -1 roll\n\tmul add\n\t4 -2 roll\n\tmul\n\t3 -1 roll\n\tadd\n} bind def\n");
	Write("/Mlprun2 {\n\t2 copy\n\tadd 2 div\n\t3 1 roll\n\texch sub\n} bind def\n");
	Write("/Mlpminmax {\n\tcvx\n\t2 index\n\t6 index\n\t2 index\n\texec\n\t{\n\t7 -3 roll\n\t4 -1 roll\n\t} if\n\t1 index\n\t5 index\n\t3 -1 roll\n\texec\n\t{\n\t4 1 roll\n");
	Write("\tpop\n\t5 -1 roll\n\taload\n\tpop pop\n\t4 -1 roll\n\taload pop\n\t[\n\t8 -2 roll\n\tpop\n\t5 -2 roll\n\tpop\n\t6 -2 roll\n\tpop\n\t5 -1 roll\n\t]\n\t4 1 roll\n\tpop\n\t}\n\t{\n\tpop pop pop\n\t} ifelse\n} bind def\n");
	Write("/Mlp1 {\n\t5 index\n\t3 index\tsub\n\t5 index\n\t2 index mul\n\t1 index\n\tle\n\t1 index\n\t0 le\n\tor\n\tdup\n\tnot\n");
	Write("\t{\n\t1 index\n\t3 index\tdiv\n\t.99999 mul\n\t8 -1 roll\n\tpop\n\t7 1 roll\n\t}\n\tif\n\t8 -1 roll\n\t2 div\n\t7 -2 roll\n\tpop sub\n\t5 index\n\t6 -3 roll\n\tpop pop\n\tmul sub\n\texch\n");
	Write("} bind def\n/intop 0 def\n/inrht 0 def\n/inflag 0 def\n/outflag 0 def\n/xadrht 0 def\n/xadlft 0 def\n/yadtop 0 def\n/yadbot 0 def\n/Minner {\n\toutflag\n\t1\n\teq\n\t{\n");
	Write("\t/outflag 0 def\n\t/intop 0 def\n\t/inrht 0 def\n\t} if\n\t5 index\n\tgsave\n\tMtmatrix setmatrix\n\tMvboxa pop\n\tgrestore\n\t3 -1 roll\n\tpop\n\tdup\n\tintop\n\tgt\n\t{\n\t/intop\n\texch def\n\t}\n\t{ pop }\n\tifelse\n\tdup\n\tinrht\n\tgt\n\t{\n\t/inrht\n\texch def\n\t}\n\t{ pop }\n\tifelse\n\tpop\n\t/inflag\n\t1 def\n} bind def\n");
	Write("/Mouter {\n\t/xadrht 0 def\n\t/xadlft 0 def\n\t/yadtop 0 def\n\t/yadbot 0 def\n\tinflag\n\t1 eq\n\t{\n\tdup\n\t0 lt\n\t{\n\tdup\n\tintop\n\tmul\n\tneg\n\t/yadtop\n\texch def\n\t} if\n\tdup\n\t0 gt\n\t{\n\tdup\n\tintop\n\tmul\n\t/yadbot\n\texch def\n\t}\n\tif\n\tpop\n\tdup\n");
	Write("\t0 lt\n\t{\n\tdup\n\tinrht\n\tmul\n\tneg\n\t/xadrht\n\texch def\n\t} if\n\tdup\n\t0 gt\n\t{\n\tdup\n\tinrht\n\tmul\n\t/xadlft\n\texch def\n\t} if\n\tpop\n\t/outflag 1 def\n\t}\n\t{ pop pop}\n\tifelse\n\t/inflag 0 def\n\t/inrht 0 def\n\t/intop 0 def\n} bind def\n");
	Write("/Mboxout {\n\toutflag\n\t1\n\teq\n\t{\n\t4 -1\n\troll\n\txadlft\n\tleadjust\n\tadd\n\tsub\n\t4 1 roll\n\t3 -1\n\troll\n\tyadbot\n");
	Write("\tleadjust\n\tadd\n\tsub\n\t3 1\n\troll\n\texch\n\txadrht\n\tleadjust\n\tadd\n\tadd\n\texch\n\tyadtop\n\tleadjust\n\tadd\n\tadd\n\t/outflag 0 def\n\t/xadlft 0 def\n\t/yadbot 0 def\n\t/xadrht 0 def\n\t/yadtop 0 def\n\t} if\n} bind def\n");
	Write("/leadjust {\n\t(m) stringwidth pop\n\t.5 mul\n} bind def\n");
	Write("/Mrotcheck {\n\tdup\n\t90\n\teq\n\t{\n\tyadbot\n\t/yadbot\n\txadrht\n\tdef\n\t/xadrht\n\tyadtop\n\tdef\n\t/yadtop\n\txadlft\n\tdef\n\t/xadlft\n\texch\n\tdef\n\t}\n\tif\n\tdup\n\tcos\n\t1 index\n\tsin\n\tCheckaux\n\tdup\n\tcos\n\t1 index\n\tsin neg\n\texch\n\tCheckaux\n\t3 1 roll\n\tpop pop\n} bind def\n");
	Write("/Checkaux {\n\t4 index\n\texch\n\t4 index\n\tmul\n\t3 1 roll\n\tmul add\n\t4 1 roll\n} bind def\n");
	Write("/Mboxrot {\n\tMrot\n\t90 eq\n\t{\n\tbrotaux\n\t4 2\n\troll\n\t}\n\tif\n\tMrot\n\t180 eq\n\t{\n\t4 2\n\troll\n\tbrotaux\n\t4 2\n\troll\n\tbrotaux\n\t}\n\tif\n\tMrot\n\t270 eq\n\t{\n\t4 2\n\troll\n\tbrotaux\n\t}\n\tif\n} bind def\n");
	Write("/brotaux {\n\tneg\n\texch\n\tneg\n} bind def\n");
	Write("/Mabsproc {\n\t0\n\tmatrix defaultmatrix\n\tdtransform idtransform\n\tdup mul exch\n\tdup mul\n\tadd sqrt\n} bind def\n");
	Write("/Mabswid {\n\tMabsproc\n\tsetlinewidth\n} bind def\n");
	Write("/Mabsdash {\n\texch\n\t[\n\texch\n\t{\n\tMabsproc\n\t}\n\tforall\n\t]\n\texch\n\tsetdash\n} bind def\n");
	Write("/MBeginOrig { Momatrix concat} bind def\n");
	Write("/MEndOrig { Mgmatrix setmatrix} bind def\n");
	Write("/sampledsound where\n{ pop}\n{ /sampledsound {\nexch\npop\nexch\n5 1 roll\nmul\n4 idiv\nmul\n2 idiv\nexch pop\nexch\n/Mtempproc exch def\n{ Mtempproc pop }\nrepeat\n} bind def\n} ifelse\n");
	Write("% Here are the short operators\n/g { setgray} bind def\n/k { setcmykcolor} bind def\n/m { moveto} bind def\n/p { gsave} bind def\n/r { setrgbcolor} bind def\n/w { setlinewidth} bind def\n/C { curveto} bind def\n/F { fill} bind def\n/L { lineto} bind def\n/P { grestore} bind def\n/s { stroke} bind def\n\n");
	//Mfill: unused
	Write("/MFill {\n\t0 0 \t\tmoveto\n\tMwidth 0 \tlineto\n\tMwidth Mheight \tlineto\n\t0 Mheight \tlineto\n\tfill\n} bind def\n\n");
	Write("/MPlotRegion {\n\t3 index\n\tMwidth mul\n\t2 index\n\tMheight mul\n\ttranslate\n\texch sub\n\tMheight mul\n\t/Mheight\n\texch def\n\texch sub\n\tMwidth mul\n\t/Mwidth\n\texch def\n} bind def\n\n");
	Write("/Mcharproc\n{\n\tcurrentfile\n\t(x)\n\treadhexstring\n\tpop\n\t0 get\n\texch\n\tdiv\n} bind def\n\n");
	Write("/Mshadeproc\n{\n\tdup\n\t3 1\n\troll\n\t{\n\tdup\n\tMcharproc\n\t3 1\n\troll\n\t} repeat\n\t1 eq\n\t{\n\tsetgray\n\t}\n\t{\n\t3 eq\n\t{\n\tsetrgbcolor\n\t}\n\t{\n\tsetcmykcolor\n\t} ifelse\n\t} ifelse\n} bind def\n\n");
	Write("/Mrectproc\n{\n\t3 index\n\t2 index\n\tmoveto\n\t2 index\n\t3 -1\n\troll\n\tlineto\n\tdup\n\t3 1\n\troll\n\tlineto\n\tlineto\n\tfill\n} bind def\n\n");
	Write("/_Mcolorimage\n{\n\t7 1\n\troll\n\tpop\n\tpop\n\tmatrix\n\tinvertmatrix\n\tconcat\n\t2 exch exp\n\t1 sub\n\t3 1 roll\n\t1 1\n\t2 index\n\t{\n\t1 1\n\t4 index\n\t{\n\tdup\n\t1 sub\n\texch\n\t2 index\n\tdup\n\t1 sub\n\texch\n\t7 index\n\t9 index\n\tMshadeproc\n\tMrectproc\n\t} for\n\tpop\n\t} for\n\tpop pop pop pop\n} bind def\n\n");
	Write("/_Mimage\n{\n\tpop\n\tmatrix\n\tinvertmatrix\n\tconcat\n\t2 exch exp\n\t1 sub\n\t3 1 roll\n\t1 1\n\t2 index\n\t{\n\t1 1\n\t4 index\n\t{\n\tdup\n\t1 sub\n\texch\n\t2 index\n\tdup\n\t1 sub\n\texch\n\t7 index\n\tMcharproc\n\tsetgray\n\tMrectproc\n\t} for\n\tpop\n\t} for\n\tpop pop pop\n} bind def\n\n");
	Write("/Mimage {\n\t4 index\n\t4 index\n\tmul 1600 gt\n\t{ image }\n\t{ _Mimage }\n\tifelse\n} def\n\n");
	Write("/Mcolorimage {\n\t6 index\n\t6 index\n\tmul 1600 gt\n\t{ colorimage }\n\t{ _Mcolorimage }\n\tifelse\n} def\n");
	Write("/Mnodistort true def\n");
	//
	Write("1.000000 1.000000 scale\n"); //Do nothing
	sprintf(m_TmpBuf,"%.6f %.6f translate\n",m_xPadMin,m_yPadMax); Write(m_TmpBuf); //Shift plot to y->+300
	Write("1.000000 -1.000000 scale\n"); //Inverse y-axis
	Write("0.000000 0.000000 translate\n"); //Do nothing
	//Note: are 'do-nothing' lines required by 'eps' standard? Or 'scale-translate' sequence is not obligatory?
	//
	Write("/Mleft\t\t0.000000 def\n/Mbottom\t0.000000 def\n");
	sprintf(m_TmpBuf,"/Mwidth\t\t%.6f def\n/Mheight\t%.6f def\n",m_xPadMax-m_xPadMin,m_yPadMax-m_yPadMin); Write(m_TmpBuf);
	Write("0 setgray\n0 setlinewidth\n/Courier findfont 12 scalefont setfont\n/Mfontsize 12 def\n/Plain /Courier findfont def\n\n");

	// write postscript header
	Write("%!\n%%Creator: VM\n%AspectRatio: "); sprintf(m_TmpBuf,"%g \n",m_aspect_ratio); Write(m_TmpBuf);
	Write("MathPictureStart\n/Mabs {\nMgmatrix idtransform\nMtmatrix dtransform\n} bind def\n/Mabsadd { Mabs\n3 -1 roll add\n3 1 roll add\nexch } bind def\n");
	//Actual font definition
	Write("%% Graphics\n");
	//fonts used
	Write("%%IncludeResource: font Courier\n%%IncludeFont: Courier\n");
	Write("%%IncludeResource: font Symbol\n%%IncludeFont: Symbol\n");
	Write("/Courier findfont 10  scalefont  setfont\n");
	//Write("/Arial findfont 10  scalefont  setfont\n");
	//
	Write("% Scaling calculations\n");
	Write("0.461538 0.153846 0.0147151 0.0255915 [\n");
	//Note: commented lines leads to additional small shrink/shift of the view!
	//Write("[.15385 .00222 -9 -9 ]\n[.15385 .00222 9 0 ]\n[.30769 .00222 -9 -9 ]\n[.30769 .00222 9 0 ]\n");
	//Write("[.46154 .00222 -6 -9 ]\n[.46154 .00222 6 0 ]\n[.61538 .00222 -6 -9 ]\n[.61538 .00222 6 0 ]\n");
	//Write("[.76923 .00222 -6 -9 ]\n[.76923 .00222 6 0 ]\n[.92308 .00222 -6 -9 ]\n[.92308 .00222 6 0 ]\n");
	//Write("[-0.0125 .14267 -6 -4.5 ]\n[-0.0125 .14267 0 4.5 ]\n[-0.0125 .27063 -12 -4.5 ]\n[-0.0125 .27063 0 4.5 ]\n");
	//Write("[-0.0125 .39859 -12 -4.5 ]\n[-0.0125 .39859 0 4.5 ]\n[-0.0125 .52654 -12 -4.5 ]\n[-0.0125 .52654 0 4.5 ]\n");
	Write("[ 0 0 0 0 ]\n"); //Last 2 values->common {x,y} shift
	sprintf(m_TmpBuf,"[ 1 %.6f 0 0 ]\n",m_aspect_ratio); Write(m_TmpBuf); //Last 2 values->"strange" {x,y}shift-to-border+scale
	Write("] MathScale\n");
	Write("% Start of Graphics\n");
	Write("1 setlinecap\n1 setlinejoin\nnewpath\n");
	Write("0 g\n");
	sprintf(m_TmpBuf,"%g Mabswid\n",m_dLineWidth); Write(m_TmpBuf); m_bLineWidthChanged=false; //Set line width
}
void CVMOutputEPS::Finalize() {
	Write("s\n"); // show previous
	// write postscript postfix
	Write("% End of Graphics\nMathPictureEnd\n");

	// write eps-postfix + end-of-line
	Write("\n%%PSTrailer\nend\ngrestore\n\n%%Trailer\n%%EOF\n");
	CVMFileOutput::Finalize();
}
void CVMOutputEPS::MoveTo(double x, double y) {
	m_xCurr=x; m_yCurr=y; sprintf(m_TmpBuf,"%g %g m\n",x,y); Write(m_TmpBuf);
}
void CVMOutputEPS::LineTo(double x, double y) {
	m_xCurr=x; m_yCurr=y; sprintf(m_TmpBuf,"%g %g L\n",x,y); Write(m_TmpBuf);
}
void CVMOutputEPS::Arc(double x, double y, double r, double phi1, double phi2) {
	Write("s\n"); //Note: required 'stroke' command before, otherwise draws Line(Old pos-->New pos)
	sprintf(m_TmpBuf,"%g %g %g %g %g arc\n",x,y,r,phi1,phi2); Write(m_TmpBuf);
}
/* //Ellips:
matrix currentmatrix
0.47619 0.238095 scale
1.05 1.06709 1 0 365.73 arc
setmatrix s 
*/
void CVMOutputEPS::TextOut(double x, double y, const char* text, int len) {
/*
*	PostScript-native text output [http://local.wasp.uwa.edu.au/~pbourke/dataformats/postscript/]:
/Times-Roman findfont
12 scalefont
setfont
newpath
100 200 moveto
(Example 3) show
---- more options:
/Times-Roman findfont
32 scalefont
setfont
100 200 translate
45 rotate
2 1 scale
newpath
0 0 moveto
(Example 4) true charpath
0.5 setlinewidth
0.4 setgray
stroke
---- but for 0..1 scaling is better something like:
/Times-Roman findfont
0.1 scalefont
setfont
0 0 translate
45 rotate
1 1 scale
newpath			<--- ?
0 0 moveto
(Example 4) true charpath
0.005 setlinewidth
0.4 setgray
stroke
*/
	//
	bool bRotate=(m_fTxtAngle!=0.f);
	if((!m_bUseMathTextMacros) || (bRotate) || (m_fBlockRotationAngle!=0.f) || (m_iFontFamily!=CVMOutput::FF_COURIER)){
		if(bRotate)StartBlockRotation(x,y,m_fTxtAngle);
		//Note: text is aligned Left+Bottom below ==> fix x,y
		if(len<0)len=CVMUtil::strLen(text);
		if(m_TextAlignX!=-1){ x-=GetLetterWidth()/2.*(m_TextAlignX+1)*len; }
		if(m_TextAlignY!=-1){ y-=GetLetterHeight()/2.*(m_TextAlignY+1); }
		sprintf(m_TmpBuf,"%g %g m\n",x,y); Write(m_TmpBuf);
		//Options before 'setfont'
		//[wFactor distorsion1 distorsion2 hFactor xShift yShift ] makefont, default: [1 0 0 1 0 0] makefont
		//[Into header] sprintf(m_TmpBuf,"%%%%IncludeResource: font Courier\n%%%%IncludeFont: Courier\n"); Write(m_TmpBuf);
		//if(m_bFontFamilyChanged){ //reset font
		const char* cFontFamily=CVMOutput::GetFontFamilyID(m_iFontFamily);
		sprintf(m_TmpBuf,"/%s findfont %g scalefont\n",cFontFamily,GetFontSizeScaled()); Write(m_TmpBuf);
		sprintf(m_TmpBuf,"setfont\n"); Write(m_TmpBuf);
		if(m_bFontFamilyChanged)m_bFontFamilyChanged=false;
		//}
		//Options before 'stroke(s)', before or after '(Text) true charpath (or 'false charpath')':
		//0.4 setgray
		//0.010 setlinewidth (width of contour line)
		//0.000 1.000 0.000 setrgbcolor
		//Options not working in case of (Text) show
		if(len>0){ sprintf(m_TmpBuf,"("); memcpy(m_TmpBuf+1,text,len); sprintf(m_TmpBuf+1+len,") show\n"); }
		else{ sprintf(m_TmpBuf,"(%s) show\n",text); }
		Write(m_TmpBuf); //,m_TextAlignX,m_TextAlignY --?
		if(bRotate)StopBlockRotation();
	}else{
		if(m_bFontFamilyChanged){ //reset font
			sprintf(m_TmpBuf,"/Courier findfont %g scalefont setfont\n",(double)GetFontSize()); Write(m_TmpBuf);
			m_bFontFamilyChanged=false; //m_bFontSizeChanged=false;
		}
		if(len>0){ sprintf(m_TmpBuf,"[("); memcpy(m_TmpBuf+2,text,len); sprintf(m_TmpBuf+2+len,")] %g %g %ld %ld Mshowa\n",x,y,m_TextAlignX,m_TextAlignY); }
		else{ sprintf(m_TmpBuf,"[(%s)] %g %g %ld %ld Mshowa\n",text,x,y,m_TextAlignX,m_TextAlignY); }
		Write(m_TmpBuf);
	}
}
void CVMOutputEPS::StartBlockRotation(double x, double y, float fBlockAngle) {
	if(fBlockAngle!=0.f){
		CVMOutput::StartBlockRotation(x,y,fBlockAngle);
		sprintf(m_TmpBuf,"gsave %g %g translate %g rotate %g %g translate\n",x,y,(double)fBlockAngle,-x,-y); Write(m_TmpBuf);
	}
}
void CVMOutputEPS::StopBlockRotation() {
	if(m_fBlockRotationAngle!=0.f){
		CVMOutput::StopBlockRotation();
		sprintf(m_TmpBuf,"grestore\n"); Write(m_TmpBuf);
	}
}
double CVMOutputEPS::GetLetterWidth() {
	double valAtFont10=((m_xPadMax>m_xPadMin)?0.026*230.375/(m_xPadMax-m_xPadMin)*(m_xMax-m_xMin):0); //original width=230.375
	double valAtFontX=valAtFont10*GetFontSize()/10;
	return valAtFontX*GetFontWidthFactor();
}
double CVMOutputEPS::GetLetterHeight() {
	double valAtFont10=((m_yPadMax-m_yPadMin)?0.03*142.3125/(m_yPadMax-m_yPadMin)*((m_yMax-m_yMin)/0.61803):0); //original height=142.3125
	return valAtFont10*GetFontSize()/10;
}
float CVMOutputEPS::SetFontSize(float sz) {
	float res=CVMOutput::SetFontSize(sz);
	if(m_bFontSizeChanged) {
		Write("s\n"); // show previous
		sprintf(m_TmpBuf,"/Courier findfont %g scalefont setfont\n",m_FontSize); Write(m_TmpBuf); m_bFontSizeChanged=false;
	}
	return res;
}
void CVMOutputEPS::SetLineWidth(double w) {
	CVMOutput::SetLineWidth(w);
	Write("s\n"); // show previous
	sprintf(m_TmpBuf,"%g Mabswid\n",m_dLineWidth); Write(m_TmpBuf); m_bLineWidthChanged=false;
}
void CVMOutputEPS::SetLineDash(int nElements, double* len) {
	Write("s\n"); // show previous
	if(nElements>0){ if(len==NULL){ nElements=0; }else if(len[0]<0.){ nElements=0; } } //set line solid
	if(nElements>=2) {
		int lBuf=sprintf(m_TmpBuf,"[ ");
		for(int i=0;i<nElements;i++){ double lI=len[i]; if(lI<0.)lI=len[0]; lBuf+=sprintf(m_TmpBuf+lBuf,"%.2g ",lI); }
		lBuf+=sprintf(m_TmpBuf+lBuf,"] 0 setdash\n"); Write(m_TmpBuf);
	}else{ sprintf(m_TmpBuf,"[ ] 0 setdash\n"); Write(m_TmpBuf); }
	//[Old]if (l1>=0) { if (l2<0) l2=l1; sprintf(m_TmpBuf,"[ %.2g %.2g ] 0 setdash\n",l1,l2); Write(m_TmpBuf); }
	//[Old]else { sprintf(m_TmpBuf,"[ ] 0 setdash\n"); Write(m_TmpBuf); }
}
void CVMOutputEPS::SetColor(CVMRGB rgb) {
	Write("s\n"); // show previous
	if(m_clrDraw!=rgb){
		CVMOutput::SetColor(rgb);
		sprintf(m_TmpBuf,"%.4g %.4g %.4g r\n",(double)rgb.m_Red,(double)rgb.m_Green,(double)rgb.m_Blue); Write(m_TmpBuf);
	}
}
void CVMOutputEPS::FillColor(double x1, double y1, double x2, double y2, CVMRGB rgb) {
	//set color
	Write("s\n"); // show previous
	sprintf(m_TmpBuf,"%.4g %.4g %.4g r\n",(double)rgb.m_Red,(double)rgb.m_Green,(double)rgb.m_Blue); Write(m_TmpBuf);
	sprintf(m_TmpBuf,"%.6g %.6g m\n%.6g %.6g L\n%.6g %.6g L\n%.6g %.6g L\nfill\n",x1,y1,x1,y2,x2,y2,x2,y1); Write(m_TmpBuf);
	//restore color
	sprintf(m_TmpBuf,"%.4g %.4g %.4g r\n",(double)m_clrDraw.m_Red,(double)m_clrDraw.m_Green,(double)m_clrDraw.m_Blue); Write(m_TmpBuf);
}
void CVMOutputEPS::FillContour(CVMRGB rgb) {
	sprintf(m_TmpBuf,"%.4g %.4g %.4g r\nfill\n",(double)rgb.m_Red,(double)rgb.m_Green,(double)rgb.m_Blue); Write(m_TmpBuf);
	//restore color
	sprintf(m_TmpBuf,"%.4g %.4g %.4g r\n",(double)m_clrDraw.m_Red,(double)m_clrDraw.m_Green,(double)m_clrDraw.m_Blue); Write(m_TmpBuf);
}
void CVMOutputEPS::BlockEnd() {
	Write("s\n");
}

/////////////////////////////////////////////////////////////////////// CVMDrawLib
void CVMDrawLib::Arrow(CVMOutput& out, double x1, double y1, double x2, double y2, double sz, int iStyle, bool bFillBcgr) {
	double xMin=0.; double xMax=1.; double yMin=0.; double yMax=1.; out.GetMargins(&xMin, &xMax, &yMin, &yMax);
	if(sz<0.)sz=(xMax-xMin)/25;
	if(iStyle<0)iStyle=0;
	double dx=(x2-x1); double dy=(y2-y1); double r=sqrt(fabs(dx*dx+dy*dy));
	double x1tmp,y1tmp,fShape,fShape2;
	if(bFillBcgr){ out.BlockEnd(); }
	if(r>0.){
		out.MoveTo(x1,y1);
		switch(iStyle) {
		case 0:
			fShape=0.2; out.LineTo(x2,y2);
			x1tmp=x2-dx*sz/r+dy*sz*fShape/r; y1tmp=y2-dy*sz/r-dx*sz*fShape/r; out.MoveTo(x1tmp,y1tmp); out.LineTo(x2,y2);
			x1tmp=x2-dx*sz/r-dy*sz*fShape/r; y1tmp=y2-dy*sz/r+dx*sz*fShape/r; out.MoveTo(x1tmp,y1tmp); out.LineTo(x2,y2);
			break;
		case 1:
			fShape=0.1; fShape2=0.3; x1tmp=x2-dx*sz/r; y1tmp=y2-dy*sz/r;
			out.MoveTo(x1+dy*sz*fShape/r,y1-dx*sz*fShape/r); out.LineTo(x1tmp+dy*sz*fShape/r,y1tmp-dx*sz*fShape/r);
			out.LineTo(x1tmp+dy*sz*fShape2/r,y1tmp-dx*sz*fShape2/r); out.LineTo(x2,y2);
			out.MoveTo(x1+dy*sz*fShape/r,y1-dx*sz*fShape/r); out.LineTo(x1-dy*sz*fShape/r,y1+dx*sz*fShape/r);
			out.LineTo(x1tmp-dy*sz*fShape/r,y1tmp+dx*sz*fShape/r); out.LineTo(x1tmp-dy*sz*fShape2/r,y1tmp+dx*sz*fShape2/r); out.LineTo(x2,y2);
			break;
		case 2:
		default:
			fShape=0.2; fShape2=0.3; x1tmp=x2-dx*sz/r; y1tmp=y2-dy*sz/r;
			out.MoveTo(x1+dy*sz*fShape/r,y1-dx*sz*fShape/r); out.LineTo(x1tmp+dy*sz*fShape/r,y1tmp-dx*sz*fShape/r);
			out.LineTo(x1tmp+dy*sz*fShape2/r,y1tmp-dx*sz*fShape2/r); out.LineTo(x2,y2);
			out.MoveTo(x1+dy*sz*fShape/r,y1-dx*sz*fShape/r); out.LineTo(x1-dy*sz*fShape/r,y1+dx*sz*fShape/r);
			out.LineTo(x1tmp-dy*sz*fShape/r,y1tmp+dx*sz*fShape/r); out.LineTo(x1tmp-dy*sz*fShape2/r,y1tmp+dx*sz*fShape2/r); out.LineTo(x2,y2);
			//break;
		}
	}
	if(bFillBcgr){
		out.FillContour();
		Arrow(out,x1,y1,x2,y2,sz,iStyle,false); //redraw arrow
	}
}

double CVMDrawLib::fAxisY		= -1.;							//[def: -1 for '0 at bottom' (+1 for '0 at top')]
bool CVMDrawLib::bParseTeX		= false;						//[def: false]
double CVMDrawLib::szShadow		= 0;							//[def: 0]
CVMRGB CVMDrawLib::clrShadow	= CVMRGB(0.5f,0.5f,0.5f);		//[def: (0.5,0.5,0.5)]
double CVMDrawLib::gapLines		= 0.6;							//[def: 0.6] distance between lines in high-of-letter units
void CVMDrawLib::Box(CVMOutput& out, double x1, double y1, double dx, double dy, const char* text) {
	if (CVMDrawLib::szShadow>0.) {
		double d=CVMDrawLib::szShadow;
		if(CVMDrawLib::fAxisY >= 0.){
			out.FillColor(x1+d,y1+dy,x1+d+dx,y1+dy+d,CVMDrawLib::clrShadow);
			out.FillColor(x1+dx,y1+d,x1+dx+d,y1+dy+d/100.,CVMDrawLib::clrShadow);
		}else{
			out.FillColor(x1+d,y1-d,x1+d+dx,y1,CVMDrawLib::clrShadow);
			out.FillColor(x1+dx,y1-d/100.,x1+dx+d,y1+dy-d,CVMDrawLib::clrShadow);
		}
	}
	//out.SetColorBackground(CVMOutput::clrWhite);
	out.Rect(x1,y1,x1+dx,y1+dy,true);
	if(text != NULL) {
		int taPrevX=out.GetTextAlignX(); if(taPrevX!=CVMOutput::TA_CENTER){ out.SetTextAlignX(CVMOutput::TA_CENTER); }
		int taPrevY=out.GetTextAlignY(); if(taPrevY!=CVMOutput::TA_CENTER){ out.SetTextAlignY(CVMOutput::TA_CENTER); }
		if(CVMDrawLib::bParseTeX) {
			out.TextOutAdv(x1+dx/2,y1+dy/2,text,true);
		}else{
			long nLines=1; long i; double hL=out.GetLetterHeight();
			for(i=0;;i++){ if(*(text+i)=='\0')break; if(*(text+i)==0x0a){ nLines++; } }
			double yLn=y1+dy/2-CVMDrawLib::fAxisY*(hL*(1.+CVMDrawLib::gapLines)*(nLines-1))/2;
			long pos0=0;
			for(i=0;;i++){ //parse lines
				if(*(text+i)=='\0')break;
				if(*(text+i)==0x0a){
					long l=i-pos0; if((i>0)&&(*(text+i-1)==0x0d)){ l--; }
					out.TextOut(x1+dx/2,yLn,(text+pos0),l); yLn+=CVMDrawLib::fAxisY*hL*(1.+CVMDrawLib::gapLines); pos0=i+1;
				}
			}
			out.TextOut(x1+dx/2,yLn,(text+pos0));
		}
		if(taPrevX!=CVMOutput::TA_CENTER){ out.SetTextAlignX(taPrevX); } if(taPrevY!=CVMOutput::TA_CENTER){ out.SetTextAlignY(taPrevY); }
	}
}

/////////////////////////////////////////////////////////////////////// CVMOutputTest
void CVMOutputTest::LetterSizeTest(CVMOutput* out) {
	std::cout<<"CVMOutputTest::LetterSizeTest"<<std::endl;
	if(out!=NULL) {
		double lw=out->GetLetterWidth(); double lh=out->GetLetterHeight();
		double xMin,xMax,yMin,yMax; out->GetMargins(&xMin,&xMax,&yMin,&yMax);
		double xC=(xMin+xMax)/2.; double yC=(yMin+yMax)/2.;
		char buf[1000]; int i;
		for(i=0;i<10;i++) {
			sprintf(buf,"%c",('A'+i));
			for(int j=0;j<10;j++){
				out->TextOut(xC*1.5+lw*i,yC/2+lh*j,buf);
			}
		}
		for(i=0;i<10;i++) {
			sprintf(buf,"%c",('a'+i));
			for(int j=0;j<10;j++){
				out->TextOut(xC*1.5+lw*i,yC*1.25+lh*j,buf);
			}
		}
	}
}
void CVMOutputTest::TestTextSizeAndAlign(CVMOutput* out) {
	std::cout<<"CVMOutputTest::TestTextSizeAndAlign"<<std::endl;
	if(out==NULL)return;
	//
	out->SetFontSize(8);
	double xMin,xMax,yMin,yMax; out->GetMargins(&xMin,&xMax,&yMin,&yMax);
	double lw=out->GetLetterWidth(); double lh=out->GetLetterHeight();
	double xCol1=(3*xMin+xMax)/4; double xCol2=(xMin+xMax)/2; double xCol3=(xMin+3*xMax)/4;
	double r=lh/8;

	bool bUseMathMacrosOrig=out->UseMathTextMacros(true);
	for(int i=0;i<4;i++){
		int iPlane=0;
		if((i==1)||(i==3)){ out->UseMathTextMacros(false); iPlane=1; }
		double yRow1=lh*2+iPlane*(yMin+yMax)/2; double yRow2=lh*5+iPlane*(yMin+yMax)/2; double yRow3=lh*8+iPlane*(yMin+yMax)/2;
		if((i==2)||(i==3)){ out->SetTextDirection(90); out->SetColor(CVMOutput::clrLawnGreen); }
		out->SetFontSize(8);

		out->SetTextAlignY(CVMOutput::TA_CENTER);
		out->Circle(xCol1,yRow1,r); out->FillContour(CVMOutput::clrGreen);
		out->SetTextAlignX(CVMOutput::TA_CENTER); out->TextOut(xCol1,yRow1,"xCyC");
		out->Circle(xCol1,yRow2,r); out->FillContour(CVMOutput::clrGreen);
		out->SetTextAlignX(CVMOutput::TA_RIGHT); out->TextOut(xCol1,yRow2,"xRyC");
		out->Circle(xCol1,yRow3,r); out->FillContour(CVMOutput::clrGreen);
		out->SetTextAlignX(CVMOutput::TA_LEFT); out->TextOut(xCol1,yRow3,"xLyC");

		out->SetTextAlignX(CVMOutput::TA_CENTER);
		out->Circle(xCol2,yRow1,r); out->FillContour(CVMOutput::clrGreen);
		out->SetTextAlignY(CVMOutput::TA_CENTER); out->TextOut(xCol2,yRow1,"xCyC");
		out->Circle(xCol2,yRow2,r); out->FillContour(CVMOutput::clrGreen);
		out->SetTextAlignY(CVMOutput::TA_UP); out->TextOut(xCol2,yRow2,"xCyU");
		out->Circle(xCol2,yRow3,r); out->FillContour(CVMOutput::clrGreen);
		out->SetTextAlignY(CVMOutput::TA_BOTTOM); out->TextOut(xCol2,yRow3,"xCyB");

		out->SetTextAlignX(CVMOutput::TA_CENTER); out->SetTextAlignY(CVMOutput::TA_UP);
		out->Circle(xCol3,yRow1,r); out->FillContour(CVMOutput::clrGreen);
		out->SetFontSize(5); out->TextOut(xCol3,yRow1,"Size 5");
		out->Circle(xCol3,yRow2,r); out->FillContour(CVMOutput::clrGreen);
		out->SetFontSize(15); out->TextOut(xCol3,yRow2,"Size 15");
		if((i==2)||(i==3)){ out->SetTextDirection(0); out->SetColor(CVMOutput::clrBlack); }
	}
	out->UseMathTextMacros(bUseMathMacrosOrig);
}

#endif //#ifdef VM_HEP_LIB_FULL
