////////////////////////////////////////////////////////
//	HEPEvtLite.cpp
//	Author: V.Makarenko (Vladimir.Makarenko@cern.ch)
////////////////////////////////////////////////////////

#include <math.h> // for sqrt() in mass()
#include <iostream>
#include <fstream>
#include <iomanip>

#include "VMLib/include/vmlib.h"

///////////////////////////////////////////// Utilites
//Library functions
bool CVMUtil::strEqu(const char* a, const char* b) {
	for(long i=0;;i++){ if(*(a+i)!=*(b+i))return false; if(*(a+i)=='\0')break; } return true;
}
long CVMUtil::strLen(const char* a) {
	for(long i=0;i<1000000000;i++){ if(*(a+i)=='\0')return i; } return -1;
}
char* CVMUtil::strCpy(const char* a) {
	if(a==NULL)return NULL; long s=CVMUtil::strLen(a); if(s<0)return NULL;
	char* res=CVMUtil::strAlloc(s);
	memcpy(res,a,s+1); //*(res+s)='\0';
	return res;
}
char* CVMUtil::strAlloc(long s) { if(s<0)s=0; return (char*)malloc(s+2); }
void CVMUtil::strFree(char* a) { if(a!=NULL)free(a); }
//
bool CVMUtil::bIsSymbol(char c) { return (((c>='a')&&(c<='z')) || ((c>='A')&&(c<='Z'))); }
bool CVMUtil::bIsNumber(char c) { return ((c>='0')&&(c<='9')); }
//
void CVMUtil::Cout(std::ostream* os, tReal val) { //auto-select form ('0.000234' or '2.34^-10')
	if(os==NULL)return;
	int fOld = os->flags(); int iPrec=os->precision();
	double fMin=1.; for(int i=0;i<(iPrec-1);i++){ fMin/=10.; }
	double fMax=1./fMin; tReal fAbs=((val>=0)?val:-val);
	if((fAbs<fMin) || (fAbs>=fMax)) {
		os->setf(std::ios_base::scientific,fOld); *os<<val;
	}else{
		os->setf(std::ios_base::fixed,fOld); *os<<val;
	}
	os->setf(fOld);
}

/*
*	--- CVMFileIn, CVMFileOut ---
*	[Used to save lines for stream operations]
*/
CVMFileIn::CVMFileIn(const char* fn, bool bBinary) {
	m_bBinaryAsDouble=false; m_bBinary=bBinary; m_bReadOK=true;
	if(fn!=NULL){ m_ifs.open(fn,std::ios::in|std::ios::binary); }
}
CVMFileIn::~CVMFileIn() { if(m_ifs.is_open()){m_ifs.close();} }
//bool CVMFileIn::IsReady() { return (m_ifs.is_open() && m_ifs.good()); }
CVMFileIn& CVMFileIn::operator>>(char& c) {
	if(IsReady()) {
		m_ifs.read((char*)&c, sizeof(c));
		//[Skips spaces]: m_ifs>>c;
		if(!m_ifs.good() || m_ifs.eof())m_bReadOK=false; //TODO: check this check!
	}else{ m_bReadOK=false; } return *this;
}
CVMFileIn& CVMFileIn::operator>>(long& val) {
	if(IsReady()) {
		if(m_bBinary){ m_ifs.read((char*)&val, sizeof(val)); }else{ m_ifs>>val; }
		if(!m_ifs.good() || m_ifs.eof())m_bReadOK=false; //TODO: check this check!
	}else{ m_bReadOK=false; } return *this;
}
CVMFileIn& CVMFileIn::operator>>(unsigned long& val) { long tmp; *this>>tmp; val=tmp; return *this; }
CVMFileIn& CVMFileIn::operator>>(int& val) { long tmp; *this>>tmp; val=tmp; return *this; }
CVMFileIn& CVMFileIn::operator>>(unsigned int& val) { long tmp; *this>>tmp; val=tmp; return *this; }
CVMFileIn& CVMFileIn::operator>>(double& val) { //Additional operation is 'tReal' != 'double'
	if(IsReady()) {
		if(m_bBinary){ m_ifs.read((char*)&val, sizeof(val)); }else{ m_ifs>>val; }
		if(!m_ifs.good() || m_ifs.eof())m_bReadOK=false; //TODO: check this check!
	}else{ m_bReadOK=false; } return *this;
}
#ifndef tReal_EQU_double
CVMFileIn& CVMFileIn::operator>>(tReal& val) { ReadReal(val,m_bBinary); return *this; }
#endif //tReal_EQU_double
bool CVMFileIn::ReadReal(tReal& val, bool bBinary) {
	if(IsReady()) {
		if(bBinary){
			if(m_bBinaryAsDouble && (sizeof(tReal)!=sizeof(double))) { //Load numbers as 'double'
				double tmp; m_ifs.read((char*)&tmp,sizeof(tmp)); val=(tReal)tmp;
			}else{ m_ifs.read((char*)&val, sizeof(val)); }
		}else{ m_ifs>>val; }
		if(!m_ifs.good() || m_ifs.eof())m_bReadOK=false; //TODO: check this check!
	}else{ m_bReadOK=false; }
	return m_bReadOK;
}
bool CVMFileIn::ReadString(char* buf, long szBuf, const char* endMark) { //skips 'endMark'
	if((endMark==NULL) || (*endMark=='\0'))endMark="\n";
	long i; char c; long szEndMark=0; for(i=0;i<1000;i++){ if(*(endMark+i)=='\0'){ szEndMark=i; break; } }
	if(szEndMark<=0)return false;
	if((buf==NULL)||(szBuf<=(szEndMark+2)))return false;
	if(!IsReady()) { m_bReadOK=false; return false; }
	for(i=0;;i++){
		if(i>=szBuf){ m_bReadOK=false; break; }
		m_ifs.read((char*)&c,sizeof(c)); //[Old:] m_ifs>>c; <--skips spaces
		*(buf+i)=c;
		if(!m_ifs.good() || m_ifs.eof()){ m_bReadOK=false; break; } //TODO: check this check!
		if(i>=(szEndMark-1)){
			bool bFitMark=true; for(int j=0;j<szEndMark;j++){ if(*(buf+i-j)!=*(endMark+szEndMark-1-j)){bFitMark=false;break;} }
			if(bFitMark){ *(buf+i+1-szEndMark)='\0'; break; }
		}
	}
	return m_bReadOK;
}
bool CVMFileIn::ReadNotEmptySymbol(char& c) { //skips spaces, tabs and line breaks
	if(!IsReady()){ m_bReadOK=false; return false; }
	bool bResOK=true;
	for(long i=0;;i++){
		char cTmp=0;
		m_ifs.read((char*)&cTmp,sizeof(cTmp)); //[Old, also good (skips spaces)]: m_ifs>>cTmp;
		if((cTmp!=' ')&&(cTmp!='\t')&&(cTmp!=0x0a)&&(cTmp!=0x0d)&&(cTmp!='\0')){ c=cTmp; break; }
		if(!m_ifs.good() || m_ifs.eof() || (i>1000000)){ m_bReadOK=false; bResOK=false; break; } //TODO: check this check!
	}
	return bResOK;
}
//
CVMFileOut::CVMFileOut(const char* fn, bool bKeepOpened, bool bBinary) {
	m_ofs=NULL; m_bBinaryAsDouble=false; m_filename=((fn!=NULL)?CVMUtil::strCpy(fn):NULL);
	m_bKeepOpened=bKeepOpened; m_bBinary=bBinary;
	if(m_filename!=NULL){ //Truncate/open file
		m_ofs=new std::ofstream(m_filename,std::ios::out|std::ios::trunc|std::ios::binary);
		if(!m_bKeepOpened){ m_ofs->close(); delete m_ofs; m_ofs=NULL; }
	}else{
		m_bKeepOpened=false;
	}
}
CVMFileOut::~CVMFileOut() { if(m_ofs!=NULL){ if(m_ofs->is_open()){m_ofs->close();} delete m_ofs; m_ofs=NULL; } }
bool CVMFileOut::PrepareStream() { //creates stream if !m_bKeepOpened
	if((!m_bKeepOpened) && (m_filename!=NULL)){
		if(m_ofs==NULL){ m_ofs=new std::ofstream(m_filename,std::ios::out|std::ios::app|std::ios::binary); }
		else if(!m_ofs->is_open()){ m_ofs->open(m_filename,std::ios::out|std::ios::app|std::ios::binary); }
	}
	return ((m_ofs!=NULL)&&(m_ofs->is_open()));
}
void CVMFileOut::ReleaseStream() { //clears stream if !m_bKeepOpened
	if((!m_bKeepOpened)&&(m_ofs!=NULL)){ m_ofs->close(); delete m_ofs; m_ofs=NULL; }
}
bool CVMFileOut::Open() {
	bool bRes=false;
	if(m_ofs!=NULL)Close();
	if(m_filename!=NULL){
		m_ofs=new std::ofstream(m_filename,std::ios::out|std::ios::app|std::ios::binary);
		bRes=m_ofs->is_open(); m_bKeepOpened=true;
	}
	return bRes;
}
void CVMFileOut::Close() {
	if(m_ofs!=NULL){ m_ofs->close(); delete m_ofs; m_ofs=NULL; } m_bKeepOpened=false;
}
bool CVMFileOut::IsReady() { return ((m_ofs!=NULL) && (m_ofs->is_open())); }
bool CVMFileOut::WriteReal(tReal val, bool bBinary) {
	if(!PrepareStream())return false;
	if(bBinary){
		if(m_bBinaryAsDouble && (sizeof(tReal)!=sizeof(double))) { //Save numbers as 'double'
			double tmp=DBLtR(val); m_ofs->write((char*)&tmp,sizeof(tmp));
		}else{ m_ofs->write((char*)&val,sizeof(val)); }
	}else{ *m_ofs<<val; }
	ReleaseStream();
	return true;
	//[Old] bool bRes=false;
	//[Old] std::ofstream* ofs=NULL; bool bDelOfs=false; if(m_ofs!=NULL){ if(m_ofs->is_open()){ ofs=m_ofs; } }
	//[] if((ofs==NULL) && (m_filename!=NULL)){ ofs=new std::ofstream(m_filename,std::ios::out|std::ios::app|std::ios::binary); bDelOfs=true; }
	//[Old] if(ofs!=NULL){
	//[Old] 	if(*ofs){
	//[Old] 		if(bBinary){
	//[Old] 			if(m_bBinaryAsDouble && (sizeof(tReal)!=sizeof(double))) { //Save numbers as 'double'
	//[Old] 				double tmp=DBLtR(val); ofs->write((char*)&tmp,sizeof(tmp));
	//[Old] 			}else{ ofs->write((char*)&val,sizeof(val)); }
	//[Old] 		}else{ *ofs<<val; }
	//[Old] 		bRes=true;
	//[Old] 	}
	//[Old] 	if(bDelOfs){ ofs->close(); delete ofs; }
	//[Old] }
	//[Old] return bRes;
}
bool CVMFileOut::WriteBuffer(const char* buf) { //opens+closes file if not open
	if(buf==NULL)return false;
	if(!PrepareStream())return false;
	bool bRes=true;
	for(long i=0;;i++){ if(i>=100000000){bRes=false;break;} char c=*(buf+i); if(c=='\0')break; m_ofs->write((char*)&c,sizeof(c)); }
	ReleaseStream();
	return bRes;
	//[Old] bool bRes=false;
	//[Old] std::ofstream* ofs=NULL; bool bDelOfs=false; if(m_ofs!=NULL){ if(m_ofs->is_open()){ ofs=m_ofs; } }
	//[Old] if((ofs==NULL) && (m_filename!=NULL)){ ofs=new std::ofstream(m_filename,std::ios::out|std::ios::app|std::ios::binary); bDelOfs=true; }
	//[Old] if(ofs!=NULL){
	//[Old] 	if(*ofs){
	//[Old] 		bRes=true;
	//[Old] 		for(long i=0;;i++){ if(i>=100000000){bRes=false;break;} char c=*(buf+i); if(c=='\0')break; ofs->write((char*)&c,sizeof(c)); }
	//[Old] 	}
	//[Old] 	if(bDelOfs){ ofs->close(); delete ofs; }
	//[Old] }
	//[Old] return bRes;
}
CVMFileOut& CVMFileOut::operator<<(long val) {
	if(PrepareStream()){ if(m_bBinary){m_ofs->write((char*)&val,sizeof(val));}else{*m_ofs<<val;} }
	ReleaseStream(); return *this;
}
//CVMFileOut& CVMFileOut::operator<<(unsigned long val) { return (*this)<<((long)val); }
CVMFileOut& CVMFileOut::operator<<(double val) { //Additional operation is 'tReal' != 'double'
	if(PrepareStream()){ if(m_bBinary){ m_ofs->write((char*)&val,sizeof(val)); }else{ *m_ofs<<val; } }
	ReleaseStream(); return *this;
}
#ifndef tReal_EQU_double
CVMFileOut& CVMFileOut::operator<<(tReal val) { WriteReal(val,m_bBinary); return *this; }
#endif //tReal_EQU_double
CVMFileOut& CVMFileOut::operator<<(const char* str) { WriteBuffer(str); return *this; }
